/**
 *
 * $Id: Include.c,v 1.2 1996/08/14 05:43:58 u27113 Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

#include <stdio.h>
#include <X11/Intrinsic.h>

#include <sys/types.h>
#include <sys/param.h>

#include <XmI/DebugUtil.h>

static char *rcsid = "$Id: Include.c,v 1.2 1996/08/14 05:43:58 u27113 Exp $";

static char **IncludeDirs = NULL;
static int nDirs = 0;

void
IncludeAddDirectory(char *d)
{
/* Special case to handle current directory */
	if (nDirs == 0) {
		nDirs++;
		IncludeDirs = (char **) XtRealloc((XtPointer)IncludeDirs, nDirs * sizeof(char *));
		IncludeDirs[nDirs-1] = XtNewString(".");
	}

	nDirs++;
	IncludeDirs = (char **) XtRealloc((XtPointer)IncludeDirs, nDirs * sizeof(char *));
	IncludeDirs[nDirs-1] = XtNewString(d);
}

FILE *
IncludeOpenFile(char *f)
{
	int	i;
	FILE	*r;
	char	*b;

	if (nDirs == 0) {
		nDirs++;
		IncludeDirs = (char **) XtRealloc((XtPointer)IncludeDirs, nDirs * sizeof(char *));
		IncludeDirs[nDirs-1] = XtNewString(".");
	}

	if (f == NULL)
		return NULL;
	if (f[0] == '/')
		return fopen(f, "r");
	b = XtMalloc(MAXPATHLEN);
	for (i=0; i<nDirs; i++) {
		sprintf(b, "%s/%s", IncludeDirs[i], f);
		if ((r = fopen(b, "r")) != NULL) {
			XtFree(b);
			return r;
		}
	}
	XtFree(b);
	return NULL;
}

char *
IncludeFindFile(char *f)
{
	int	i;
	FILE	*r;
	char	*b;

	if (nDirs == 0) {
		nDirs++;
		IncludeDirs = (char **) XtRealloc((XtPointer)IncludeDirs, nDirs * sizeof(char *));
		IncludeDirs[nDirs-1] = XtNewString(".");
	}

	XdbDebug(__FILE__, NULL, "IncludeFindFile(%s)\n", f);

	if (f == NULL)
		return NULL;
	if (f[0] == '/')
		return XtNewString(f);
	b = XtMalloc(MAXPATHLEN);
	for (i=0; i<nDirs; i++) {
		sprintf(b, "%s/%s", IncludeDirs[i], f);
		if ((r = fopen(b, "r")) != NULL) {
			XdbDebug(__FILE__, NULL, "IncludeFindFile => %s\n", b);
			fclose(r);
			return b;
		} else {
			XdbDebug(__FILE__, NULL, "IncludeFindFile : not in %s\n", IncludeDirs[i]);
		}
	}
	XtFree(b);
	return NULL;
}
