/**
 *
 * $Id: Uil.c,v 1.2 1996/08/14 05:44:00 u27113 Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#include <Xm/Xm.h>
#include <XmI/DebugUtil.h>

#include "UilMrm.h"
#include "Uil.h"
#include "Include.h"
#include "Identifier.h"
#include "Object.h"
#include "Procedures.h"
#include "Module.h"
#include "Arguments.h"

static char *rcsid = "$Id: Uil.c,v 1.2 1996/08/14 05:44:00 u27113 Exp $";

int yydebug;
extern FileData Files[];
int yyparse (void);

int IDNumber = 1;
static char OutputFileName[256];

uil_global_t	*UilGlobal;

int
main(int argc, char **argv)
{
  int Generate_Binary_Code = False;
  int c;
  FILE *input;
  
  OutputFileName[0] = 0;
  yydebug = 0;
  UilGlobal = (uil_global_t *)XtMalloc(sizeof(uil_global_t));

  while (-1 != (c = getopt(argc, argv, "dcI:mo:sv:w"))) {
    switch (c) {
    case 'c':
    case 'd':
      yydebug = 1;
      break;
    case 'I':	
      XdbDebug(__FILE__, NULL, "Include %s\n",optarg);
      IncludeAddDirectory(optarg);
      break;
    case 'm':

      Generate_Binary_Code = True;
      break;
    case 'o':	
      strcpy(OutputFileName,optarg);
      break;
    case 's':	
      fprintf(stderr,"set locale\n");
      break;
    case 'V':	
      puts(rcsid);
      exit(0);
      break;
    case 'v':	
      fprintf(stderr,"generate listing to %s\n",
	      optarg);
      break;
    case 'w':	
      fprintf(stderr,"surpress all warnings\n");
      break;
    case 'W':	
      fprintf(stderr,"WMD\n");
      break;
    case '?':
      exit(1);
      break;
    default:
      fprintf(stderr,"?? getopt returned character code 0%o \n",c);
      exit(1);
      break;
    }
  }

  if (optind >= argc) {
    fprintf(stderr, "\n\nsevere: no source file specified\n");
    exit(1);
  }

  input = freopen(argv[optind],"r",stdin);
  if (input == NULL){
    fprintf(stderr,"\n\nsevere: error opening source file: %s\n",
	    argv[optind]);
    exit(1);
  }

  strcpy(Files[0].Name,argv[optind]);
  Files[0].lineno = 1;
  if (OutputFileName[0] == '\0')
    strcpy(OutputFileName, "a.uid");

/*
 * Parse input
 */
  yyparse();

/*
 * Generate output
 */
  (void) freopen(OutputFileName, "w", stdout);
  fprintf(stdout, "%%LessTif%%UID%%Version %s%%\n", LesstifVERSION_STRING);
  ModulePrint(stdout);
  TreePrint((tree)UilGlobal->Objects, ObjectPrint);
  TreePrint((tree)UilGlobal->Identifiers, IdentifierPrint);
  TreePrint((tree)UilGlobal->Procedures, ProcedurePrint);
  fclose(stdout);

/*
 * Die
 */
  exit(0);
  /*NOTREACHED*/
}
