%{
/**
 *
 * $Id: lex.l,v 1.3 1996/08/14 05:44:01 u27113 Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

/*
 * Originally written by Geoffrey W. Ritchey
 */ 
static char *rcsid = "$Id: lex.l,v 1.3 1996/08/14 05:44:01 u27113 Exp $";

#include <string.h>

#include "UilMrm.h"
#include "Uil.h"

#define YYSTYPE char *

#include "misc.h"
#include "y.tab.h"
#include "Include.h"

#include <XmI/DebugUtil.h>

/* Needed for yacc, not for bison */
extern YYSTYPE yylval;

#define MAX_INCLUDE_DEPTH 10
YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];

FileData Files[MAX_INCLUDE_DEPTH];

char *FileName = Files[0].Name;
int LineNumber = 1;

int include_stack_ptr = 0;
%}

/*
 * Define start condition
 * This thing is to be used in parsing C style comments.
 * The code to parse these comments is borrowed from the flexdoc(1) manual page.
 */
%x comment

%%

<<EOF>> {
	LineNumber = Files[include_stack_ptr].lineno;
	if (--include_stack_ptr < 0) {
		yyterminate();
	}
	else
		yy_switch_to_buffer(include_stack[include_stack_ptr]);
	FileName = Files[include_stack_ptr].Name;
}
		
include[ \t]+file[ \t]+('|\(?\")[^'"]*('|\"\)?)\; {
			char *s;

			if (include_stack_ptr >= MAX_INCLUDE_DEPTH){
				fprintf(stderr,"Includes nested too deeply\n");
				exit(1);
			}
			s = strpbrk(yytext,"'\"")+1;
			s[strcspn(s,"\"'")] = 0;
			include_stack[include_stack_ptr++] = YY_CURRENT_BUFFER;
			strcpy(Files[include_stack_ptr].Name,s);
			Files[include_stack_ptr].lineno = LineNumber;
			FileName = Files[include_stack_ptr].Name;
			LineNumber = 1;
			yyin = IncludeOpenFile(s);
			if (!yyin)
				perror(s);
			yy_switch_to_buffer(yy_create_buffer(yyin,YY_BUF_SIZE));
		}
">>"		{
			return RIGHT_SHIFT;
		}
"<<"		{
			return LEFT_SHIFT;
		}
\"[^"]*\"	{
			yylval = Store((char *)yytext);
			return STRING;
		}
'[^']*'		{
			yytext[strlen(yytext)-1] = yytext[0] = '"';
			yylval = Store(yytext);
			return STRING;
		}
-?[0-9]+\.[0-9]*	{
			yylval = Store(yytext);
			return FLOAT_LITERAL;
		}
-?[0-9]+	{
			yylval = Store(yytext);
			return INTEGER_LITERAL;
		}
any		{
			return ANY;
		}
argument	{
			return ARGUMENT;
		}
arguments	{
			return ARGUMENTS;
		}
asciz_string_table	{
			return ASCIZ_STRING_TABLE;
		}
asciz_table	{
			return ASCIZ_TABLE;
		}
background[ \t]+color	{
			return BACKGROUND_COLOR;
		}
background	{
			return BACKGROUND;
		}
bool		{
			return BOOL;
		}
boolean		{
			return BOOLEAN;
		}
boolean_type	{
			return BOOLEAN;
		}
callbacks	{
			return CALLBACKS;
		}
case_insensitive {
			return CASE_INSENSITIVE;
		}
case_sensitive {
			return CASE_SENSITIVE;
		}
char_set	{
			return CHARACTER_SET;
		}
character_set	{
			return CHARACTER_SET;
		}
color_table	{
			return COLOR_TABLE;
		}
color		{
			return COLOR;
		}
compound_string_component	{
			return COMPOUND_STRING_COMPONENT;
		}
compound_string_table	{
			return COMPOUND_STRING_TABLE;
		}
compound_string	{
			return COMPOUND_STRING;
		}
controls	{
			return CONTROLS;
		}
end 		{
			return END;
		}
exported	{
			return EXPORTED;
		}
false		{
			yylval = 0;
			return T_FALSE;
		}
float		{
			return FLOAT;
		}
font		{
			return FONT;
		}
fontset	{
			return FONTSET;
		}
font_table	{
			return FONT_TABLE;
		}
foreground[ \t]+color	{
			return FOREGROUND_COLOR;
		}
foreground	{
			return FOREGROUND;
		}
gadget		{
			return GADGET;
		}
icon 		{
			return ICON;
		}
identifier	{
			return IDENTIFIER;
		}
imported	{
			return IMPORTED;
		}
integer		{
			return INTEGER;
		}
integer_table	{
			return INTEGER_TABLE;
		}
keysym 		{
			return KEYSYM;
		}
list		{
			return LIST;
		}
managed		{
			return MANAGED;
		}
module		{
			return MODULE;
		}
names		{
			return NAMES;
		}
object		{
			return OBJECT;
		}
objects		{
			return OBJECTS;
		}
private		{
			return PRIVATE;
		}
procedure	{
			return PROCEDURE;
		}
procedures	{
			return PROCEDURES;
		}
reason		{
			return REASON;
		}
rgb		{
			return RGB;
		}
right_to_left	{
			return RIGHT_TO_LEFT;
		}
separate	{
			return SEPARATE;
		}
single_float	{
			return SINGLE_FLOAT;
		}
sixteen_bit	{
			return SIXTEEN_BIT;
		}
string		{
			return STRING;
		}
string_table	{
			return STRING_TABLE;
		}
translation_table {
			return TRANSLATION_TABLE;
		}
true		{
			yylval = (char *)1;
			return T_TRUE;
		}
unmanaged	{
			return UNMANAGED;
		}
user_defined	{
			return USER_DEFINED;
		}
value		{
			return VALUE;
		}
version		{
			return VERSION;
		}
wide_character	{
			return WIDE_CHARACTER;
		}
widget		{
			return WIDGET;
		}
xbitmapfile	{
			return XBITMAPFILE;
		}
xpixmapfile	{
			return XPIXMAPFILE;
		}

[a-zA-Z][a-zA-Z_0-9]* {
			yylval = (char *)Store(yytext);
			return LEX_ID;
		}

"/*"			BEGIN(comment);
<comment>[^*\n]*	/* Eat */
<comment>"*"+[^*/\n]*	/* Eat */
<comment>\n		LineNumber++;
<comment>"*"+"/"	BEGIN(INITIAL);

"!"[^\n]*	{
		}
[ \t]+
\n		LineNumber++;

.		{
			return yytext[0];
		}

%%
