%{
/**
 *
 * $Id: yacc.y,v 1.2 1996/08/14 05:44:02 u27113 Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

/* 
 * Derived from work by Geoffrey W. Ritchey
 */ 

static char *rcsid = "$Id: yacc.y,v 1.2 1996/08/14 05:44:02 u27113 Exp $";

#include <stdio.h>
#include <stdarg.h>

#include "UilMrm.h"
#include "Uil.h"
#include "Tree.h"
#include "Procedures.h"
#include "Object.h"
#include "Module.h"
#include "Identifier.h"
#include "Value.h"
#include "Arguments.h"
#include "UilDebug.h"
#include "Control.h"
#include "Features.h"
#include "Expression.h"

#include <X11/Intrinsic.h>

extern yytext;
extern int LineNumber;
extern char *FileName;

#include <XmI/DebugUtil.h>

%}

%token ANY
%token ARGUMENT
%token ARGUMENTS
%token ASCIZ_STRING_TABLE
%token ASCIZ_TABLE
%token BACKGROUND
%token BACKGROUND_COLOR
%token BOOL
%token BOOLEAN
%token CALLBACKS
%token CASE_INSENSITIVE
%token CASE_SENSITIVE
%token CHARACTER_SET
%token CLASS_REC_NAME
%token COLOR
%token COLOR_TABLE
%token COMPOUND_STRING
%token COMPOUND_STRING_COMPONENT
%token COMPOUND_STRING_TABLE
%token CONTROLS
%token END
%token EXPORTED
%token T_FALSE
%token FLOAT
%token FLOAT_LITERAL
%token FONT
%token FONTSET
%token FONT_TABLE
%token FOREGROUND
%token FOREGROUND_COLOR
%token GADGET
%token ICON
%token LEX_ID
%token IDENTIFIER
%token IMPORTED
%token INTEGER
%token INTEGER_LITERAL
%token INTEGER_TABLE
%token KEYSYM
%token LEFT_SHIFT
%token LIST
%token MANAGED
%token MODULE
%token NAMES
%token OBJECT
%token OBJECTS
%token PIXMAP
%token PRIVATE
%token PROCEDURE
%token PROCEDURES
%token REASON
%token RGB
%token RIGHT_SHIFT
%token RIGHT_TO_LEFT
%token SEPARATE
%token SINGLE_FLOAT
%token SIXTEEN_BIT
%token STRING
%token STRING_TABLE
%token TRANSLATION_TABLE
%token T_TRUE
%token UNMANAGED
%token USER_DEFINED
%token VALUE
%token VERSION
%token WIDE_CHARACTER
%token WIDGET
%token XBITMAPFILE
%token XPIXMAPFILE

%%

input:
	MODULE ID module_header module_body END MODULE ';'
		{
			ModuleName($2);
		}
;

module_header:
	  VERSION '=' STRING module_header
		{
			ModuleVersion($3);
			XdbDebug(__FILE__, NULL, "ModuleHeader VERSION\n");
		}
	| NAMES '=' CASE_SENSITIVE module_header
		{
			ModuleCaseSensitive(True);
			XdbDebug(__FILE__, NULL, "ModuleHeader NAMES\n");
		}
	| NAMES '=' CASE_INSENSITIVE module_header
		{
			ModuleCaseSensitive(False);
			XdbDebug(__FILE__, NULL, "ModuleHeader NAMES\n");
		}
	| CHARACTER_SET '=' ID module_header
		{
			ModuleCharacterSet($3);
			XdbDebug(__FILE__, NULL, "ModuleHeader CHAR_SET\n");
		}
	| OBJECTS '=' '{' object_list '}' module_header
		{
			XdbDebug(__FILE__, NULL, "ModuleHeader OBJECTS\n");
		}
	|
		{
			/* empty - no action */
		}
;

module_body:
	  module_body VALUE value_sequence
		{
			XdbDebug(__FILE__, NULL, "ModuleBody ValueSection\n");

#if 0
			TreeInsert((tree *)&UilGlobal->Values, (tree)$3);
			$$ = $1;
#endif
		}
	| module_body PROCEDURE procedure_list
		{
			XdbDebug(__FILE__, NULL, "ModuleBody ProcedureSection\n");

			TreeInsert((tree *)&UilGlobal->Procedures, (tree)$3);
			$$ = $1;
		}
	| module_body LIST list
		{
			XdbDebug(__FILE__, NULL, "ModuleBody List\n");

			TreeInsert((tree *)&UilGlobal->Lists, (tree)$3);
			$$ = $1;
		}
	| module_body OBJECT object
		{
			XdbDebug(__FILE__, NULL, "ModuleBody OBJECT\n");

			TreeInsert((tree *)&UilGlobal->Objects, $3);
			$$ = $1;
		}
	| module_body IDENTIFIER identifier_list
		{

#if 0
			TreeInsert((tree *)&UilGlobal->Identifiers, (tree)$3);
#endif
			$$ = $1;

			XdbDebug(__FILE__, NULL, "ModuleBody IDENTIFIER\n");
		}
	|
		{
			$$ = NULL;
		}
;

identifier_list:
	  ID ';' identifier_list
		{
			XdbDebug(__FILE__, NULL, "Identifier %s\n", $2);
			$$ = Identifier($1);
		}
	|
		{
			/* empty - no action */
			$$ = NULL;
		}
;

value_sequence:
	  ID ':' value_expression ';' value_sequence
		{
			Value($1, $3, 0);
		}
	| ID ':' EXPORTED value_expression ';' value_sequence
		{
			Value($1, $4, UIL_VALUE_EXPORTED);
		}
	| ID ':' PRIVATE value_expression ';' value_sequence
		{
			Value($1, $4, UIL_VALUE_PRIVATE);
		}
	| ID ':' IMPORTED value_type ';' value_sequence
		{
			ValueImport($1, $4);
		}
	|
		{
			/* empty - no action */
		}
;

value_type:
	  ANY
		{
			$$ = t_any;
		}
	| ARGUMENT
		{
			$$ = t_argument;
		}
	| BOOLEAN
		{
			$$ = t_boolean;
		}
	| COLOR
		{
			$$ = t_color;
		}
	| COLOR_TABLE
		{
			$$ = t_color_table;
		}
	| COMPOUND_STRING
		{
			$$ = t_compound_string;
		}
	| FLOAT
		{
			$$ = t_float;
		}
	| FONT
		{
			$$ = t_font;
		}
	| FONT_TABLE
		{
			$$ = t_font_table;
		}
	| FONTSET
		{
			$$ = t_fontset;
		}
	| ICON
		{
			$$ = t_icon;
		}
	| INTEGER
		{
			$$ = t_integer;
		}
	| INTEGER_TABLE
		{
			$$ = t_integer_table;
		}
	| KEYSYM
		{
			$$ = t_keysym;
		}
	| REASON
		{
			$$ = t_reason;
		}
	| SINGLE_FLOAT
		{
			$$ = t_single_float;
		}
	| STRING
		{
			$$ = t_string;
		}
	| STRING_TABLE
		{
			$$ = t_string_table;
		}
	| TRANSLATION_TABLE
		{
			$$ = t_translation_table;
		}
	| WIDE_CHARACTER
		{
			$$ = t_wide_character;
		}
	| WIDGET
		{
			$$ = t_widget;
		}
;

ID:
	  LEX_ID
;

expression:	value_expression;

value_expression :
	mult
	| value_expression '-' mult
		{
		}
	| value_expression '+' mult
		{
		}
	| value_expression RIGHT_SHIFT mult
		{
		}
	| value_expression LEFT_SHIFT mult
		{
		}
	| value_expression '&' mult
		{
		}
	| value_expression '|' mult
		{
		}
;

mult:
	prim
	| mult '*' prim
	| mult '/' prim
	| mult '^' prim
		{
		}
;

prim:
	  COLOR_TABLE '(' color_expression_list ')'
		{
		}
	| '~' prim
		{
		}
	| '-' prim
		{
		}
	| '+' prim
		{
		}
	| ICON '(' COLOR_TABLE '=' ID ',' row_sequence ')'
		{
		}
	| ICON '(' row_sequence ')'
		{
		}
	| boolean_literal
		{
			XdbDebug(__FILE__, NULL, "ValueExpression BooleanLiteral\n");
			$$ = $1;
		}
	| INTEGER_LITERAL
		{
			XdbDebug(__FILE__, NULL, "ValueExpression IntegerLiteral %s\n", $1);
			$$ = IntegerLiteral($1);
		}
	| FLOAT_LITERAL
		{
			$$ = FloatLiteral($1);
		}
	| STRING
		{
			$$ = StringLiteral($1);
		}
	| COMPOUND_STRING '(' expression ',' property_list ')'
		{
			XdbDebug(__FILE__, NULL, "CompoundString with property list\n");
		}
	| COMPOUND_STRING '(' expression ')'
		{
			XdbDebug(__FILE__, NULL, "CompoundString\n");
		}
	| COMPOUND_STRING_COMPONENT '(' component_type ',' ID ')'
		{
		}
	| COMPOUND_STRING_COMPONENT '(' component_type ')'
		{
		}
	| COMPOUND_STRING_TABLE '(' expression_list ')'
		{
		}
	| ASCIZ_TABLE '(' expression_list ')'
		{
		}
	| ASCIZ_STRING_TABLE '(' expression_list ')'
		{
		}
	| WIDE_CHARACTER '(' expression ')'
		{
		}
	| CLASS_REC_NAME '(' expression ')'
		{
		}
	| INTEGER_TABLE '(' expression_list ')'
		{
		}
	| ARGUMENT '(' expression ')'
		{
		}
	| ARGUMENT '(' expression ',' argument_type ')'
		{
		}
	| REASON '(' expression ')'
		{
		}
	| TRANSLATION_TABLE '(' expression_list ')'
		{
		}
	| ID
		{
			XdbDebug(__FILE__, NULL, "ValueExpression ID (%s)\n", $1);
		}
	| FONT '(' expression ')'
		{
			XdbDebug(__FILE__, NULL, "ValueExpression FONT (%s)\n", $3);
		}
	| FONT '(' expression ',' CHARACTER_SET '=' character_set ')'
		{
			XdbDebug(__FILE__, NULL, "ValueExpression FONT (%s), charset %s\n", $3, $7);
		}
	| '#' character_set STRING
		{
			XdbDebug(__FILE__, NULL, "#(CharSet %s, String %s)\n", $2, $3);
		}
	| CHARACTER_SET '(' expression ')'
	| CHARACTER_SET '(' expression ',' property_list ')'
	| KEYSYM '(' STRING ')'
	| COLOR '(' expression ')'
	| COLOR '(' expression ',' FOREGROUND ')'
	| COLOR '(' expression ',' BACKGROUND ')'
	| RGB '(' expression ',' expression ',' expression ')'
	| XBITMAPFILE '(' expression ')'
	| XPIXMAPFILE '(' expression ')'
	| SINGLE_FLOAT '(' FLOAT_LITERAL ')'
	| FLOAT '(' FLOAT_LITERAL ')'
	| FONTSET '(' expression_list ')'
	| FONTSET '(' expression_list ',' CHARACTER_SET '=' character_set ')'
	| FONT_TABLE '(' font_expression_list ')'
	| '(' value_expression ')'
		{
		}
;

font_expression_list:
	  font_expression
	| font_expression ',' font_expression_list
;

font_expression:
	  character_set '=' ID
	| FONT '(' expression ')'
	| FONT '(' expression ',' CHARACTER_SET '=' expression ')'
	| FONTSET '(' expression_list ')'
	| FONTSET '(' expression_list ',' CHARACTER_SET '=' expression ')'
;

character_set:
	  ID
		{
			XdbDebug(__FILE__, NULL, "CharacterSet %s\n", $1);
		}
	| CHARACTER_SET '(' ID ')'
		{
			XdbDebug(__FILE__, NULL, "CharacterSet (%s)\n", $3);
		}
;

argument_type:
	  ANY
		{
			$$ = t_any;
		}
	| ASCIZ_STRING_TABLE
		{
			$$ = t_asciz_string_table;
		}
	| BOOLEAN
		{
			$$ = t_boolean;
		}
	| COLOR
		{
			$$ = t_color;
		}
	| COMPOUND_STRING
		{
			$$ = t_compound_string;
		}
	| FLOAT
		{
			$$ = t_float;
		}
	| FONT
		{
			$$ = t_font;
		}
	| FONT_TABLE
		{
			$$ = t_font_table;
		}
	| FONTSET
		{
			$$ = t_fontset;
		}
	| ICON
		{
			$$ = t_icon;
		}
	| INTEGER
		{
			$$ = t_integer;
		}
	| INTEGER_TABLE
		{
			$$ = t_integer_table;
		}
	| KEYSYM
		{
			$$ = t_keysym;
		}
	| PIXMAP
		{
			$$ = t_pixmap;
		}
	| REASON
		{
			$$ = t_reason;
		}
	| SINGLE_FLOAT
		{
			$$ = t_single_float;
		}
	| STRING
		{
			$$ = t_string;
		}
	| STRING_TABLE
		{
			$$ = t_string_table;
		}
	| TRANSLATION_TABLE
		{
			$$ = t_translation_table;
		}
	| WIDE_CHARACTER
		{
			$$ = t_wide_character;
		}
	| WIDGET
		{
			$$ = t_widget;
		}
;

expression_list:
	  expression
		{
			$$ = $1;
		}
	| expression_list ',' expression
		{
			$$ = TreeAppend($1, $3);
		}
;

component_type:
	ID
;

property_list:
	  SEPARATE '=' expression
		{
			XdbDebug(__FILE__, NULL, "Separate = ...\n");
		}
	| RIGHT_TO_LEFT '=' expression
		{
			XdbDebug(__FILE__, NULL, "RightToLeft = ...\n");
		}
	| SEPARATE '=' expression ',' property_list
		{
			XdbDebug(__FILE__, NULL, "Separate = ...\n");
		}
	| RIGHT_TO_LEFT '=' expression ',' property_list
		{
			XdbDebug(__FILE__, NULL, "RightToLeft = ...\n");
		}
	| SIXTEEN_BIT '=' expression
	| SIXTEEN_BIT '=' expression ',' property_list
;

boolean_literal:
	  T_TRUE
		{
			XdbDebug(__FILE__, NULL, "Boolean True\n");
			$$ = BooleanLiteral(True);
		}
	| T_FALSE
		{
			XdbDebug(__FILE__, NULL, "Boolean False\n");
			$$ = BooleanLiteral(False);
		}
;

row_sequence :
	  STRING
	| STRING ',' row_sequence
;

color_expression_list:
	  color_expression '=' STRING
	| color_expression '=' STRING ',' color_expression_list
;

color_expression:
	  BACKGROUND_COLOR
	| FOREGROUND_COLOR
	| STRING
	| color_function
;

color_function:
	  COLOR '(' STRING ')'
	| COLOR '(' STRING ',' BACKGROUND ')'
	| COLOR '(' STRING ',' FOREGROUND ')'
;

object_list:
	  ID '=' WIDGET ';' object_list
		{
			XdbDebug(__FILE__, NULL, "OBJECT_LIST WIDGET %s\n", $1);
		}
	| ID '=' GADGET ';' object_list
		{
			XdbDebug(__FILE__, NULL, "OBJECT_LIST GADGET %s\n", $1);
		}
	|
		{
			XdbDebug(__FILE__, NULL, "OBJECT_LIST (NULL)\n");
		}
;

procedure_list:
	  ID '(' type_list ')' ';' procedure_list
		{
			XdbDebug(__FILE__, NULL, "ProcedureList %s(...)\n", $2);
			$$ = ProcedureListWithType($2, $4);
		}
	| ID '(' ')' ';' procedure_list
		{
			XdbDebug(__FILE__, NULL, "ProcedureList %s()\n", $2);
			$$ = ProcedureListEmpty($2);
		}
	| ID '(' type_list ')' ';'
		{
			XdbDebug(__FILE__, NULL, "ProcedureList %s (...)\n", $1);
			$$ = ProcedureListWithType($1, $3);
		}
	| ID ';'
		{
			XdbDebug(__FILE__, NULL, "ProcedureList %s\n", $1);
			$$ = ProcedureList($1);
		}
	| ID '(' ')' ';' 
		{
			XdbDebug(__FILE__, NULL, "ProcedureList %s()\n", $2);
			$$ = ProcedureListEmpty($2);
		}
;

type_list:
	  type_list ',' type
		{
			XdbDebug(__FILE__, NULL, "NOOP line %d\n", __LINE__);
		}
	| type
		{
			XdbDebug(__FILE__, NULL, "NOOP line %d\n", __LINE__);
		}
;

type:
	  INTEGER
		{
			XdbDebug(__FILE__, NULL, "Type INTEGER %s\n", $1);
		}
	| FLOAT 
		{
			XdbDebug(__FILE__, NULL, "Type FLOAT %s\n", $1);
		}
	| STRING 
		{
			XdbDebug(__FILE__, NULL, "Type STRING %s\n", $1);
		}
	| BOOLEAN 
		{
			XdbDebug(__FILE__, NULL, "Type BOOLEAN %s\n", $1);
		}
	| COMPOUND_STRING 
		{
			XdbDebug(__FILE__, NULL, "Type COMPOUND_STRING %s\n", $1);
		}
	| COLOR 
		{
			XdbDebug(__FILE__, NULL, "Type COLOR %s\n", $1);
		}
	| FONT_TABLE
		{
			XdbDebug(__FILE__, NULL, "Type FONT_TABLE %s\n", $1);
		}
	| ANY
		{
			XdbDebug(__FILE__, NULL, "Type ANY %s\n", $1);
		}
	| WIDGET
		{
			XdbDebug(__FILE__, NULL, "Type WIDGET %s\n", $1);
		}
	| ID
		{
			XdbDebug(__FILE__, NULL, "Type %s\n", $1);
		}
;

list:
	  ID ':' ARGUMENTS '{' list_arg '}' ';' list
		{
			XdbDebug(__FILE__, NULL, "List ARGUMENTS\n");
			$$ = NULL;
		}
	| ID ':' CONTROLS '{' control_list_arg '}' ';' list
		{
			XdbDebug(__FILE__, NULL, "List CONTROLS\n");
			$$ = NULL;
		}
	| ID ':' '{' list_arg '}' ';' list
		{
			XdbDebug(__FILE__, NULL, "List\n");
			$$ = NULL;
		}
	| ID ':' CALLBACKS '{' callback_list_arg '}' ';' list
		{
			XdbDebug(__FILE__, NULL, "List CALLBACKS\n");
			$$ = NULL;
		}
	| ID ':' PROCEDURES '{' procedure_list_arg '}' ';' list
		{
			XdbDebug(__FILE__, NULL, "List PROCEDURES\n");
			$$ = NULL;
		}
	|
		{
			XdbDebug(__FILE__, NULL, "List (empty)\n");
			$$ = NULL;
		}

list_arg:
	  ID '=' expression ';' list_arg
		{
			XdbDebug(__FILE__, NULL, "List element\n");
		}
	|
		{
			XdbDebug(__FILE__, NULL, "List element (empty)\n");
		}
;

callback_list_arg:
	  ID '=' PROCEDURE ID '(' value_expression ')' ';' callback_list_arg
		{
			XdbDebug(__FILE__, NULL, "CallbackList %s = Procedure %s(...)\n", $1, $4);
		}
	| ID '=' PROCEDURE ID '(' ')' ';' callback_list_arg
		{
			XdbDebug(__FILE__, NULL, "CallbackList %s = Procedure %s()\n", $1, $4);
		}
	| ID '=' PROCEDURE ID ';' callback_list_arg
		{
			XdbDebug(__FILE__, NULL, "CallbackList %s = Procedure %s\n", $1, $4);
		}
	|
		{
			XdbDebug(__FILE__, NULL, "CallbackList (empty)\n");
		}
;

control_list_arg:
	MANAGED object_definition_list control_list_arg
		{
			/* FIX ME - Is this a real case or did I invent this ? */
			/* It doesn't seem reasonable that the MANAGED keyword would
			 * apply to more than one object */
			XdbDebug(__FILE__, NULL, "ControlListArg MANAGED\n");
			ObjectManage($2);
			$$ = TreeAppend($2, $3);
		}
	| UNMANAGED object_definition_list control_list_arg
		{
			/* FIX ME - Is this a real case or did I invent this ? */
			/* It doesn't seem reasonable that the UNMANAGED keyword would
			 * apply to more than one object */
			XdbDebug(__FILE__, NULL, "ControlListArg UNMANAGED\n");
			ObjectUnmanage($2);
			$$ = TreeAppend($2, $3);
		}
	| ID ':' MANAGED object_definition_list control_list_arg
		{
			/* FIX ME - Is this a real case or did I invent this ? */
			/* It doesn't seem reasonable that the MANAGED keyword would
			 * apply to more than one object */
			XdbDebug(__FILE__, NULL, "ControlListArg %s : MANAGED\n", $1);
		}
	| ID ':' UNMANAGED object_definition_list control_list_arg
		{
			/* FIX ME - Is this a real case or did I invent this ? */
			/* It doesn't seem reasonable that the UNMANAGED keyword would
			 * apply to more than one object */
			XdbDebug(__FILE__, NULL, "ControlListArg %s : UNMANAGED\n", $1);
		}
	| ID ':' object_definition_list control_list_arg
		{
			XdbDebug(__FILE__, NULL, "ControlListArg %s\n", $1);
		}
	| CONTROLS ID ';' control_list_arg
		{
			XdbDebug(__FILE__, NULL, "ControlListArg - refer to other list %s\n", $2);
			$$ = ControlListAppend($4, ControlListFromList($2));
		}
	| CONTROLS ID ';' 
;

object_definition_list:
	object
;

procedure_list_arg:
	  ID '(' value_expression ')' ';' procedure_list_arg
		{
			XdbDebug(__FILE__, NULL, "Procedures %s (value_expression)\n", $1);
		}
	| ID '(' ')' ';' procedure_list_arg
		{
			XdbDebug(__FILE__, NULL, "Procedures %s ( )\n", $1);
		}
	| ID ';' procedure_list_arg
		{
			XdbDebug(__FILE__, NULL, "Procedures %s\n", $1);
		}
	|
		{
			XdbDebug(__FILE__, NULL, "Procedures (empty)\n");
		}
;

object:
	  ID ':' IMPORTED ID ';'
		{
			XdbDebug(__FILE__, NULL, "OBJECT %s IMPORTED %s\n", $1, $4);
			$$ = ObjectImported($1, $4);
		}
	| ID ':' ID '{' features '}' ';'
		{
			XdbDebug(__FILE__, NULL, "OBJECT %s : %s\n", $1, $3);
			$$ = ObjectNormal($1, $3, UIL_OT_UNKNOWN, $5);
		}
	| ID ':' ID WIDGET '{' features '}' ';'
		{
			XdbDebug(__FILE__, NULL, "OBJECT %s %s WIDGET\n", $1, $3);
			$$ = ObjectNormal($1, $3, UIL_OT_WIDGET, $6);
		}
	| ID ':' EXPORTED ID '{' features '}' ';'
		{
			XdbDebug(__FILE__, NULL, "OBJECT %s EXPORTED %s\n", $1, $4);
			$$ = ObjectExported($1, $4, UIL_OT_WIDGET, $6);
		}
	| ID ':' USER_DEFINED PROCEDURE ID '{' features '}' ';'
		{
			XdbDebug(__FILE__, NULL, "OBJECT %s USER_DEFINED PROCEDURE %s\n", $1, $5);
			$$ = ObjectUserDefined($1, $5, UIL_OT_UNKNOWN, $7);
		}
	| ID ID ';'
		{
			XdbDebug(__FILE__, NULL, "OBJECT %s %s\n", $1, $2);
			$$ = ObjectNormal($1, $2, UIL_OT_UNKNOWN, NULL);
		}
	| ID '{' features '}' ';'
		{
			XdbDebug(__FILE__, NULL, "OBJECT %s : %s\n", $1, $3);
			$$ = ObjectNormal($1, NULL, UIL_OT_UNKNOWN, $3);
		}
	| ID WIDGET '{' features '}' ';'
		{
			XdbDebug(__FILE__, NULL, "OBJECT %s WIDGET\n", $1);
			$$ = ObjectNormal($1, NULL, UIL_OT_WIDGET, $4);
		}
	| USER_DEFINED PROCEDURE ID '{' features '}' ';'
		{
			XdbDebug(__FILE__, NULL, "OBJECT %s USER_DEFINED PROCEDURE %s\n", $1, $5);
		}
;

features:
	  features controls
		{
			XdbDebug(__FILE__, NULL, "Features CONTROLS\n");
			$$ = FeaturesControls($1, $2);
		}
	|  features arguments
		{
			XdbDebug(__FILE__, NULL, "Features ARGUMENTS\n");
			$$ = FeaturesArguments($1, $2);
		}
	|  features callbacks 
		{
			XdbDebug(__FILE__, NULL, "Features CALLBACKS\n");
			$$ = FeaturesCallbacks($1, $2);
		}
	| 
		{
			XdbDebug(__FILE__, NULL, "Features (empty)\n");
			$$ = NULL;
		}
;

controls:
	  CONTROLS '{' control_list '}' ';'
		{
			$$ = $3;
		}
	| CONTROLS ID ';'
		{
			$$ = $2;
		}
;

control_list:
	  control_list ID ID ';'  
		{
			YYSTYPE	o;
			o = ObjectNormal($3, $2, UIL_OT_UNKNOWN, NULL);
			TreeInsert((tree *)&UilGlobal->Objects, o);
			$$ = TreeAppend($1, ControlListObject($3, o));
		}
	| control_list ID '{' features '}' ';'
		{
			YYSTYPE	o;
			/* untyped */
			o = ObjectNormal($2, NULL, UIL_OT_UNKNOWN, $4);
			TreeInsert((tree *)&UilGlobal->Objects, o);
			$$ = TreeAppend($1, ControlListObject($2, o));
		}
	| control_list ID UNMANAGED '{' features '}' ';'
		{
			YYSTYPE	o;
			o = ObjectNormal($2, NULL, UIL_OT_UNKNOWN, $5);
			ObjectUnmanage(o);
			TreeInsert((tree *)&UilGlobal->Objects, o);
			$$ = TreeAppend($1, ControlListObject($2, o));
		}
	| control_list MANAGED ID '{' features '}' ';' 
		{
			YYSTYPE	o;
			o = ObjectNormal($3, NULL, UIL_OT_UNKNOWN, $5);
			ObjectManage(o);
			TreeInsert((tree *)&UilGlobal->Objects, o);
			$$ = TreeAppend($1, ControlListObject($3, o));
		}
	| control_list UNMANAGED ID '{' features '}' ';' 
		{
			YYSTYPE	o;
			o = ObjectNormal($3, NULL, UIL_OT_UNKNOWN, $5);
			ObjectUnmanage(o);
			TreeInsert((tree *)&UilGlobal->Objects, o);
			$$ = TreeAppend($1, ControlListObject($3, o));
		}
	| control_list ID ':' ID '{' features '}' ';' 
		{
			YYSTYPE	o;
			o = ObjectNormal($2, $4, UIL_OT_UNKNOWN, $6);
			TreeInsert((tree *)&UilGlobal->Objects, o);
			$$ = TreeAppend($1, ControlListObject($2, o));
		}
	| control_list ID ':' ID WIDGET '{' features '}' ';' 
		{
			YYSTYPE	o;
			o = ObjectNormal($2, $4, UIL_OT_WIDGET, $7);
			TreeInsert((tree *)&UilGlobal->Objects, o);
			$$ = TreeAppend($1, ControlListObject($2, o));
		}
	| control_list ID ':' ID GADGET '{' features '}' ';' 
		{
			YYSTYPE	o;
			o = ObjectNormal($2, $4, UIL_OT_GADGET, $7);
			TreeInsert((tree *)&UilGlobal->Objects, o);
			$$ = TreeAppend($1, ControlListObject($2, o));
		}
	| control_list MANAGED ID ID ';' 
		{
			YYSTYPE	o;
			o = ObjectNormal($4, $3, UIL_OT_UNKNOWN, NULL);
			ObjectManage(o);
			TreeInsert((tree *)&UilGlobal->Objects, o);
			$$ = TreeAppend($1, ControlListObject($4, o));
		}
	| control_list UNMANAGED ID ID ';' 
		{
			YYSTYPE	o;
			o = ObjectNormal($4, $3, UIL_OT_UNKNOWN, NULL);
			ObjectUnmanage(o);
			TreeInsert((tree *)&UilGlobal->Objects, o);
			$$ = TreeAppend($1, ControlListObject($4, o));
		}
	| control_list USER_DEFINED ID ';' 
		{
			YYSTYPE	o;
			o = ObjectUserDefined($3, NULL, UIL_OT_UNKNOWN, NULL);
			TreeInsert((tree *)&UilGlobal->Objects, o);
			$$ = TreeAppend($1, ControlListObject($3, o));
		}
	| 
		{
			$$ = NULL;
		}
;

arguments:
	  ARGUMENTS '{' argument_list '}' ';'
			{
				XdbDebug(__FILE__, NULL, "Arguments (list)\n");
				$$ = $3;
			}
	| ARGUMENTS ID ';'
			{
				XdbDebug(__FILE__, NULL, "Arguments %s\n", $2);
				$$ = ArgumentsFromList($2);
			}
;

argument_list:
	ID '=' expression ';' argument_list
		{
			XdbDebug(__FILE__, NULL, "ArgumentList ID(%s) = expr\n", $1);
			$$ = ArgumentConcat($5, ArgumentIdIsExpression($1, $3));
#if 0
			XdbDebug(__FILE__, NULL, "ArgumentList ID(%s) = expr\n", $1);
			$$ = ArgumentIdIsExpression($2, $4);
#endif
		}
	| ID '=' ID ID ';' argument_list /* another widget */
		{
			XdbDebug(__FILE__, NULL, "ArgumentList ID = ID ID\n");
			$$ = ArgumentConcat($6, ArgumentIdIsIdId($1, $3, $4));
#if 0
			XdbDebug(__FILE__, NULL, "ArgumentList ID = ID ID\n");
			$$ = ArgumentIdIsIdId($1, $3, $4);
#endif
		}
	| ARGUMENTS ID ';' argument_list
		{
			XdbDebug(__FILE__, NULL, "ArgumentList ARGUMENTS ID\n");
			$$ = ArgumentConcat($4, ArgumentsFromList($2));
#if 0
			XdbDebug(__FILE__, NULL, "ArgumentList ARGUMENTS ID\n");
			$$ = ArgumentsFromList($2);
#endif
		}
	|
		{
			$$ = NULL;
		}
;

compound_string:
	  compound_string '&' COMPOUND_STRING '(' STRING ',' SEPARATE '=' BOOL ')'
		{
		}
	| compound_string '&' COMPOUND_STRING '(' STRING ')'
		{
		}
	| compound_string '&' ID
		{
			XdbDebug(__FILE__, NULL, "NOOP line %d\n", __LINE__);
		}
	| compound_string '&' STRING
		{
		}
	| COMPOUND_STRING '(' STRING ',' SEPARATE '=' BOOL ')'
		{
		}
	| COMPOUND_STRING '(' STRING ')'
		{	
		}
;

callbacks:
	  CALLBACKS '{' callback_list '}' ';'
		{
			$$ = $3;
		}
	| CALLBACKS ID  ';'
		{
			$$ = $3;
		}
;

callback_list:
	  callback_list ID '=' PROCEDURE ID '(' proc_argument_list ')' ';'
		{
		}
	| callback_list ID '=' PROCEDURES '{' procedure_list '}' ';'
		{
			XdbDebug(__FILE__, NULL, "NOOP line %d\n", __LINE__);
		}
	| 
		{
			$$ = NULL;
		}
;

proc_argument_list: proc_argument_list ',' INTEGER
		{
			XdbDebug(__FILE__, NULL, "NOOP line %d\n", __LINE__);
		}
	| proc_argument_list ',' ID
		{
			XdbDebug(__FILE__, NULL, "NOOP line %d\n", __LINE__);
		}
	| proc_argument_list ',' STRING
		{
			XdbDebug(__FILE__, NULL, "NOOP line %d\n", __LINE__);
		}
	| INTEGER
		{
			XdbDebug(__FILE__, NULL, "NOOP line %d\n", __LINE__);
		}
	| ID
		{
			XdbDebug(__FILE__, NULL, "NOOP line %d\n", __LINE__);
		}
	| STRING
		{
			XdbDebug(__FILE__, NULL, "NOOP line %d\n", __LINE__);
		}
	| compound_string
		{
			XdbDebug(__FILE__, NULL, "NOOP line %d\n", __LINE__);
		}
	|	
		{
			$$ = NULL;
		}
;


%%


extern optind;
extern char *optarg;

void yyerror(char *s)
{
  fprintf(stderr,"%s:%d: %s\n",FileName,LineNumber,s);
  fprintf(stderr," current token \'%s\'\n",yytext);
}

yywrap()
{
}

void yyExit(int line, char *file, char *fmt, ...)
{
  va_list ap;

  yyerror("");
  va_start (ap, fmt);
  fprintf(stderr,"%s:%d:",file,line);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  exit(1);
}
