/**
 *
 * $Id: MrmRead.c,v 1.2 1996/08/14 05:45:12 u27113 Exp $
 * 
 * Copyright (C) 1996 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static char rcsid[] = "$Id: MrmRead.c,v 1.2 1996/08/14 05:45:12 u27113 Exp $";

#include <X11/Xos.h>
#include <XmI/DebugUtil.h>

#include "UilMrm.h"

#include <stdarg.h>
#include "uil.h"
#include "lookup.h"
#include "misc.h"

#include <sys/stat.h>
#include <sys/file.h>

/*
 * Implementation idiocies
 *	This really means FIX ME
 */
#define	MAX_LINE_LENGTH	512		/* Longer than any UID line */

/*
 * Global data structures
 */
uil_global_t	**UilModules = NULL;
int		nModules = 0;
static char	line[MAX_LINE_LENGTH];


/* FIX ME this isn't nearly right yet */
/* ANSI C Concatenation */
static char *uid_default_path =
"%U%S:"			"%U:"
"%a/%L/uid/%N/%U%S:"	"%a/%L/uid/%N/%U:"
"%a/%l/uid/%N/%U%S:"	"%a/%l/uid/%N/%U:"
"%a/uid/%N/%U%S:"	"%a/uid/%N/%U:"
"%a/%L/uid/%U%S:"	"%a/%L/uid/%U:"
"%a/%l/uid/%U%S:"	"%a/%l/uid/%U:"
"%a/uid/%U%S:"		"%a/uid/%U:"
"%H/uid/%U%S:"		"%H/uid/%U:"
"%H/%U%S:"		"%H/%U:"
"%D/%L/uid/%N/%U%S:"	"%D/%L/uid/%N/%U:"
"%D/%l/uid/%N/%U%S:"	"%D/%l/uid/%N/%U:"
"%D/uid/%N/%U%S:"	"%D/uid/%N/%U:"
"%D/%L/uid/%U%S:"	"%D/%L/uid/%U:"
"%D/%l/uid/%U%S:"	"%D/%l/uid/%U:"
;

static SubstitutionRec subst[] = {
/* $HOME */		{ 'H', NULL },
/* file name */		{ 'U', NULL },
/* app class name */	{ 'N', NULL },
/* $XAPPLRESDIR */	{ 'a', NULL },
};

/*
 * Current <<<XXX>>>
 *	For use while reading the uid.
 */
uil_global_t	*UilGlobal = NULL;
uil_objects_t	*UilObject = NULL;

/*
 * Initialisation
 *	According to the spec, you can call this after MrmOpenHierarchy,
 *	but you *must* call it before creating widgets.
 *	It may have to do with type converters. Again, see the spec.
 */
void
MrmInitialize ()
{
	UilModules = NULL;			/* Not necessary */
}

/*
 * Read UID statements
 *	These functions are all declared static (i.e. local).
 *
 * Return values :
 *	-1 for failure
 *	0 or some positive value (ID) for success
 */
static int
UidHeaderRead(char *p)
{
	return 0;
}

static int
ModuleRead(char *p)
{
	uil_global_t	*m;
	char		*p, *q;
	int		i;

/* Get a module structure */
	m = (uil_global_t *) XtMalloc(sizeof(uil_global_t));
	UilGlobal = m;

/* Expand the array storing all module pointers */
	nModules++;
	UilModules = (uil_global_t **) XtRealloc((XtPointer)UilModules,
			sizeof(uil_global_t *) * nModules);
	UilModules[nModules - 1] = m;

/* Read the line, field 1 (module name) */
	p = line + 8;			/* Go past <MODULE> */
	for (q=p; *q != ','; q++) ;
	m->module_name = XtMalloc(q - p + 1);
	for (i=0; p+i < q; i++)
		m->module_name[i] = p[i];
	m->module_name[i] = '\0';

/* Field 2 - version string */
	p = q+1;			/* First character of version */
	for (q=p; *q != ','; q++) ;
	m->uil_version = XtMalloc(q - p + 1);
	for (i=0; p+i < q; i++)
		m->uil_version[i] = p[i];
	m->uil_version[i] = '\0';

/* Field 3 - Case sensitive */
	p = ++q;
	if (*p == 's')
		m->case_sensitive = True;
	else
		m->case_sensitive = False;

/* Field 4 - Character set */
	p = q + 2;
	for (q=p; *q != '\n'; q++) ;
	m->character_set = XtMalloc(q - p + 1);
	for (i=0; p+i < q; i++)
		m->character_set[i] = p[i];
	m->character_set[i] = '\0';

	XdbDebug(__FILE__, NULL, "Module %s, Uil Version %s, Case %s, Charset %s\n",
		m->module_name,
		m->uil_version,
		m->case_sensitive ? "Sensitive" : "Insensitive",
		m->character_set);

	return 0;
}

/*
 * Read a string starting at p, until but not including a comma,
 *	return a pointer to the comma.
 */
static char *
_Read_String(char *p, char **r, char comma)
{
	int	i;
	char	*q;

	for (q=p; *q && *q != comma && *q != '\n'; q++) ;
	*r = XtMalloc(q - p + 1);
	for (i=0; p+i < q; i++)
		(*r)[i] = p[i];
	(*r)[i] = '\0';
	return q;
}

static char *
_ReadString(char *p, char **r)
{
	return _Read_String(p, r, ',');
}

#define	CheckComma(p)		\
	if (*p != ',')		\
		return -1;

#define	CheckEol(p)		\
	if (*p != '\n')		\
		return -1;

static int
ArgumentRead(char *p)
{
	char	*b;

	p = _ReadString(p, &b);
	XdbDebug(__FILE__, NULL, "Argument %s\n", b);
	XtFree(b);
	return 0;
}

static int
ControlListRead(char *p)
{
	XdbDebug(__FILE__, NULL, "ControlList\n");
	return 0;
}

static int
ObjectRead(char *p)
{
	char	*n, *wt, w, m;
	tree	t;

	p = _ReadString(p, &n);		/* name */
	CheckComma(p);
	p = _ReadString(p+1, &wt);	/* widget type */
	CheckComma(p);
	p++;
	w = *p;
	p++;
	CheckComma(p);
	p++;
	m = *p;
	p++;
	CheckEol(p);

	t = ObjectNormal(n, wt, (w == 'w') ? UIL_OT_WIDGET : UIL_OT_GADGET, NULL);
	if (m == 'm')
		ObjectManage(t);
	else
		ObjectUnmanage(t);

	XdbDebug(__FILE__, NULL, "Object %s %s %c %c\n", n, wt, w, m);
	return 0;
}

static struct {
	char *code;
	int (*f)(char *);
} _decode [] = {
 { "%LessTif%",	UidHeaderRead },
 { "<MODULE>",	ModuleRead },
 { "<WID>",	ObjectRead },
 { "<ARG>",	ArgumentRead },
 { "<CTRL>",	ControlListRead },
};

static Boolean
pred(String p)
{
	struct stat status;

	XdbDebug(__FILE__, NULL, "Pred(%s)\n", p);

	return (access(p, R_OK) == 0 &&
		stat(p, &status) == 0 &&
		S_ISDIR(status.st_mode) == 0);
}

/*
 * Read a bunch of UID files
 */
Cardinal 
MrmOpenHierarchyPerDisplay(Display *display,
			   MrmCount num_files,
			   String name_list[],
			   MrmOsOpenParamPtr os_params_list[],
			   MrmHierarchy *hierarchy)
{
	int	i, j;
	FILE	*f;
	char	*r;

	XdbDebug(__FILE__, NULL, "MrmOpenHierarchyPerDisplay\n");

/* Handle substitution record */
	subst[0].match = 'H';
	subst[0].substitution = getenv("HOME");
	subst[2].match = 'N';
	subst[2].substitution = "??";
	subst[3].match = 'a';
	subst[3].substitution = getenv("XAPPRESDIR");

/* Open files */
	for (i=0; i<num_files; i++) {
	    subst[1].match = 'U';
	    subst[1].substitution = name_list[i];

	    if (display) {
		char *path = getenv("UIDPATH");
		char *p = XtResolvePathname(display,
	/* type %T */		"uid",
	/* name %N */		name_list[i],
	/* suffix %S */		".uid",
	/* path */		path ? path : uid_default_path,
	/* substitutions */	subst, 4,
	/* predicate */		pred);
		if (!p)
		    p = XtResolvePathname(display,
	/* type %T */		"uid",
	/* name %N */		name_list[i],
	/* suffix %S */		NULL,
	/* path */		path ? path : uid_default_path,
	/* substitutions */	subst, 4,
	/* predicate */		pred);

		XdbDebug(__FILE__, NULL, "MrmOpenHierarchyPerDisplay: XtResolvePathname found %s\n", p);

		if (p) {
		    f = fopen(p, "r");
		    XtFree(p);
		} else
		    f = fopen(name_list[i], "r");
	    } else
		f = fopen(name_list[i], "r");
	    if (f == NULL) {
		return MrmNOT_FOUND;
	    }

	/* Read a line */
	    while ((r = fgets(line, MAX_LINE_LENGTH, f)) != NULL) {
		/* Decode the line */
		/* FIX ME - this should be more efficient.
		   The string codes in UID should be numbers */
		for (j=0; j<XtNumber(_decode); j++) {
		    int l = strlen(_decode[j].code);
		    if (strncmp(line, _decode[j].code, l) == 0) {
			int	id = _decode[j].f(line + l);
			if (id < 0)
				return MrmFAILURE;
		    }
		}
	    }
	}

	return MrmSUCCESS;
}

/* This thing is obsolete */
Cardinal
MrmOpenHierarchy(MrmCount num_files,
		String *name_list,
		MrmOsOpenParamPtr *os_ext_list,
		MrmHierarchy *hierarchy_id_return)
{
	return MrmOpenHierarchyPerDisplay(NULL, num_files, name_list,
			os_ext_list, hierarchy_id_return);
}

Cardinal
MrmCloseHierarchy(MrmHierarchy h)
{
	/* Free all the data structures */
	/* FIX ME */
	return MrmSUCCESS;
}
