/**
 *
 * $Id: Object.c,v 1.1.1.1 1996/08/06 05:51:34 toshok Exp $
 *
 * Copyright (C) 1995, 1996 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static char *rcsid = "$Id: Object.c,v 1.1.1.1 1996/08/06 05:51:34 toshok Exp $";

#include <stdio.h>
#include "UilMrm.h"
/* #include "Uil.h" */
#include "Tree.h"

#include "Arguments.h"
#include "Control.h"

/*
 * TreeInsert commented out everywhere because they were a way to get the
 * development going. In real life they need to be called from the layer
 * above this one (i.e. from the yacc.y file).
 */
YYSTYPE
ObjectImported(char *a, char *b)
{
	uil_objects_t	*s = (uil_objects_t *)XtMalloc(sizeof(uil_objects_t));

	s->type = s_objects;
	s->name = XtNewString(a);

	return (uil_objects_t *)s;
}

/*
 * ObjectNormal
 *	a : name of the object
 *	b : its type (e.g. XmPushButton)
 *	ot : UIL_OT_UNKNOWN, UIL_OT_WIDGET, or UIL_OT_GADGET
 *	features : features structure
 */
YYSTYPE
ObjectNormal(char *a, char *b, int ot, YYSTYPE features)
{
	uil_objects_t	*s = (uil_objects_t *)XtMalloc(sizeof(uil_objects_t)),
			*f = (uil_objects_t *)features;

	s->name = XtNewString(a);
	s->type = s_objects;
	s->widget_class_name = b;

	switch (ot) {
	case UIL_OT_GADGET:
		s->is_widget = False;
		break;
	case UIL_OT_WIDGET:
	case UIL_OT_UNKNOWN:
	default:
		s->is_widget = True;
	}

	if (f) {
		s->arguments = f->arguments;
		s->controls = f->controls;
		s->callbacks = f->callbacks;
		XtFree((XtPointer)f);
	} else {
		s->arguments = NULL;
		s->controls = NULL;
		s->callbacks = NULL;
	}
	return (uil_objects_t *)s;
}

YYSTYPE
ObjectExported(char *a, char *b, int ot, YYSTYPE features)
{
	return ObjectNormal(a, b, ot, features);
}

YYSTYPE
ObjectUserDefined(char *a, char *b, int ot, YYSTYPE features)
{
	return ObjectNormal(a, b, ot, features);
}

void
ObjectPrint(FILE *o, tree l)
{
	uil_objects_t	*p = (uil_objects_t *)l;

	fprintf(o, "<WID>%s,%s,%c,%c\n",
		p->name,
		p->widget_class_name,
		p->is_widget ? 'w' : 'g',
		p->is_managed ? 'm' : 'u');

	TreePrint(p->arguments, ArgumentPrint);
	TreePrint(p->controls, ControlListPrint);
}

void
ObjectManage(tree t)
{
	uil_objects_t	*p = (uil_objects_t *)t;

	p->is_managed = True;
}

void
ObjectUnmanage(tree t)
{
	uil_objects_t	*p = (uil_objects_t *)t;

	p->is_managed = False;
}
