/**
 *
 * $Id: Tree.c,v 1.2 1996/08/14 05:45:15 u27113 Exp $
 *
 * Copyright (C) 1995, 1996 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static char *rcsid = "$Id: Tree.c,v 1.2 1996/08/14 05:45:15 u27113 Exp $";

/*
 * For now, this is a completely inefficient tree implementation.
 * No balancing whatsoever.
 *
 * "Write a bad but working implementation first, then optimise"
 */

#include <stdio.h>
#include <XmI/DebugUtil.h>
#include "Tree.h"

void
TreeInsert(tree *parent, tree id)
{
	int	r;
	tree	p;

	if (parent == NULL) {
#ifdef	DEBUG
		_XmError(NULL, "uil: TreeInsert(NULL, _)\n");
		/*NOTREACHED*/
#else
		_XmWarning(NULL, "uil: TreeInsert(NULL, _)\n");
		return;
#endif
	}
	if (id == NULL) {
#ifdef DEBUG
		_XmError(NULL, "uil: TreeInsert(_, NULL)\n");
		/*NOTREACHED*/
#else
		_XmWarning(NULL, "uil: TreeInsert(_, NULL)\n");
		return;
#endif
	}

	p = *parent;

	if (p == NULL) {
		XdbDebug(__FILE__, NULL, "TreeInsert(%s)\n", id->name);
		p = *parent = id;
		p->left = p->right = NULL;
		return;
	}

	r = strcmp(p->name, id->name);
	if (r > 0)
		TreeInsert(&(p->left), id);
	else if (r < 0)
		TreeInsert(&(p->right), id);
	else {
		XdbDebug(__FILE__, NULL, "TreeInsert: duplicate %s\n", id->name);
	}
}

void
TreePrint(tree t, void (*f)(FILE *o, tree l))
{
	if (t == NULL)
		return;
	if (t->left)
		TreePrint(t->left, f);
	if (t->name)
		f(stdout, t);
	if (t->right)
		TreePrint(t->right, f);
}

tree
TreeAppend(tree a, tree b)
{
	tree	t;

	if (a == NULL)
		return b;
	else if (b == NULL)
		return a;

	for (t=a; t->right != NULL; t=t->right) ;
	t->right = b;

	return a;
}
