/**
 *
 * $Id: UilMrm.h,v 1.1.1.1 1996/08/06 05:51:34 toshok Exp $
 *
 * Copyright (C) 1995, 1996 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

#ifndef	_UIL_MRM_H_
#define	_UIL_MRM_H_

/*
 * This file contains the definitions of the structure that is built
 * in memory when the uil compiler reads a uil source file.
 *
 * This structure closely resembles but is not equal to the UID file
 * format.
 *
 * The libMrm library reads a UID file and builds the same structure
 * in memory that the uil compiler had.
 */

#include <X11/Intrinsic.h>
#include "Tree.h"

/*
 * Almost sure this is going to change soon ...
 */
#define YYSTYPE char *


/* Real stuff now ... */
typedef	struct uil_arguments_s	uil_arguments_t;
typedef struct uil_callbacks_s	uil_callbacks_t;
typedef struct uil_controls_s	uil_controls_t;
typedef struct uil_expression_s	uil_expression_t;
typedef	struct uil_identifier_s	uil_identifier_t;
typedef struct uil_lists_s	uil_lists_t;
typedef struct uil_procedures_s	uil_procedures_t;
typedef struct uil_objects_s	uil_objects_t;
typedef struct uil_values_s	uil_values_t;

/*
 * Only one of these in the uil compiler,
 * but one per module read in libMrm.
 */
typedef struct uil_global_s {
	char	*uil_version;
	char	*module_name;
	Boolean	case_sensitive;
	char	*character_set;

	uil_arguments_t		*Arguments;
	uil_identifier_t	*Identifiers;
	uil_objects_t		*Objects;
	uil_procedures_t	*Procedures;
	uil_callbacks_t		*Callbacks;
	uil_lists_t		*Lists;
	uil_values_t		*Values;
} uil_global_t;

/*
 * Types defined in UIL(5)
 */
typedef enum uil_type {
	t_any,
	t_argument,
	t_asciz_string_table,
	t_boolean,
	t_color,
	t_color_table,
	t_compound_string,
	t_float,
	t_font,
	t_font_table,
	t_fontset,
	t_icon,
	t_integer,
	t_integer_table,
	t_keysym,
	t_pixmap,
	t_reason,
	t_single_float,
	t_string,
	t_string_table,
	t_translation_table,
	t_wide_character,
	t_widget
} uil_type;

typedef enum structure_type {
	s_unknown,
	s_arguments,
	s_callbacks,
	s_controls,
	s_expression,
	s_identifier,
	s_lists,
	s_objects,
	s_procedures,
	s_values
} what ;

typedef enum uil_expression {
	e_unknown,
	e_literal,
	e_function
} uil_expression;

/*
 * We need some system dependency here.
 * According to the spec, this needs to be a 32-bit type.
 * FIX ME
 */

typedef	long	UilValue;

/*
 * Details of the uil
 */
struct uil_arguments_s {
	char			*name;
	uil_arguments_t		*left, *right;
	what			type;
	XtArgVal		value;
};

struct uil_callbacks_s {
	char			*name;
	uil_callbacks_t		*left, *right;
	what			type;
	uil_procedures_t	*proc;
	uil_values_t		*parameters;
};

struct uil_controls_s {
	char			*name;
	uil_controls_t		*left, *right;
	what			type;
	uil_objects_t		*object;
};

struct uil_expression_s {
	char			*name;
	uil_expression_t	*left, *right;
	what			type;
	uil_expression		expression_type;
	uil_type		value_type;
	UilValue		value;
};

struct uil_identifier_s {
	char			*name;
	uil_identifier_t	*left, *right;
	what			type;
};

struct uil_lists_s {
	char			*name;
	uil_lists_t		*left, *right;
	what			type;
};

struct uil_objects_s {
	char			*name;
	uil_objects_t		*left, *right;
	what			type;
	Boolean			is_widget;
	Boolean			is_managed;
	char			*widget_class_name;	/* where should this be translated ? */
	WidgetClass		widget_class;
	uil_arguments_t		*arguments;
	uil_callbacks_t		*callbacks;
	uil_controls_t		*controls;
	uil_lists_t		*lists;		/* ?? */
};

struct uil_procedures_s {
	char			*name;
	uil_procedures_t	*left, *right;
	what			type;
};

struct uil_values_s {
	char			*name;		/* may be NULL, for literal */
	uil_values_t		*left, *right;
	what			type;
	uil_type		uil_type;
	XtArgVal		value;
};

union uil_all {
	tree			tree;
	uil_arguments_t		arguments;
	uil_callbacks_t		callbacks;
	uil_controls_t		controls;
	uil_identifier_t	identifier;
	uil_lists_t		lists;
	uil_objects_t		objects;
	uil_procedures_t	procedures;
	uil_values_t		values;
};

#define	UIL_OT_UNKNOWN	0
#define	UIL_OT_WIDGET	1
#define	UIL_OT_GADGET	2

#define	UIL_VALUE_EXPORTED	1
#define	UIL_VALUE_PRIVATE	2

#endif
