/**
 *
 * $Id: CascadeBG.c,v 1.3 1996/08/12 02:13:53 miers Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static char rcsid[] = "$Id: CascadeBG.c,v 1.3 1996/08/12 02:13:53 miers Exp $";

#include <LTconfig.h>
#include <XmI/XmI.h>
#include <XmI/DebugUtil.h>

#include <Xm/XmP.h>
#include <Xm/CacheP.h>
#include <Xm/CascadeB.h> /* For XmCascadeButtonHighlight */
#include <Xm/CascadeBP.h>
#include <Xm/CascadeBGP.h>
#include <Xm/ManagerP.h>
#include <Xm/MenuShell.h>
#include <Xm/RowColumnP.h>
#include <Xm/ScreenP.h>
#include <Xm/MenuUtilP.h>
#include <X11/Xfuncs.h>

/* Forward Declarations */

static void class_initialize();
static void class_part_initialize(WidgetClass class);
static void initialize(Widget request, Widget new, ArgList args, Cardinal *num_args);
static void destroy(Widget w);
static void expose(Widget w, XEvent *event, Region region);
static Boolean set_values(Widget current, Widget request, Widget new, ArgList args, Cardinal *num_args);
static void get_values_hook(Widget w, ArgList args, Cardinal *num_args);

static void input_dispatch(Widget gadget, XEvent *event, Mask event_mask);

static void secondary_object_create(Widget request, Widget new, ArgList args, Cardinal *num_args);
static void initialize_prehook(Widget request, Widget new, ArgList args, Cardinal *num_args);
static void initialize_posthook(Widget request, Widget new, ArgList args, Cardinal *num_args);
static Boolean set_values_prehook(Widget old, Widget request, Widget new, ArgList args, Cardinal *num_args);
static Boolean set_values_posthook(Widget old, Widget request, Widget new, ArgList args, Cardinal *num_args);
static void get_values_prehook(Widget new, ArgList args, Cardinal *num_args);
static void get_values_posthook(Widget new, ArgList args, Cardinal *num_args);
static Cardinal get_sec_res_data(WidgetClass wc, XmSecondaryResourceData **data);

/*
 * This used to be a private function, now we use it for both CascadeButtonWidget
 * and CascadeButtonGadget.
 * Implemented in CascadeB.c
 */
extern void _XmCBMenuBarSelect(Widget w, XEvent *event, String *params, Cardinal *num_params);
extern void _XmCBMenuBarDoSelect(Widget w, XEvent *event, String *params, Cardinal *num_params);
extern int _XmCascadeBCacheCompare(XtPointer A, XtPointer B);


/*
 * cache resources
 */
#define Offset(field) XtOffsetOf(XmCascadeButtonGCacheObjRec, cascade_button_cache.field)
static XtResource cache_resources[] = {
    {
	XmNcascadePixmap, XmCPixmap, XmRGadgetPixmap,
	sizeof(Pixmap), Offset(cascade_pixmap),
	XmRImmediate, (XtPointer)XmUNSPECIFIED_PIXMAP
    },
    {
	XmNmappingDelay, XmCMappingDelay, XmRInt,
	sizeof(int), Offset(map_delay),
	XmRImmediate, (XtPointer)180
    }
};

XmCascadeButtonGCacheObjClassRec xmCascadeButtonGCacheObjClassRec = {
    /* Object class part */
    {
	/* superclass            */ (WidgetClass) &xmLabelGCacheObjClassRec,
        /* class_name            */ "XmCascadeButtonGCacheObjClass",
	/* widget_size           */ sizeof(XmCascadeButtonGCacheObjRec),
	/* class_initialize      */ NULL,
	/* class_part_initialize */ NULL,
	/* class_inited          */ FALSE,
	/* initialize            */ NULL,
	/* initialize_hook       */ NULL,
	/* realize               */ NULL,
	/* actions               */ NULL,
	/* num_actions           */ 0,
	/* resources             */ cache_resources,
	/* num_resources         */ XtNumber(cache_resources),
	/* xrm_class             */ NULLQUARK,
	/* compress_motion       */ 0,
	/* compress_exposure     */ 0,
	/* compress_enterleave   */ 0,
	/* visible_interest      */ 0,
	/* destroy               */ NULL,
	/* resize                */ NULL,
	/* expose                */ NULL,
	/* set_values            */ NULL,
	/* set_values_hook       */ NULL,
	/* set_values_almost     */ NULL,
	/* get_values_hook       */ NULL,
	/* accept_focus          */ NULL,
	/* version               */ XtVersion,
	/* callback offsets      */ NULL,
	/* tm_table              */ NULL,
	/* query_geometry        */ NULL,
        /* display_accelerator   */ NULL,
	/* extension             */ NULL
    },
    /* XmExtObject part */
    {
        /* syn_resources      */ NULL,
        /* num_syn_resources  */ 0,
        /* extension          */ NULL
    },
    /* LabelGCacheObj part */
    {
	/* foo                */ 0
    },
    /* CascadeButtonGCacheObj part */
    {
	/* foo                */ 0
    }
};

#undef Offset
#define Offset(field) XtOffsetOf(XmCascadeButtonGadgetRec, cascade_button.field)

/* Resources for the cascadebutton class */
static XtResource resources[] = {
    {
	XmNactivateCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(activate_callback),
	XmRCallback, (XtPointer)NULL
    },
    {
	XmNcascadingCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(cascade_callback),
	XmRCallback, (XtPointer)NULL
    },
    {
	XmNsubMenuId, XmCMenuWidget, XmRMenuWidget,
	sizeof(Widget), Offset(submenu),
	XmRMenuWidget, (XtPointer)NULL
    },
    /* resources we override from XmGadget */
    {
	XmNshadowThickness, XmCShadowThickness, XmRHorizontalDimension,
	sizeof(Dimension), XtOffsetOf(XmCascadeButtonGadgetRec, gadget.shadow_thickness),
	XmRImmediate, (XtPointer)2
    },
    {
	XmNtraversalOn, XmCTraversalOn, XmRBoolean,
	sizeof(Boolean), XtOffsetOf(XmCascadeButtonGadgetRec, gadget.traversal_on),
	XmRImmediate, (XtPointer)True
    },
    {
	XmNhighlightThickness, XmCHighlightThickness, XmRHorizontalDimension,
	sizeof(Dimension), XtOffsetOf(XmCascadeButtonGadgetRec, gadget.highlight_thickness),
	XmRImmediate, (XtPointer)2
    },
};

static XmBaseClassExtRec _XmCascadeBGRectClassExtRec = {
    /* next_extension            */ NULL,
    /* record_type               */ NULLQUARK,                             
    /* version                   */ XmBaseClassExtVersion,
    /* size                      */ sizeof(XmBaseClassExtRec),
    /* initialize_prehook        */ initialize_prehook,
    /* set_values_prehook        */ set_values_prehook,
    /* initialize_posthook       */ initialize_posthook,
    /* set_values_posthook       */ set_values_posthook,
    /* secondary_object_class    */ (WidgetClass)&xmCascadeButtonGCacheObjClassRec,
    /* secondary_object_create   */ secondary_object_create,
    /* get_secondary_resources   */ get_sec_res_data,
    /* fast_subclass             */ { 0 },
    /* get_values_prehook        */ get_values_prehook,
    /* get_values_posthook       */ get_values_posthook,
    /* class_part_init_prehook   */ NULL,
    /* class_part_init_posthook  */ NULL,
    /* ext_resources             */ NULL,
    /* compiled_ext_resources    */ NULL,
    /* num_ext_resources         */ 0,
    /* use_sub_resources         */ FALSE,
    /* widget_navigable          */ XmInheritWidgetNavigable,
    /* focus_change              */ XmInheritFocusChange,
    /* wrapper_data              */ NULL
};

static XmCacheClassPart cache_part = {
    /* cache head part */
    {
        /* next         */ NULL,
        /* prev         */ NULL,
        /* ref_count    */ 0
    },
    _XmCacheCopy,
    _XmCacheDelete,
    _XmCascadeBCacheCompare
};

static XmGadgetClassExtRec _XmCascadeBGGadgetClassExtRec = {
    /* next_extension            */ NULL,
    /* record_type               */ NULLQUARK,
    /* version                   */ XmGadgetClassExtVersion,
    /* size                      */ sizeof(XmGadgetClassExtRec),
    /* widget_baseline_proc      */ XmInheritBaselineProc,
    /* display_rect_proc         */ XmInheritDisplayRectProc,
};

static void DelayedArm(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void CheckDisarm(Widget w, XEvent *event, String *params, Cardinal *num_params);

XmCascadeButtonGadgetClassRec xmCascadeButtonGadgetClassRec = {
    /* RectObj class part */
    {
	/* superclass            */ (WidgetClass) &xmLabelGadgetClassRec,
	/* class_name            */ "XmCascadeButtonGadget",
	/* widget_size           */ sizeof(XmCascadeButtonGadgetRec),
	/* class_initialize      */ class_initialize,
	/* class_part_initialize */ class_part_initialize,
	/* class_inited          */ FALSE,
	/* initialize            */ initialize,
	/* initialize_hook       */ NULL,
	/* realize               */ NULL,
	/* actions               */ NULL,
	/* num_actions           */ 0,
	/* resources             */ resources,
	/* num_resources         */ XtNumber(resources),
	/* xrm_class             */ NULLQUARK,
	/* compress_motion       */ FALSE,
	/* compress_exposure     */ XtExposeCompressMaximal,
	/* compress_enterleave   */ FALSE,
 	/* visible_interest      */ FALSE,
	/* destroy               */ destroy,
	/* resize                */ XtInheritResize,
	/* expose                */ expose,
	/* set_values            */ set_values,
	/* set_values_hook       */ NULL,
	/* set_values_almost     */ XtInheritSetValuesAlmost,
	/* get_values_hook       */ get_values_hook,
	/* accept_focus          */ NULL,
	/* version               */ XtVersion,
	/* callback offsets      */ NULL,
	/* tm_table              */ NULL,
	/* query_geometry        */ XtInheritQueryGeometry,
	/* display_accelerator   */ NULL,
	/* extension             */ (XtPointer)&_XmCascadeBGRectClassExtRec
    },
    /* XmGadget part */
    {
	/* border_highlight   */ XmInheritBorderHighlight, 
	/* border_unhighlight */ XmInheritBorderUnhighlight, /* fix me */
	/* arm_and_activate   */ NULL, /* fix me */
	/* input_dispatch     */ input_dispatch,
	/* visual_change      */ NULL, /* fix me */
	/* syn_resources      */ NULL,
	/* num_syn_resources  */ 0,
	/* cache_part         */ &cache_part,
	/* extension          */ (XtPointer)&_XmCascadeBGGadgetClassExtRec
    },
    /* XmLabelGadget part */
    {
        /* setOverrideCallback */ NULL, /* fix me */
        /* menuProcs           */ NULL, /* fix me */
	/* extension           */ NULL
    },
    /* XmCascadeButtonGadget part */
    {
	/* extension */ NULL
    },
};

WidgetClass xmCascadeButtonGadgetClass = (WidgetClass)&xmCascadeButtonGadgetClassRec;

/********************************* CACHE PART *******************************/
static void
secondary_object_create(Widget request,
	   Widget new,
	   ArgList args,
	   Cardinal *num_args)
{
    XmBaseClassExt *bce;
    XtPointer nsec, rsec;
    XmWidgetExtData ed;
    int size;

    XdbDebug(__FILE__, new, "CascadeButtonGCacheRec %s being initialized.\n",
	     XtName(new));

    bce = _XmGetBaseClassExtPtr(XtClass(new), XmQmotif);

    size = (*bce)->secondaryObjectClass->core_class.widget_size;
    nsec = _XmExtObjAlloc(size);
    rsec = _XmExtObjAlloc(size);

    ((XmExtRec *)nsec)->object.self = (Widget)nsec;
    ((XmExtRec *)nsec)->object.widget_class = (*bce)->secondaryObjectClass;
    ((XmExtRec *)nsec)->object.parent = XtParent(new);
    ((XmExtRec *)nsec)->object.xrm_name = new->core.xrm_name;
    ((XmExtRec *)nsec)->object.being_destroyed = False;
    ((XmExtRec *)nsec)->object.destroy_callbacks = NULL;
    ((XmExtRec *)nsec)->object.constraints = NULL;

    ExtObj_LogicalParent(nsec) = new;
    ExtObj_ExtensionType(nsec) = XmCACHE_EXTENSION;

    XtGetSubresources(new, nsec, NULL, NULL,
                      (*bce)->secondaryObjectClass->core_class.resources,
                      (*bce)->secondaryObjectClass->core_class.num_resources,
                      args, *num_args);

    ed = (XmWidgetExtData)XtMalloc(sizeof(XmWidgetExtDataRec));
    ed->widget = (Widget)nsec;
    ed->reqWidget = (Widget)rsec;

    bcopy(nsec, rsec, size);
    ((XmExtRec *)rsec)->object.self = (Widget)rsec;

    _XmPushWidgetExtData(new, ed, XmCACHE_EXTENSION);

    LabG_Cache(new) = &(((XmLabelGCacheObject)nsec)->label_cache);
    LabG_Cache(request) = &(((XmLabelGCacheObject)rsec)->label_cache);

    CBG_Cache(new) =
	&(((XmCascadeButtonGCacheObject)nsec)->cascade_button_cache);
    CBG_Cache(request) =
	&(((XmCascadeButtonGCacheObject)rsec)->cascade_button_cache);
}

int
_XmCascadeBCacheCompare(XtPointer A, XtPointer B)
{
    return !bcmp(((XmCascadeButtonGCacheObjPart *)A),
		 ((XmCascadeButtonGCacheObjPart *)B),
		 sizeof(XmCascadeButtonGCacheObjPart));
}

int
_XmArrowPixmapCacheCompare(XtPointer A, XtPointer B)
{
    return 0;
}

void
_XmArrowPixmapCacheDelete(XtPointer data)
{
}

/*
 * Assumes that the existing pixmaps are freed.
 */
void
_XmCreateArrowPixmaps(Widget w)
{
	Pixmap		apm, pm;
	GC		bgc, tgc, fgc;
	Display		*dpy;
	int		depth;
	Dimension	ww, hh, st;
	Screen		*scr;
	Window		win;
	Pixel		tsc, bsc, fc, hc;
	XGCValues	values;
	XtGCMask	mask, dynamic, dontcare;

	tsc = XmParentTopShadowColor(w);	/* top shadow color */
	bsc = XmParentBottomShadowColor(w);	/* bottom shadow color */
	fc = XmParentForeground(w);		/* fill color */
	hc = XmParentHighlightColor(w);		/* highlight color */

	values.foreground = bsc;
	values.background = XmParentBackground(w);
	mask = GCForeground | GCBackground;
	dynamic = GCClipMask | GCClipXOrigin | GCClipYOrigin;
	dontcare = ~(mask | dynamic);
	bgc = XtAllocateGC(w, 0, mask, &values, dynamic, dontcare);

	values.foreground = tsc;
	values.background = XmParentBackground(w);
	mask = GCForeground | GCBackground;
	dynamic = GCClipMask | GCClipXOrigin | GCClipYOrigin;
	dontcare = ~(mask | dynamic);
	tgc = XtAllocateGC(w, 0, mask, &values, dynamic, dontcare);

	values.foreground = values.background = XmParentBackground(w);
	mask = GCForeground | GCBackground;
	dynamic = GCForeground | GCClipMask | GCClipXOrigin | GCClipYOrigin;
	dontcare = ~(mask | dynamic);
	fgc = XtAllocateGC(w, 0, mask, &values, dynamic, dontcare);

	dpy = XtDisplay(w);
	scr = XtScreen(w);
	depth = DefaultDepthOfScreen(scr);

	ww = 2 * XtHeight(w) / 3;
	hh = XtHeight(w) / 2;
	st = 2;

	win = RootWindowOfScreen(scr);

/* Create pixmaps, fill them up with _XmDrawArrow */
	pm = XCreatePixmap(dpy, win, ww, hh, depth);
	apm = XCreatePixmap(dpy, win, ww, hh, depth);

	XdbDebug(__FILE__, w, "_XmCreateArrowPixmaps -> CascadePixmap 0x%X ArmedPixmap 0x%X\n",
		pm, apm);

	XFillRectangle(dpy, pm, fgc, 0, 0, ww, hh);
	XFillRectangle(dpy, apm, fgc, 0, 0, ww, hh);

	_XmDrawArrow(dpy, pm,
		/* GC's */		bgc, tgc, fgc,
		/* x, y, w, h */	0, 0, ww, hh,
		/* shadowthickness */	st,
		/* direction */		XmARROW_RIGHT);

	XSetForeground(dpy, fgc, hc);
/* Armed pixmap */
	_XmDrawArrow(dpy, apm,
		/* GC's */		bgc, tgc, fgc,
		/* x, y, w, h */	0, 0, ww, hh,
		/* shadowthickness */	st,
		/* direction */		XmARROW_RIGHT);

/* Copy them into the widget/gadget */
	if (XmIsGadget(w)) {
		CBG_ArmedPixmap(w) = apm;
		CBG_CascadePixmap(w) = pm;
	} else {
		CB_ArmedPixmap(w) = apm;
		CB_CascadePixmap(w) = pm;
	}
/* Caching ?? */

	XtReleaseGC(w, bgc);
	XtReleaseGC(w, fgc);
	XtReleaseGC(w, tgc);
}

/******************************** GADGET PART *******************************/
static void
class_initialize()
{
    XtResourceList combined, labels;
    int ncom;
    Cardinal nlabels;

    /* don't let the nulls fool you.  look at the header file -- the arg
     * isn't used. */
    ClassCacheHead(CBG_ClassCachePart(NULL)).prev =
	&ClassCacheHead(CBG_ClassCachePart(NULL));
    ClassCacheHead(CBG_ClassCachePart(NULL)).next =
	&ClassCacheHead(CBG_ClassCachePart(NULL));

    _XmCascadeBGRectClassExtRec.record_type = XmQmotif;

    /*
     * Label subclasses (ToggleBG, PushBG, CascadeBG) have a problem.  Since
     * we do all the subpart manipulation in the pre- and post- hooks, and
     * since those hooks aren't chained, we have to either make multiple
     * calls to XtGetSubresources/Xt[Get|Set]Subvalues, or merge the resource
     * lists.  Since I just wrote _XmTransformSubresources, seems like a
     * waste not to use it.
     */
    ncom = XtNumber(cache_resources) +
           xmLabelGCacheObjClassRec.object_class.num_resources;

    _XmTransformSubResources(xmLabelGCacheObjClassRec.object_class.resources,
                             xmLabelGCacheObjClassRec.object_class.num_resources,
                             &labels, &nlabels);

    combined = (XtResourceList)XtMalloc(sizeof(XtResource) * ncom);
    bcopy(labels, combined, nlabels * sizeof(XtResource));
    bcopy(cache_resources,
          &combined[nlabels],
          XtNumber(cache_resources) * sizeof(XtResource));
    XtFree((char *)labels);

    xmCascadeButtonGCacheObjClassRec.object_class.resources = combined;
    xmCascadeButtonGCacheObjClassRec.object_class.num_resources = ncom;
}

static void
class_part_initialize(WidgetClass widget_class)
{
    _XmFastSubclassInit(widget_class, XmCASCADE_BUTTON_GADGET_BIT);
}

static void
initialize_prehook(Widget request,
           Widget new,
           ArgList args,
           Cardinal *num_args)
{
    XdbDebug(__FILE__, new, "CascadeBG InitializePrehook\n");
}

static void
initialize_posthook(Widget request,
           Widget new,
           ArgList args,
           Cardinal *num_args)
{
    XmWidgetExtData ext;

    XdbDebug(__FILE__, new, "CascadeBG InitializePosthook\n");

    /* don't let the null fool you */
    LabG_Cache(new) = _XmCachePart(LabG_ClassCachePart(NULL),
                                   (XtPointer)LabG_Cache(new),
                                   sizeof(XmLabelGCacheObjPart));
    CBG_Cache(new) = _XmCachePart(CBG_ClassCachePart(NULL),
                                  (XtPointer)CBG_Cache(new),
                                  sizeof(XmCascadeButtonGCacheObjPart));

    _XmPopWidgetExtData(new, &ext, XmCACHE_EXTENSION);
    _XmExtObjFree((XtPointer)ext->widget);
    _XmExtObjFree((XtPointer)ext->reqWidget);
    XtFree((char *)ext);
}

static void
initialize(Widget request,
	   Widget new,
	   ArgList args,
	   Cardinal *num_args)
{
    Widget rc;

    if (! XmIsManager(XtParent(new)))
	_XmError(new, "parent should be manager.");

    CBG_SetArmed(new, False);

    rc = XtParent(new);

    if (!XmIsRowColumn(rc))
	_XmError(new, "Cascade gadget parent must be a RowColumn.");

    if (RC_Type(rc) != XmMENU_BAR      && RC_Type(rc) != XmMENU_POPUP &&
	RC_Type(rc) != XmMENU_PULLDOWN && RC_Type(rc) != XmMENU_OPTION)
	_XmError(new, "Cascade gadget parent is incorrect type.");

    if (RC_Type(rc) != XmMENU_BAR)	/* ??? */
	_XmCreateArrowPixmaps(new);

    LabG_Highlight(new) = 0;

    CBG_Timer(new) = 0;

    G_EventMask(new) = XmARM_EVENT | XmACTIVATE_EVENT |
			XmENTER_EVENT | XmLEAVE_EVENT |
			XmFOCUS_IN_EVENT | XmFOCUS_OUT_EVENT;

    G_TraversalOn(new) = True;	/* Danny 18/5/1996 - FIX ME (not sure) */

}

static void
destroy(Widget w)
{
    _XmCacheDelete(CBG_Cache(w));
}

static Boolean
set_values_prehook(Widget old,
                   Widget request,
                   Widget new,
                   ArgList args,
                   Cardinal *num_args)
{
    XmBaseClassExt *bce;
    XmWidgetExtData ed;
    int size;
    XtPointer nsec, rsec;

    bce = _XmGetBaseClassExtPtr(XtClass(new), XmQmotif);
    size = (*bce)->secondaryObjectClass->core_class.widget_size;

    nsec = _XmExtObjAlloc(size);
    rsec = _XmExtObjAlloc(size);

    ((XmExtRec *)nsec)->object.self = (Widget)nsec;
    ((XmExtRec *)nsec)->object.widget_class = (*bce)->secondaryObjectClass;
    ((XmExtRec *)nsec)->object.parent = XtParent(new);
    ((XmExtRec *)nsec)->object.xrm_name = new->core.xrm_name;
    ((XmExtRec *)nsec)->object.being_destroyed = False;
    ((XmExtRec *)nsec)->object.destroy_callbacks = NULL;
    ((XmExtRec *)nsec)->object.constraints = NULL;

    ExtObj_LogicalParent(nsec) = new;
    ExtObj_ExtensionType(nsec) = XmCACHE_EXTENSION;

    bcopy(LabG_Cache(new),
          &((XmLabelGCacheObject)nsec)->label_cache,
          sizeof(XmLabelGCacheObjPart));
    bcopy(CBG_Cache(new),
          &((XmCascadeButtonGCacheObject)nsec)->cascade_button_cache,
          sizeof(XmCascadeButtonGCacheObjPart));

    ed = (XmWidgetExtData)XtMalloc(sizeof(XmWidgetExtDataRec));
    ed->widget = (Widget)nsec;
    ed->reqWidget = (Widget)rsec;

    _XmPushWidgetExtData(new, ed, XmCACHE_EXTENSION);

    _XmGadgetImportSecondaryArgs(new, args, num_args);

    XtSetSubvalues((XtPointer)nsec,
                    (*bce)->secondaryObjectClass->core_class.resources,
                    (*bce)->secondaryObjectClass->core_class.num_resources,
                    args, *num_args);

    bcopy(nsec, rsec, size);

    LabG_Cache(new) = &(((XmCascadeButtonGCacheObject)nsec)->label_cache);
    LabG_Cache(request) = &(((XmCascadeButtonGCacheObject)rsec)->label_cache);
    CBG_Cache(new) =
	&(((XmCascadeButtonGCacheObject)nsec)->cascade_button_cache);
    CBG_Cache(request) =
	&(((XmCascadeButtonGCacheObject)rsec)->cascade_button_cache);

    _XmExtImportArgs((Widget)nsec, args, num_args);

    return FALSE;
}

static Boolean
set_values_posthook(Widget old,
                   Widget request,
                   Widget new,
                   ArgList args,
                   Cardinal *num_args)
{
    XmWidgetExtData ext;

    if (!_XmLabelCacheCompare(LabG_Cache(new), LabG_Cache(old))) {

	_XmCacheDelete((XtPointer)LabG_Cache(old));

	LabG_Cache(new) = _XmCachePart(LabG_ClassCachePart(NULL),
				       (XtPointer)LabG_Cache(new),
				       sizeof(XmLabelGCacheObjPart));
    }
    else
	LabG_Cache(new) = LabG_Cache(old);

    if (!_XmCascadeBCacheCompare(CBG_Cache(new), CBG_Cache(old))) {

	_XmCacheDelete((XtPointer)CBG_Cache(old));

	CBG_Cache(new) = _XmCachePart(CBG_ClassCachePart(NULL),
				      (XtPointer)CBG_Cache(new),
				      sizeof(XmCascadeButtonGCacheObjPart));
    }
    else
	CBG_Cache(new) = CBG_Cache(old);

    _XmPopWidgetExtData(new, &ext, XmCACHE_EXTENSION);

    _XmExtObjFree((XtPointer)ext->widget);
    _XmExtObjFree((XtPointer)ext->reqWidget);

    XtFree( (char *) ext);

    return FALSE;
}

static Boolean
set_values(Widget old,
	   Widget request,
	   Widget new,
	   ArgList args,
	   Cardinal *num_args)
{
    Boolean refresh_needed = False;

    XdbDebug(__FILE__, new, "set_values()\n");

    if (CBG_CascadePixmap(old) != CBG_CascadePixmap(new))
    {
        _XmFreeScratchPixmap((XmScreen)XmGetXmScreen(XtScreenOfObject(old)), CBG_CascadePixmap(old));
        _XmFreeScratchPixmap((XmScreen)XmGetXmScreen(XtScreenOfObject(old)), CBG_ArmedPixmap(old));

	_XmCreateArrowPixmaps(new);

	refresh_needed = True;
    }

    return refresh_needed;
}

static void
get_values_prehook(Widget new, ArgList args, Cardinal *num_args)
{
    XmBaseClassExt *bce;
    XmWidgetExtData ed;
    int size;
    XtPointer nsec;

    bce = _XmGetBaseClassExtPtr(XtClass(new), XmQmotif);
    size = (*bce)->secondaryObjectClass->core_class.widget_size;

    nsec = _XmExtObjAlloc(size);

    bcopy(LabG_Cache(new),
          &((XmLabelGCacheObject)nsec)->label_cache,
          sizeof(XmLabelGCacheObjPart));
    bcopy(CBG_Cache(new),
          &((XmCascadeButtonGCacheObject)nsec)->cascade_button_cache,
          sizeof(XmCascadeButtonGCacheObjPart));

    /*
     * don't do this and ResInd will blow up.
     */
    ((XmExtRec *)nsec)->object.self = (Widget)nsec;
    ((XmExtRec *)nsec)->object.widget_class = (*bce)->secondaryObjectClass;
    ((XmExtRec *)nsec)->object.parent = XtParent(new);
    ((XmExtRec *)nsec)->object.xrm_name = new->core.xrm_name;
    ((XmExtRec *)nsec)->object.being_destroyed = False;
    ((XmExtRec *)nsec)->object.destroy_callbacks = NULL;
    ((XmExtRec *)nsec)->object.constraints = NULL;

    ExtObj_LogicalParent(nsec) = new;
    ExtObj_ExtensionType(nsec) = XmCACHE_EXTENSION;

    ed = (XmWidgetExtData)XtMalloc(sizeof(XmWidgetExtDataRec));
    ed->widget = (Widget)nsec;

    _XmPushWidgetExtData(new, ed, XmCACHE_EXTENSION);

    XtGetSubvalues((XtPointer)nsec,
                    (*bce)->secondaryObjectClass->core_class.resources,
                    (*bce)->secondaryObjectClass->core_class.num_resources,
                    args, *num_args);

    _XmExtGetValuesHook((Widget)nsec, args, num_args);
}

static void
get_values_posthook(Widget new, ArgList args, Cardinal *num_args)
{
     XmWidgetExtData ext;

     _XmPopWidgetExtData(new, &ext, XmCACHE_EXTENSION);

     _XmExtObjFree((XtPointer)ext->widget);

     XtFree( (char *) ext);
}


static void
get_values_hook(Widget w, ArgList args, Cardinal *num_args)
{
    XdbDebug(__FILE__, w, "get_values_hook()\n");
}

static void
expose(Widget w,
       XEvent *event,
       Region region)
{
#define superclass (&xmLabelGadgetClassRec)
    (*superclass->rect_class.expose)(w, event, region);
#undef superclass

    XdbDebug(__FILE__, w, "expose() Armed(%d)\n", CBG_IsArmed(w));

    if (CBG_IsArmed(w) || LabG_MenuType(w) == XmMENU_OPTION)
	XmCascadeButtonHighlight(w, True);
    else
	XmCascadeButtonHighlight(w, False);
}

static Cardinal
get_sec_res_data(WidgetClass wc, XmSecondaryResourceData **data)
{
    /* FIXME */

    return _XmSecondaryResourceData(&_XmCascadeBGRectClassExtRec,
                                    data, NULL, NULL, NULL, NULL);
}

static void
StartDrag(Widget w,
	  XEvent *event,
	  String *params,
	  Cardinal *num_params)
{
    XmCascadeButtonGadget cw = (XmCascadeButtonGadget) w;

    XdbDebug(__FILE__, w, "StartDrag()\n");

    /* Arm the cascade button, post the submenu, and enable mouse
       traversal on it. */

    /* If the submenu is already active, disable keyboard traversal
       and set it to mouse traversal */

    if (CBG_Submenu(cw) == NULL) {
	XdbDebug(__FILE__, w, "StartDrag: no CBG_Submenu\n");
	return;
    }

    if (XmIsMenuShell(XtParent(CBG_Submenu(cw))))
    {
	Position my_x, my_y;

	XtTranslateCoords(w, 0, 0, &my_x, &my_y);

	XtMoveWidget(XtParent(CBG_Submenu(cw)),
                     my_x + XtWidth(w) - LabG_Shadow(cw),
                     my_y);

	XtCallCallbackList(w,
			   CBG_CascadeCall(w),
			   NULL);

	XtPopup(XtParent(CBG_Submenu(cw)), XtGrabNonexclusive);
    }
    else
    {
	_XmWarning(w, 
                   "CascadeButton's popup must be a subclass of XmMenuShell\n");
        return;
    }

    CBG_SetArmed(cw, True);

    expose(w, NULL, (Region)NULL);
}

static void 
CascadePopupHandler(XtPointer clientData, XtIntervalId *id)
{
    Widget w = (Widget)clientData;

    CBG_Timer(w) = 0;

    if (CBG_Submenu(w) && XmIsMenuShell(XtParent(CBG_Submenu(w))))
    {
	Position my_x, my_y;
	
	/* make sure the menu is popped down */

/* This seems to have problems */	
/*	XtUnmapWidget(CBG_Submenu(w));*/
	
	/* position the row column inside the menushell*/
	XtMoveWidget(CBG_Submenu(w),
		     0,0);

	/* now move the menushell */
	XtTranslateCoords(w, 0, 0, &my_x, &my_y);
	XtMoveWidget(XtParent(CBG_Submenu(w)),
                     my_x + XtWidth(w) - LabG_Shadow(w),
                     my_y);

	XtCallCallbackList(w,
			   CBG_CascadeCall(w),
			   NULL);

	RC_CascadeBtn(CBG_Submenu(w)) = w;
	XdbDebug0(__FILE__, CBG_Submenu(w), "RC_CascadeBtn(%s) -> %s\n",
		XtName(CBG_Submenu(w)), XtName(w));
	RC_PopupPosted(XtParent(w)) = XtParent(CBG_Submenu(w));

	XtPopup(XtParent(CBG_Submenu(w)), XtGrabNonexclusive);

	XtMapWidget(CBG_Submenu(w));
    }
    else if (CBG_Submenu(w))
    {
	_XmWarning(w,
                   "CascadeButton's popup must be a subclass of XmMenuShell\n");
        return;
    }
}

static void 
DelayedArm(Widget w, 
           XEvent *event, 
           String *params, 
           Cardinal *num_params)
{
    XdbDebug(__FILE__, w, "DelayedArm()\n");

    XAllowEvents(XtDisplay(w), SyncPointer, CurrentTime);

    if (_XmGetInDragMode(w)) 
    {
	CBG_Timer(w) = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
				       CBG_MapDelay(w),
				       CascadePopupHandler,
				       (XtPointer)w);

	XmCascadeButtonHighlight(w, True);
	
	CBG_SetArmed(w, True);

	MGR_SelectedGadget(XtParent(w)) = (XmGadget)w;
    }
}

static void 
CheckDisarm(Widget w, 
	    XEvent *event, 
	    String *params, 
	    Cardinal *num_params)
{
    XdbDebug(__FILE__, w, "CheckDisarm()\n");

    if (_XmGetInDragMode(w)) 
    {
	if (CBG_Timer(w))
	{
	    XtRemoveTimeOut(CBG_Timer(w));
	    CBG_Timer(w) = 0;
	}

	XmCascadeButtonHighlight(w, False);
	
	CBG_SetArmed(w, False);
    }
}

static void 
input_dispatch(Widget gadget, 
	       XEvent *event, 
	       Mask event_mask)
{
    Cardinal num_params = 0;

    switch (event_mask)
    {
    case XmARM_EVENT:
	if (LabG_MenuType(gadget) == XmMENU_BAR)
	    _XmCBMenuBarSelect(gadget, event, NULL, &num_params);
	else
	    StartDrag(gadget, event, NULL, &num_params);
	break;
    case XmACTIVATE_EVENT:
	if (LabG_MenuType(gadget) == XmMENU_BAR)
  	    _XmCBMenuBarDoSelect(gadget, event, NULL, &num_params);
	else
	    StartDrag(gadget, event, NULL, &num_params);
	break;
    case XmENTER_EVENT:
	if (LabG_MenuType(gadget) == XmMENU_PULLDOWN)
	    DelayedArm(gadget, event, NULL, &num_params);
	break;
    case XmLEAVE_EVENT:
	if (LabG_MenuType(gadget) == XmMENU_PULLDOWN)
	    CheckDisarm(gadget, event, NULL, &num_params);
    case XmFOCUS_IN_EVENT:
	_XmFocusInGadget(gadget, event, NULL, &num_params);
	break;
    case XmFOCUS_OUT_EVENT:
	_XmFocusOutGadget(gadget, event, NULL, &num_params);
	break;
    }
}

Widget
XmCreateCascadeButtonGadget(Widget parent,
		    char *name,
		    Arg *arglist,
		    Cardinal argcount)
{
    return XtCreateWidget(name,
			  xmCascadeButtonGadgetClass,
			  parent,
			  arglist,
			  argcount);
}
