/**
 *
 * $Id: DebugUtil.c,v 1.3 1996/08/12 02:13:56 miers Exp $
 * 
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static char rcsid[] = "$Id: DebugUtil.c,v 1.3 1996/08/12 02:13:56 miers Exp $";

#include <LTconfig.h>
#include <XmI/XmI.h>

#include <X11/Xos.h>
#include <Xm/GadgetP.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/Xresource.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <Xm/VendorSEP.h>

/* Protect ourselves (for now) */

#ifdef	PRINT_STATE
static char *_XdbState(Widget w)
{
	if (XtIsRealized(w)) {
		if (XtIsManaged(w))
			return "realized, managed";
		else
			return "realized, not managed";
	} else {
		if (XtIsManaged(w))
			return "not realized, not managed";
		else
			return "not realized, not managed";
	}
}
#endif

static void _XdbPrintTree(Widget w, int level)
{
	int	i;
	CompositeWidget	cw = (CompositeWidget)w;

	if (w == NULL)
		return;

	for (i=0; i<level; i++)
		fprintf(stderr, "\t");
	fprintf(stderr, "%s : %p/%ld", XtName(w), w, XtWindow(w));
	fprintf(stderr, "(%s) geo %d %d %d %d",
		w->core.widget_class->core_class.class_name,
		XtX(w), XtY(w), XtWidth(w), XtHeight(w));
#ifdef	PRINT_STATE
	fprintf(stderr, " state: %s %s",
		_XdbState(w), w->core.mapped_when_managed ? "mwm": "");
#endif
	fprintf(stderr, "\n");
	if (XtIsSubclass(w, compositeWidgetClass))
		for (i=0; i<cw->composite.num_children; i++)
			_XdbPrintTree(cw->composite.children[i], level+1);

	for (i=0; i<cw->core.num_popups; i++)
		_XdbPrintTree(cw->core.popup_list[i], level+1);
}

void XdbPrintTree(Widget w)
{
	_XdbPrintTree(w, 0);
}

void XdbPrintCompleteTree(Widget w)
{
	Widget	ww = w;

	while (ww) {
		w = ww;
		ww = XtParent(w);
	}

	_XdbPrintTree(w, 0);
}

char * XdbGeometryResult2String(XtGeometryResult r)
{
	switch (r) {
	case XtGeometryYes:	return "Yes";
	case XtGeometryNo:	return "No";
	case XtGeometryAlmost:	return "Almost";
	case XtGeometryDone:	return "Done";
	default:		return "(invalid geometry result)";
	}
}

char *
XdbAttachment2String(int a)
{
	switch (a) {
	case XmATTACH_FORM:		return "XmATTACH_FORM";
	case XmATTACH_OPPOSITE_FORM:	return "XmATTACH_OPPOSITE_FORM";
	case XmATTACH_WIDGET:		return "XmATTACH_WIDGET";
	case XmATTACH_OPPOSITE_WIDGET:	return "XmATTACH_OPPOSITE_WIDGET";
	case XmATTACH_NONE:		return "XmATTACH_NONE";
	case XmATTACH_POSITION:		return "XmATTACH_POSITION";
	default :			return "(invalid attachment)";
	}
}

/*
 * See if 'fn' is one of the elements in 'l'.
 * L is a colon-separated list of source files.
 * The ".c" suffix for sources is not required (i.e. Form is equivalent
 * to Form.c).
 *
 * As of post-0.50, the "fn" can contain a directory prefix.
 * Needs to be dealt with.
 */
static Boolean
ValidateSource(char *fn, char *l)
{
#ifndef	LESSTIF_PRODUCTION
	char	*s, *p, *q;
	int	i;

	if (l == NULL)
		return False;

	if (strcmp(l, "all") == 0)
		return True;

/* remove directory prefix */
	s = strrchr(fn, '/');
	if (s && *(s+1) != '\0')
		fn = s+1;
	i = strlen(fn);

	s = l;
	while ((p = strstr(s, fn)) != NULL) {
		q = p + i;
		if (*q == '\0' || *q == '.' || *q == ':')
			return True;
		s = q;
	}
#endif
	return False;
}

/*
 * This is the new style debugging support routine for LessTif.
 *
 * It allows us to choose at run time which sources should spit out
 * their guts.
 */
char *_XdbFindDebugSources(Widget w)
{
	return getenv("DEBUGSOURCES");
}

void XdbDebug(char *fn, Widget w, char *fmt, ...)
{
#ifndef LESSTIF_PRODUCTION
	va_list	ap;
	char	*dbs = _XdbFindDebugSources(w);

	if (ValidateSource(fn, dbs)) {
		if (w)
		    fprintf(stderr, "%s %s: ",
			w->core.widget_class->core_class.class_name,
			XtName(w));
		else
		    fprintf(stderr, "(null widget): ");
		va_start(ap, fmt);
		vfprintf(stderr, fmt, ap);
		va_end(ap);
		fflush(stderr);
	}
#endif
}

void XdbDebug2(char *fn, Widget w, Widget c, char *fmt, ...)
{
#ifndef LESSTIF_PRODUCTION
	va_list	ap;
	char	*dbs = _XdbFindDebugSources(w);

	if (ValidateSource(fn, dbs)) {
		if (w && c)
		    fprintf(stderr, "%s %s (child %s): ",
			w->core.widget_class->core_class.class_name,
			XtName(w), XtName(c));
		else if (w)
		    fprintf(stderr, "%s %s (child NULL): ",
			w->core.widget_class->core_class.class_name,
			XtName(w));
		else
		    fprintf(stderr, "(null widget): ");
		va_start(ap, fmt);
		vfprintf(stderr, fmt, ap);
		va_end(ap);
	}
#endif
}

void XdbDebug0(char *fn, Widget w, char *fmt, ...)
{
#ifndef LESSTIF_PRODUCTION
	va_list	ap;
	char	*dbs = _XdbFindDebugSources(w);

	if (ValidateSource(fn, dbs)) {
		va_start(ap, fmt);
		vfprintf(stderr, fmt, ap);
		va_end(ap);
	}
#endif
}

Boolean XdbInDebug(char *fn, Widget w)
{
#ifdef	LESSTIF_PRODUCTION
	return False;
#else
	return ValidateSource(fn, _XdbFindDebugSources(w));
#endif
}

char *
XdbMenuEnum2String(int f)
{
    switch(f) {
	case XmMENU_POPDOWN:		return "XmMENU_POPDOWN";
	case XmMENU_PROCESS_TREE:	return "XmMENU_PROCESS_TREE";
	case XmMENU_TRAVERSAL:		return "XmMENU_TRAVERSAL";
	case XmMENU_SHELL_POPDOWN:	return "XmMENU_SHELL_POPDOWN";
	case XmMENU_CALLBACK:		return "XmMENU_CALLBACK";
	case XmMENU_BUTTON:		return "XmMENU_BUTTON";
	case XmMENU_CASCADING:		return "XmMENU_CASCADING";
	case XmMENU_SUBMENU:		return "XmMENU_SUBMENU";
	case XmMENU_ARM:		return "XmMENU_ARM";
	case XmMENU_DISARM:		return "XmMENU_DISARM";
	case XmMENU_BAR_CLEANUP:	return "XmMENU_BAR_CLEANUP";
	case XmMENU_STATUS:		return "XmMENU_STATUS";
	case XmMENU_MEMWIDGET_UPDATE:	return "XmMENU_MEMWIDGET_UPDATE";
	case XmMENU_BUTTON_POPDOWN:	return "XmMENU_BUTTON_POPDOWN";
	case XmMENU_RESTORE_EXCLUDED_TEAROFF_TO_TOPLEVEL_SHELL:
					return "XmMENU_RESTORE_EXCLUDED_TEAROFF_TO_TOPLEVEL_SHELL";
	case XmMENU_RESTORE_TEAROFF_TO_TOPLEVEL_SHELL:
					return "XmMENU_RESTORE_TEAROFF_TO_TOPLEVEL_SHELL";
	case XmMENU_RESTORE_TEAROFF_TO_MENUSHELL:
					return "XmMENU_RESTORE_TEAROFF_TO_MENUSHELL";
	case XmMENU_GET_LAST_SELECT_TOPLEVEL:
					return "XmMENU_GET_LAST_SELECT_TOPLEVEL";
	case XmMENU_TEAR_OFF_ARM:	return "XmMENU_TEAR_OFF_ARM";
	default:			return "??";
    }
}

char *
XdbBoolean2String(int b)
{
	if (b)
		return "True";
	return "False";
}

char *
XdbXmString2String(XmString xms)
{
	char *s = NULL;

	XmStringGetLtoR(xms, XmFONTLIST_DEFAULT_TAG, &s);

	return s;
}

char *
XdbPacking2String(unsigned char p)
{
	static char res[40];

	switch(p) {
	case XmPACK_COLUMN:	return "XmPACK_COLUMN";
	case XmPACK_TIGHT:	return "XmPACK_TIGHT";
	case XmPACK_NONE:	return "XmPACK_NONE";
	default :
		sprintf(res, "Invalid packing %d", p);
		return res;
	}
}

char *
XdbRcType2String(unsigned char t)
{
	static char res[40];

	switch(t) {
	case XmWORK_AREA:	return "XmWORK_AREA";
	case XmMENU_BAR:	return "XmMENU_BAR";
	case XmMENU_PULLDOWN:	return "XmMENU_PULLDOWN";
	case XmMENU_POPUP:	return "XmMENU_POPUP";
	case XmMENU_OPTION:	return "XmMENU_OPTION";
	default :
		sprintf(res, "Invalid RC Type %d", t);
		return res;
	}
}
