/**
 *
 * $Id: RowColumn.c,v 1.7 1996/08/21 02:06:29 miers Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static char rcsid[] = "$Id: RowColumn.c,v 1.7 1996/08/21 02:06:29 miers Exp $";

#include <LTconfig.h>
#include <XmI/XmI.h>
#include <XmI/DebugUtil.h>

#include <X11/Xfuncs.h>
#include <Xm/XmP.h>
#include <Xm/BaseClassP.h>
#include <Xm/CascadeBP.h>
#include <Xm/CascadeBGP.h>
#include <Xm/DrawnBP.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/ManagerP.h>
#include <Xm/MenuShellP.h>
#include <Xm/MenuUtilP.h>
#include <Xm/PushBP.h>
#include <Xm/PushBGP.h>
#include <Xm/RepType.h>
#include <Xm/RowColumnP.h>
#include <Xm/ScreenP.h>
#include <Xm/Separator.h>
#include <Xm/SeparatoG.h>
#include <Xm/TearOffBP.h>
#include <Xm/ToggleBP.h>
#include <Xm/ToggleBGP.h>
#include <Xm/TransltnsP.h>

#include <stdio.h>
#include <string.h>

/*
 * builtin widgets for option menus
 */
#define RC_OPTION_LABEL		"OptionLabel"
#define RC_OPTION_CBG		"OptionButton"

/* Forward Declarations */

static void class_initialize();
static void class_part_initialize(WidgetClass class);
static void initialize(Widget request, Widget new, ArgList args, 
		       Cardinal * num_args);
static void initialize_prehook(Widget request, Widget new, ArgList args, 
			       Cardinal * num_args);
static void destroy(Widget w);
static void resize(Widget w);
static void expose(Widget w, XEvent * event, Region region);
static XtGeometryResult query_geometry(Widget w, XtWidgetGeometry * proposed, 
				       XtWidgetGeometry * answer);
static Boolean set_values(Widget current, Widget request, Widget new, 
			  ArgList args, Cardinal * num_args);
static XtGeometryResult RcGeometryManager(Widget w, XtWidgetGeometry * request,
					 XtWidgetGeometry * reply);
static void change_managed(Widget w);
static void insert_child(Widget w);
static void XmRowColumnLayout(Widget rc, int mode, Widget cw, XtWidgetGeometry *cg, XtWidgetGeometry *rcgp);
static void DoLayoutHT(Widget rc, int mode, Widget cw, XtWidgetGeometry *cg, XtWidgetGeometry *rcg);
static void DoLayoutVT(Widget rc, int mode, Widget cw, XtWidgetGeometry *cg, XtWidgetGeometry *rcg);
static void DoLayoutVC(Widget rc, int mode, Widget cw, XtWidgetGeometry *cg, XtWidgetGeometry *rcg);
static void DoLayoutHC(Widget w);
static void prefered_size(Widget rc, Dimension *needed_width, 
			  Dimension *needed_height);
static void _XmRadioCallback(Widget w, XtPointer cd, XtPointer cbs);
static void _XmEntryCallback(Widget w, XtPointer cd, XtPointer cbs);
static void initialize_boxes(Widget w);

static void _XmFromMenuPost(Widget widget, int offset, XtArgVal *value);
static XmImportOperator _XmToMenuPost(Widget widget, int offset, 
				     XtArgVal *value);

static Cardinal _XmRowColumnOrderProc(Widget widget);

#define Offset(field) XtOffsetOf(XmRowColumnRec, row_column.field)
#define MGR_Offset(field) XtOffsetOf(XmRowColumnRec, manager.field)
#define COMP_Offset(field) XtOffsetOf(XmRowColumnRec, composite.field)

/* Resources for the RowColumn class */
static XtResource resources[] =
{
    {
	XmNresizeWidth, XmCResizeWidth, XmRBoolean,
	sizeof(Boolean), Offset(resize_width),
	XmRImmediate, (XtPointer) True
    },
    {
	XmNresizeHeight, XmCResizeHeight, XmRBoolean,
	sizeof(Boolean), Offset(resize_height),
	XmRImmediate, (XtPointer) True
    },
    {
	XmNwhichButton, XmCWhichButton, XmRWhichButton,
	sizeof(unsigned int), Offset(postButton),
	XmRImmediate, (XtPointer) XmUNSPECIFIED
	/* add support for this!!! */
    },
    {
	XmNmenuPost, XmCMenuPost, XmRString,
	sizeof(String), Offset(menuPost),
	XmRString, (XtPointer) NULL
    },
    {
	XmNadjustLast, XmCAdjustLast, XmRBoolean,
	sizeof(Boolean), Offset(adjust_last),
	XmRImmediate, (XtPointer) True
    },
    {
	XmNmarginWidth, XmCMarginWidth, XmRHorizontalDimension,
	sizeof(Dimension), Offset(margin_width),
	XtRCallProc, (XtPointer) _XmRowColumnMarginDefault
	/*	XmRImmediate, (XtPointer)XmINVALID_DIMENSION*/
    },
    {
	XmNmarginHeight, XmCMarginHeight, XmRVerticalDimension,
	sizeof(Dimension), Offset(margin_height),
	XtRCallProc, (XtPointer) _XmRowColumnMarginDefault
	/*XmRImmediate, (XtPointer)XmINVALID_DIMENSION*/
    },
    {
	XmNentryCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(entry_callback),
	XmRCallback, (XtPointer) NULL
    },
    {
	XmNmapCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(map_callback),
	XmRCallback, (XtPointer) NULL
    },
    {
	XmNunmapCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(unmap_callback),
	XmRCallback, (XtPointer) NULL
    },
    {
	XmNorientation, XmCOrientation, XmROrientation,
	sizeof(unsigned char), Offset(orientation),
	XmRImmediate, (XtPointer) XmVERTICAL
	/*FIXME: Should be XmROrientation, (XtPointer)some whacko thing here */
    },
    {
	XmNspacing, XmCSpacing, XmRHorizontalDimension,
	sizeof(Dimension), Offset(spacing),
	XtRCallProc, (XtPointer) _XmRowColumnSpacingDefault
	/*XmRImmediate, (XtPointer)XmINVALID_DIMENSION*/
    },
    {
	XmNentryBorder, XmCEntryBorder, XmRHorizontalDimension,
	sizeof(Dimension), Offset(entry_border),
	XmRImmediate, (XtPointer) 0
    },
    {
	XmNisAligned, XmCIsAligned, XmRBoolean,
	sizeof(Boolean), Offset(do_alignment),
	XmRImmediate, (XtPointer) True
	/* FIXME: Motif has XmRBoolean, (XtPointer)whacko value */
    },
    {
	XmNentryAlignment, XmCAlignment, XmRAlignment,
	sizeof(unsigned char), Offset(entry_alignment),
	XtRImmediate, (XtPointer) XmALIGNMENT_BEGINNING
	/* FIXME: Motif has XmRAlignment, (XtPointer)whacko value */
    },
    {
	XmNadjustMargin, XmCAdjustMargin, XmRBoolean,
	sizeof(Boolean), Offset(adjust_margin),
	XtRImmediate, (XtPointer) True
	/* FIXME: Motif has XmRBoolean, (XtPointer)whacko value */
    },
    {
	XmNnumColumns, XmCNumColumns, XmRShort,
	sizeof(short), Offset(num_columns),
	XtRImmediate, (XtPointer) 1
	/* FIXME: Motif has XmRShort, (XtPointer)whacko value */
    },
    {
	XmNradioBehavior, XmCRadioBehavior, XmRBoolean,
	sizeof(Boolean), Offset(radio),
	XmRImmediate, (XtPointer) False
	/* FIXME: Motif has XmRBoolean, (XtPointer)whacko value */
    },
    {
	XmNradioAlwaysOne, XmCRadioAlwaysOne, XmRBoolean,
	sizeof(Boolean), Offset(radio_one),
	XmRImmediate, (XtPointer) True
	/* FIXME: Motif has XmRBoolean, (XtPointer)whacko value */
    },
    {
	XmNisHomogeneous, XmCIsHomogeneous, XmRBoolean,
	sizeof(Boolean), Offset(homogeneous),
	XtRCallProc, (XtPointer) _XmRowColumnIsHomogeneousDefault
	/* FIXME: Motif has XmRBoolean, (XtPointer)whacko value */
    },
    {
	XmNentryClass, XmCEntryClass, XmRWidgetClass,
	sizeof(WidgetClass), Offset(entry_class),
	XmRCallProc, (XtPointer) _XmRowColumnEntryClassDefault
	/* Motif has XmRWidgetClass, (XtPointer)NULL */
    },
    {
	XmNrowColumnType, XmCRowColumnType, XmRRowColumnType,
	sizeof(unsigned char), Offset(type),
	XtRImmediate, (XtPointer) XmWORK_AREA
	/* FIXME: Motif has XmRRowColumnType, (XtPointer)whacko value */
    },
    {
	XmNmenuHelpWidget, XmCMenuWidget, XmRMenuWidget,
	sizeof(Widget), Offset(help_pushbutton),
	XmRImmediate, (XtPointer) NULL
	/* FIXME: Motif has XmRMenuWidget, (XtPointer)whacko value */
    },
    {
	XmNlabelString, XmCXmString, XmRXmString,
	sizeof(XmString), Offset(option_label),
	XmRXmString, (XtPointer) NULL
    },
    {
	XmNsubMenuId, XmCMenuWidget, XmRMenuWidget,
	sizeof(Widget), Offset(option_submenu),
	XmRImmediate, (XtPointer) NULL
	/* FIXME: Motif has XmRMenuWidget, (XtPointer)whacko value */
    },
    {
	XmNmenuHistory, XmCMenuWidget, XmRMenuWidget,
	sizeof(Widget), Offset(memory_subwidget),
	XmRImmediate, (XtPointer) NULL
	/* FIXME: Motif has XmRMenuWidget, (XtPointer)whacko value */
    },
    {
	XmNpopupEnabled, XmCPopupEnabled, XmRBoolean,
	sizeof(Boolean), Offset(popup_enabled),
	XmRImmediate, (XtPointer) True
	/* FIXME: Motif has XmRBoolean, (XtPointer)whacko value */
    },
    {
	XmNmenuAccelerator, XmCAccelerators, XmRString,
	sizeof(String), Offset(menu_accelerator),
	XtRCallProc, (XtPointer) _XmRowColumnMenuAcceleratorDefault
	/* FIXME: Motif has XmRString, (XtPointer)whacko values */
    },
    {
	XmNmnemonic, XmCMnemonic, XmRKeySym,
	sizeof(KeySym), Offset(mnemonic),
	XmRImmediate, (XtPointer) NULL
    },
    {
	XmNmnemonicCharSet, XmCMnemonicCharSet, XmRString,
	sizeof(String), Offset(mnemonicCharSet),
	XtRImmediate, (XtPointer) XmFONTLIST_DEFAULT_TAG
    },
    {
	XmNshadowThickness, XmCShadowThickness, XmRHorizontalDimension,
	sizeof(Dimension), MGR_Offset(shadow_thickness),
	XmRImmediate, (XtPointer) 0
	/* FIXME: Motif has XmRImmediate, (XtPointer)XmINVALID_DIMENSION */
    },
    {
	XmNpostFromList, XmCPostFromList, XmRWidgetList,
	sizeof(WidgetList), Offset(postFromList),
	XmRWidgetList, (XtPointer) NULL
    },
    {
	XmNpostFromCount, XmCPostFromCount, XmRInt,
	sizeof(int), Offset(postFromCount),
	XmRImmediate, (XtPointer) XmUNSPECIFIED
    },
    {
	XmNnavigationType, XmCNavigationType, XmRNavigationType,
	sizeof(XmNavigationType), MGR_Offset(navigation_type),
	XmRImmediate, (XtPointer) ((XmNavigationType) XmUNSPECIFIED)
    },
    {
      XmNentryVerticalAlignment, XmCVerticalAlignment, XmRVerticalAlignment,
	sizeof(unsigned char), Offset(entry_vertical_alignment),
	XtRImmediate, (XtPointer) XmALIGNMENT_BEGINNING
	/* FIXME: Motif has XmRVerticalAlignment, (XtPointer)whacko values */
    },
    {
	XmNtearOffModel, XmCTearOffModel, XmRTearOffModel,
	sizeof(unsigned char), Offset(TearOffModel),
	XtRImmediate, (XtPointer) XmTEAR_OFF_DISABLED
	/* FIXME: Motif has XmRTearOffModel, (XtPointer)whacko values */
    },
    {
	XmNtearOffMenuActivateCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(tear_off_activated_callback),
	XmRCallback, (XtPointer) NULL
    },
    {
	XmNtearOffMenuDeactivateCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(tear_off_deactivated_callback),
	XmRCallback, (XtPointer) NULL
    },
    {
	XmNinsertPosition, XmCInsertPosition, XmRFunction,
	sizeof(XtOrderProc), COMP_Offset(insert_position),
	XmRImmediate, (XtPointer) _XmRowColumnOrderProc,
    },
    {
	XmNpacking, XmCPacking, XmRPacking,
	sizeof(unsigned char), Offset(packing),
	XmRImmediate, (XtPointer) XmUNSPECIFIED		/* To be overruled in initialize */
    }
};

static XmSyntheticResource syn_resources[] =
{
    {
	XmNmnemonicCharSet,
	sizeof(String), Offset(mnemonicCharSet),
	NULL /* FIXME */ , NULL
    },
    {
	XmNmenuAccelerator,
	sizeof(String), Offset(menu_accelerator),
	NULL /* FIXME */ , NULL
    },
    {
	XmNmenuPost,
	sizeof(String), Offset(menuPost),
	_XmFromMenuPost, _XmToMenuPost,
    },
    {
	XmNlabelString,
	sizeof(XmString), Offset(option_label),
	NULL /* FIXME */ , NULL
    },
    {
	XmNspacing,
	sizeof(Dimension), Offset(spacing),
	_XmFromHorizontalPixels, _XmToHorizontalPixels
    },
    {
	XmNmarginHeight,
	sizeof(Dimension), Offset(margin_height),
	_XmFromVerticalPixels, _XmToVerticalPixels
    },
    {
	XmNmarginWidth,
	sizeof(Dimension), Offset(margin_width),
	_XmFromHorizontalPixels, _XmToHorizontalPixels
    },
    {
	XmNentryBorder,
	sizeof(Dimension), Offset(entry_border),
	_XmFromHorizontalPixels, _XmToHorizontalPixels
    }
};

#undef Offset

#define Offset(field) XtOffsetOf(XmRowColumnConstraintRec, row_column.field)

static XtResource rowColumnConstraintResources[] =
{
    {
	XmNpositionIndex, XmCPositionIndex, XmRShort,
	sizeof(short), Offset(position_index),
	XmRImmediate, (XtPointer) XmLAST_POSITION
    }
};

/* not all actually written yet... */
#define STATIC_ACTION(fn) static void (fn)(Widget, XEvent*, String*, Cardinal*)
#define ACTION(fn) void (fn)(Widget, XEvent*, String*, Cardinal*)

STATIC_ACTION(MenuBarEnter);
STATIC_ACTION(MenuBarLeave);
STATIC_ACTION(MenuBarGadgetSelect);
STATIC_ACTION(MenuBtnDown);
STATIC_ACTION(MenuBtnUp);
ACTION(_XmMenuHelp);
STATIC_ACTION(MenuEnter);
STATIC_ACTION(MenuUnmap);
STATIC_ACTION(MenuFocusIn);
STATIC_ACTION(MenuFocusOut);
STATIC_ACTION(DoBtnEventCleanupReplay);

static Boolean ExternalBtnEvent(Widget w, XEvent * event);
static void _XmPopupButtonPressHandler(Widget w, XtPointer client_data,
				       XEvent *event, Boolean *cont);
#undef STATIC_ACTION
#undef ACTION

char _XmRowColumn_menu_traversal_table[] =
"<Unmap>:              MenuUnmap()\n\
 <EnterWindow>Normal:  MenuEnter()\n\
 <FocusIn>:            MenuFocusIn()\n\
 <FocusOut>:           MenuFocusOut()\n\
 <Key>osfHelp:         MenuHelp()\n\
 <Key>osfLeft:         MenuGadgetTraverseLeft()\n\
 <Key>osfRight:        MenuGadgetTraverseRight()\n\
 <Key>osfUp:           MenuGadgetTraverseUp()\n\
 <Key>osfDown:         MenuGadgetTraverseDown()";

char _XmRowColumn_bar_table[] =
"<BtnDown>:            MenuBtnDown()\n\
 <BtnUp>:              MenuBtnUp()\n\
 <Key>osfSelect:       MenuBarGadgetSelect()\n\
 <Key>osfActivate:     MenuBarGadgetSelect()\n\
 <Key>osfHelp:         MenuHelp()\n\
 <Key>osfCancel:       MenuGadgetEscape()\n\
 ~s ~m ~a <Key>Return: MenuBarGadgetSelect()\n\
 ~s ~m ~a <Key>space:  MenuBarGadgetSelect()";

char _XmRowColumn_option_table[] =
"<BtnDown>:             MenuBtnDown()\n\
 <BtnUp>:               MenuBtnUp()\n\
 <Key>osfActivate:      ManagerParentActivate()\n\
 <Key>osfCancel:        ManagerParentCancel()\n\
 <Key>osfSelect:        ManagerGadgetSelect()\n\
 <Key>osfHelp:          MenuHelp()\n\
 ~s ~m ~a <Key>Return:  ManagerParentActivate()\n\
 ~s ~m ~a <Key>space:   ManagerGadgetSelect()";

char _XmRowColumn_menu_table[] =
"<BtnDown>:             MenuBtnDown()\n\
 <BtnUp>:               MenuBtnUp()\n\
 <Key>osfActivate:      ManagerGadgetSelect()\n\
 <Key>osfSelect:        ManagerGadgetSelect()\n\
 <Key>osfCancel:        MenuGadgetEscape()\n\
 <Key>osfHelp:          MenuHelp()\n\
 ~s ~m ~a <Key>Return:  ManagerGadgetSelect()\n\
 ~s ~m ~a <Key>space:   ManagerGadgetSelect()";

static XtActionsRec actions[] =
{
    {"MenuBarEnter", MenuBarEnter},
    {"MenuBarLeave", MenuBarLeave},
    {"MenuBarGadgetSelect", MenuBarGadgetSelect},
    {"MenuBtnDown", MenuBtnDown},
    {"MenuBtnUp", MenuBtnUp},
    {"MenuFocusIn", MenuFocusIn},
    {"MenuFocusOut", MenuFocusOut},
    {"MenuHelp", _XmMenuHelp},
    {"MenuUnmap", MenuUnmap},
    {"MenuEnter", MenuEnter},
    {"MenuGadgetEscape", _XmMenuEscape},
    {"MenuGadgetTraverseLeft", _XmRC_GadgetTraverseLeft},
    {"MenuGadgetTraverseRight", _XmRC_GadgetTraverseRight},
    {"MenuGadgetTraverseUp", _XmRC_GadgetTraverseUp},
    {"MenuGadgetTraverseDown", _XmRC_GadgetTraverseDown},
};

static XmBaseClassExtRec _XmRowColumnCoreClassExtRec =
{
    /* next_extension            */ NULL,
    /* record_type               */ NULLQUARK,
    /* version                   */ XmBaseClassExtVersion,
    /* size                      */ sizeof(XmBaseClassExtRec),
    /* initialize_prehook        */ initialize_prehook,
    /* set_values_prehook        */ NULL, /* FIXME */
    /* initialize_posthook       */ NULL, /* FIXME */
    /* set_values_posthook       */ NULL, /* FIXME */
    /* secondary_object_class    */ NULL, /* FIXME */
    /* secondary_object_create   */ NULL, /* FIXME */
    /* get_secondary_resources   */ NULL, /* FIXME */
    /* fast_subclass             */ {0},  /* FIXME */
    /* get_values_prehook        */ NULL, /* FIXME */
    /* get_values_posthook       */ NULL, /* FIXME */
    /* class_part_init_prehook   */ NULL, /* FIXME */
    /* class_part_init_posthook  */ NULL, /* FIXME */
    /* ext_resources             */ NULL, /* FIXME */
    /* compiled_ext_resources    */ NULL, /* FIXME */
    /* num_ext_resources         */ 0,    /* FIXME */
    /* use_sub_resources         */ FALSE,/* FIXME */
    /* widget_navigable          */ XmInheritWidgetNavigable,
    /* focus_change              */ XmInheritFocusChange,
    /* wrapper_data              */ NULL
};

static CompositeClassExtensionRec rcCompositeExt =
{
    /* next_extension */ NULL,
    /* record_type    */ NULLQUARK,
    /* version        */ XtCompositeExtensionVersion,
    /* record_size    */ sizeof(CompositeClassExtensionRec),
    /* accepts_objects */ True,
#if XtSpecificationRelease >= 6
    /* allows_change_managed_set */ True
#endif
};

static XmManagerClassExtRec _XmRowColumnMClassExtRec =
{
    /* next_extension            */ NULL,
    /* record_type               */ NULLQUARK,
    /* version                   */ XmManagerClassExtVersion,
    /* record_size               */ sizeof(XmManagerClassExtRec),
    /* traversal_children        */ NULL /* FIXME */
};

XmRowColumnClassRec xmRowColumnClassRec =
{
    /* Core class part */
    {
	/* superclass            */ (WidgetClass) & xmManagerClassRec,
	/* class_name            */ "XmRowColumn",
	/* widget_size           */ sizeof(XmRowColumnRec),
	/* class_initialize      */ class_initialize,
	/* class_part_initialize */ class_part_initialize,
	/* class_inited          */ FALSE,
	/* initialize            */ initialize,
	/* initialize_hook       */ NULL,
	/* realize               */ XtInheritRealize,
	/* actions               */ actions,
	/* num_actions           */ XtNumber(actions),
	/* resources             */ resources,
	/* num_resources         */ XtNumber(resources),
	/* xrm_class             */ NULLQUARK,
	/* compress_motion       */ TRUE,
	/* compress_exposure     */ XtExposeCompressMultiple,
	/* compress_enterleave   */ TRUE,
	/* visible_interest      */ TRUE,
	/* destroy               */ destroy,
	/* resize                */ resize,
	/* expose                */ expose,
	/* set_values            */ set_values,
	/* set_values_hook       */ NULL,
	/* set_values_almost     */ XtInheritSetValuesAlmost,
	/* get_values_hook       */ NULL,
	/* accept_focus          */ NULL,
	/* version               */ XtVersion,
	/* callback offsets      */ NULL,
	/* tm_table              */ XtInheritTranslations,
	/* query_geometry        */ query_geometry,
	/* display_accelerator   */ XtInheritDisplayAccelerator,
	/* extension             */ (XtPointer) & _XmRowColumnCoreClassExtRec
    },
    /* Composite class part */
    {
	/* geometry manager */ RcGeometryManager,
	/* change_managed   */ change_managed,
	/* insert_child     */ insert_child,
	/* delete_child     */ XtInheritDeleteChild,
	/* extension        */ (XtPointer) & rcCompositeExt,
    },
    /* Constraint class part */
    {
	/* subresources      */ rowColumnConstraintResources,
	/* subresource_count */ XtNumber(rowColumnConstraintResources),
	/* constraint_size   */ sizeof(XmRowColumnConstraintRec),
	/* initialize        */ NULL,
	/* destroy           */ NULL,
	/* set_values        */ NULL,
	/* extension         */ NULL,
    },
    /* XmManager class part */
    {
	/* translations           */ XmInheritTranslations,
	/* syn_resources          */ syn_resources,
	/* num_syn_resources      */ XtNumber(syn_resources),
	/* syn_constraint_res     */ NULL,
	/* num_syn_constraint_res */ 0,
	/* parent_process         */ XmInheritParentProcess,
	/* extension              */ (XtPointer) & _XmRowColumnMClassExtRec
    },
    /* XmRowColumn Area part */
    {
	/* menuProcedures   */ NULL, /* FIXME */
	/* armAndActivate   */ NULL, /* FIXME */
	/* traversalHandler */ NULL, /* FIXME */
	/* extension        */ NULL, /* FIXME */
    },
};

WidgetClass xmRowColumnWidgetClass = (WidgetClass) & xmRowColumnClassRec;

static void
class_initialize()
{
    _XmRowColumnCoreClassExtRec.record_type = XmQmotif;
}

static void
class_part_initialize(WidgetClass widget_class)
{
    CompositeClassExtension ext, *extptr;
    XmRowColumnWidgetClass rc_class = (XmRowColumnWidgetClass) widget_class;

    extptr = (CompositeClassExtension *) 
      _XmGetClassExtensionPtr((XmGenericClassExt *) 
			         &(rc_class->composite_class.extension),
			      NULLQUARK);

    if (extptr == NULL || *extptr == NULL) {
	ext = (CompositeClassExtension) XtNew(CompositeClassExtensionRec);
	if (ext != NULL) {
	    ext->next_extension = rc_class->composite_class.extension;
	    ext->record_type = NULLQUARK;
	    ext->version = XtCompositeExtensionVersion;
	    ext->record_size = sizeof(CompositeClassExtensionRec);
	    ext->accepts_objects = True;
#if XtSpecificationRelease >= 6
	    ext->allows_change_managed_set = True;
#endif
	    rc_class->composite_class.extension = (XtPointer) ext;
	}
    }

    _XmFastSubclassInit(widget_class, XmROW_COLUMN_BIT);
}

static void
initialize_prehook(Widget request,
		   Widget new,
		   ArgList args,
		   Cardinal * num_args)
{
    if (RC_Type(new) == XmWORK_AREA)
	MGR_NavigationType(new) = XmTAB_GROUP;
    else
	MGR_NavigationType(new) = XmNONE;

#if 1
/*
 * I think all cases of RC_Type turn out to be traverable...
 * Danny 23/6/96
 */
    MGR_TraversalOn(new) = True;
#else
    if (RC_Type(new) == XmWORK_AREA || RC_Type(new) == XmMENU_OPTION)
	MGR_TraversalOn(new) = True;
    else
	MGR_TraversalOn(new) = False;
#endif
}

static void
initialize(Widget request,
	   Widget new,
	   ArgList args,
	   Cardinal * num_args)
{
    if (RC_Type(new) == XmMENU_BAR || RC_Type(new) == XmMENU_POPUP
	|| RC_Type(new) == XmMENU_PULLDOWN) {
	MGR_ShadowThickness(new) = 2;
    }

    /* now, we install our translations, 
     * depending on the row column type */

    if (RC_Type(new) != XmWORK_AREA) {
	XtOverrideTranslations(new,
			  XtParseTranslationTable(_XmRowColumn_menu_table));
	XtOverrideTranslations(new,
		XtParseTranslationTable(_XmRowColumn_menu_traversal_table));
    }

    if (RC_Type(new) == XmMENU_OPTION)
	XtOverrideTranslations(new,
			XtParseTranslationTable(_XmRowColumn_option_table));
    else if (RC_Type(new) == XmMENU_BAR)
	XtOverrideTranslations(new,
			   XtParseTranslationTable(_XmRowColumn_bar_table));

    RC_Boxes(new) = NULL;	/* no initial children */

    /* Internal widgets */

    RC_CascadeBtn(new) = NULL;
    RC_MemWidget(new) = NULL;
    RC_PopupPosted(new) = NULL;
    RC_LastSelectToplevel(new) = NULL;

    /* Tear off stuff */
    RC_SetTornOff(new, 0);
    RC_SetFromInit(new, 0);
    RC_SetTearOffDirty(new, 0);
    RC_SetTearOffActive(new, 0);
    RC_ParentShell(new) = NULL;
    RC_TearOffControl(new) = NULL;

    if (RC_Type(new) == XmMENU_OPTION) {
	XtVaCreateManagedWidget(RC_OPTION_LABEL,
				xmLabelGadgetClass,
				new,
				XmNlabelString, RC_OptionLabel(new),
				NULL);

	XtVaCreateManagedWidget(RC_OPTION_CBG,
				xmCascadeButtonGadgetClass,
				new,
				XmNsubMenuId, RC_OptionSubMenu(new),
				NULL);
    }

    /* just to initialize it. */
    RC_SetFromResize(new, 0);

    /* menuPost stuff. */
    if (RC_MenuPost(new) == NULL) 
      if (RC_Type(new) == XmMENU_OPTION
	  || RC_Type(new) == XmWORK_AREA
	  || RC_Type(new) == XmMENU_BAR)
	RC_MenuPost(new) =  "<Button1Down>";

    if (RC_Packing(new) == (unsigned char) XmUNSPECIFIED) {
/*
 * This is what the XmRowColumn manual page says
 *	(end of the text describing XmNpacking).
 */
	if (RC_RadioBehavior(new) && RC_Type(new) == XmWORK_AREA)
	    RC_Packing(new) = XmPACK_COLUMN;
	else if (RC_Type(new) == XmMENU_OPTION)
	    RC_Packing(new) = XmPACK_TIGHT;
	else
	    RC_Packing(new) = XmPACK_TIGHT;
    }

    /* Be verbose about it ... */
    if (XdbInDebug(__FILE__, new)) 
    {
	XdbDebug(__FILE__, 
		 new, 
		 "Initialize: RadioBehavior %s, RC_Type %s => Packing %s\n",
		 XdbBoolean2String(RC_RadioBehavior(new)), 
		 XdbRcType2String(RC_Type(new)),
		 XdbPacking2String(RC_Packing(new)));
    }
}

static void
destroy(Widget w)
{
    if (RC_Boxes(w))
	XtFree((char *)RC_Boxes(w));

    if (RC_TearOffControl(w))
	XtDestroyWidget(RC_TearOffControl(w));
}

static Boolean
set_values(Widget old,
	   Widget request,
	   Widget new,
	   ArgList args,
	   Cardinal * num_args)
{
    Boolean need_refresh = False;

    XdbDebug(__FILE__, new, "RowColumn set_values()\n");

    if ((RC_Orientation(old) != RC_Orientation(new))
		|| (RC_Packing(old) != RC_Packing(new))) {

	XtFree((char *)RC_Boxes(old));

	initialize_boxes(new);

	XmRowColumnLayout(new, Mode_Normal, NULL, NULL, NULL);

	need_refresh = True;
    }

    if ((RC_TearOffModel(new) != RC_TearOffModel(old))
	&& (RC_Type(new) == XmMENU_POPUP
	    || RC_Type(new) == XmMENU_PULLDOWN))
    {
	if (RC_TearOffModel(new) == XmTEAR_OFF_ENABLED)
	    RC_TearOffControl(new) = 
	      XtVaCreateManagedWidget("TearOffControl",
				      xmTearOffButtonWidgetClass,
				      new,
				      XmNpositionIndex, 0, 
				      NULL);
	else
	    XtDestroyWidget(MGR_Children(old)[0]); /* FIX ME */
    }

    if (RC_Type(new) == XmMENU_OPTION 
	&& RC_OptionSubMenu(new) != RC_OptionSubMenu(old)) 
    {
	int	i;

	/* Oh yes this is very efficient :-( */

	for (i=0; i<MGR_NumChildren(new); i++)
	    if (XtIsSubclass(MGR_Children(new)[i], 
			     xmCascadeButtonGadgetClass)) 
            {
		XtVaSetValues(MGR_Children(new)[i], 
			      XmNsubMenuId, 
			      RC_OptionSubMenu(new), 
			      NULL);
		XdbDebug2(__FILE__, 
			  new, 
			  MGR_Children(new)[i], 
			  "Assign SubMenuId (%s)\n",
			  XtName(RC_OptionSubMenu(new)));
		break;
	    }
    }

    if (XtSensitive(new) != XtSensitive(old)) {
	XdbDebug(__FILE__, new, "SetValues: sensitive changes to %d\n", XtSensitive(new));
#if 0
/* Doesn't seem to be necessary */
	if (XtSensitive(new)) {
	    int i;
	    for (i=0; i<MGR_NumChildren(new); i++)
		XtSetSensitive(MGR_Children(new)[i], True);
	}
#endif
	need_refresh = True;
    }

    return need_refresh;
}

static void
expose(Widget w,
       XEvent * event,
       Region region)
{
    _XmDrawShadows(XtDisplay(w),
		   XtWindow(w),
		   MGR_TopShadowGC(w),
		   MGR_BottomShadowGC(w),
		   0, 0,
		   XtWidth(w),
		   XtHeight(w),
		   MGR_ShadowThickness(w),
		   XmSHADOW_OUT);

    /* display the gadgets, if there are any */
    _XmRedisplayGadgets(w, event, region);
}

static XtGeometryResult
query_geometry(Widget w,
	       XtWidgetGeometry * request,
	       XtWidgetGeometry * reply)
{
    XtGeometryResult result = XtGeometryYes;
    Dimension width, height;

    XdbDebug(__FILE__, w, "RC QueryGeometry\n");

/* Prefered_size should be obsoleted soon, and replaced with
    XmRowColumnLayout(w, Mode_Test, NULL, NULL, &rcg);
*/
    prefered_size(w, &width, &height);

    if (XdbInDebug(__FILE__, w)) 
    {
	XdbDebug(__FILE__, w, 
		 "prefered_size calculates w/h %d %d\n", 
		 width, height);

	if (request->request_mode & CWWidth)
	    XdbDebug(__FILE__, w, "preferred width %d\n", request->width);
	if (request->request_mode & CWHeight)
	    XdbDebug(__FILE__, w, "preferred height %d\n", request->height);
    }

    reply->width = width;
    reply->height = height;

    /* Not interested in more flags */
    request->request_mode &= CWWidth | CWHeight;	

    if (request->request_mode == 0) {
	reply->request_mode = CWWidth | CWHeight;
	XdbDebug(__FILE__, w, 
		 "RC QueryGeometry(NULL) => XtGeometryYes, %d %d\n", 
		 width, height);
	return result;
    }

    if (request->request_mode & CWWidth)
	if (request->width < width) 
        {
	    result = XtGeometryAlmost;
	    reply->width = width;
	    reply->request_mode |= CWWidth;
	}
    if (request->request_mode & CWHeight)
	if (request->height < height) 
        {
	    result = XtGeometryAlmost;
	    reply->height = height;
	    reply->request_mode |= CWHeight;
	}

    reply->width = width;
    reply->height = height;
    reply->request_mode = CWWidth | CWHeight;

    XdbDebug(__FILE__, w, 
	     "RC QueryGeometry() => %s, %d %d\n",
	     XdbGeometryResult2String(result), width, height);

    return result;
}

static void
initialize_boxes(Widget w)
{
    int i;
    XmRCKidGeometry kid_geometry;

    RC_Boxes(w) = (XmRCKidGeometry) XtCalloc(MGR_NumChildren(w), 
					     sizeof(XmRCKidGeometryRec));

    for (i = 0; i < MGR_NumChildren(w); i++) 
    {
	Dimension *baselines;
	int number_of_baselines;

	kid_geometry = &(RC_Boxes(w)[i]);
	kid_geometry->kid = MGR_Children(w)[i];

	if (kid_geometry->kid == NULL)
	  continue;

	XtQueryGeometry(kid_geometry->kid, NULL, &(RC_Boxes(w)[i].box));

	kid_geometry->margin_top =
	    kid_geometry->margin_bottom = 0;

	XtVaGetValues(kid_geometry->kid,
		      XmNmarginTop, &kid_geometry->margin_top,
		      XmNmarginBottom, &kid_geometry->margin_bottom,
		      NULL);

	if (XmWidgetGetBaselines(kid_geometry->kid, 
				 &baselines, 
				 &number_of_baselines)) 
        {
	    /* we set the baseline of the child to the lowest baseline of
	     * the widget */
	    kid_geometry->baseline = baselines[number_of_baselines - 1];

	    XtFree((char *)baselines);
	}
	else
	    kid_geometry->baseline = 0;		/* is this right ? */
    }
}

/* This thing to be obsoleted real soon now */
/* FIX ME */
static void
prefered_size_HT(Widget w,
		 Dimension *needed_width,
		 Dimension *needed_height)
{
    int i;
    Dimension current_x, current_y, max_width, max_height;
    XmRCKidGeometry kid_geometry;
    Widget kid;

    current_x = RC_MarginW(w) + MGR_ShadowThickness(w);
    current_y = RC_MarginH(w) + MGR_ShadowThickness(w);
    max_width = 0;
    max_height = 0;

    if (!RC_Boxes(w)) 
    {
	*needed_width = 2 * (RC_MarginW(w) + MGR_ShadowThickness(w));
	*needed_height = 2 * (RC_MarginH(w) + MGR_ShadowThickness(w));
	return;
    }

    for (i = 0; i < MGR_NumChildren(w); i++) 
    {
	kid_geometry = &(RC_Boxes(w)[i]);
	kid = kid_geometry->kid;

	if (!kid || !XtIsManaged(kid))
	    continue;

	if (XtIsRealized((Widget) w) 
	    && current_x + kid_geometry->box.width > XtWidth(w)) 
        {
	    current_y += max_height + RC_Spacing(w);
	    current_x = RC_MarginW(w) + MGR_ShadowThickness(w);
	}

	/* don't add the spacing after the last child */
	if (i == MGR_NumChildren(w) - 1)	
	    current_x += kid_geometry->box.width;
	else
	    current_x += RC_Spacing(w) + kid_geometry->box.width;

	if (current_x > max_width)
	    max_width = current_x;

	if (kid_geometry->box.height + current_y > max_height)
	    max_height = kid_geometry->box.height + current_y;
    }

    *needed_width = max_width + RC_MarginW(w) + MGR_ShadowThickness(w);
    *needed_height = max_height + RC_MarginH(w) + MGR_ShadowThickness(w);
}

static void
prefered_size_HC(Widget w,
		 Dimension *needed_width,
		 Dimension *needed_height)
{
    int i;
    int number_per_row;
    Dimension width_of_widgets = 0;
    Dimension *height_of_row;

    if (RC_NCol(w) == 0)
	/* what the hell.  punt and go to .. */
	prefered_size_HT(w, needed_width, needed_height);

    XdbDebug(__FILE__, w, "prefered_size_HC()\n");

#define ROW_OF_THIS_WIDGET(i) (int)((float)(i) / number_per_row + 0.5)

    /* we need to round up */
    number_per_row = (int)((float)MGR_NumChildren(w) / RC_NCol(w) + 0.5);

    height_of_row = (Dimension *) XtCalloc(RC_NCol(w), sizeof(Dimension));

    for (i = 0; i < RC_NCol(w); i++)
	height_of_row[i] = 0;

    /* first, figure out the width of each column and the height of each row */

    for (i = 0; i < MGR_NumChildren(w); i++) 
    {
	XmRCKidGeometry kid_geometry = &(RC_Boxes(w)[i]);
	Widget kid = kid_geometry->kid;

	if (!kid || !XtIsManaged(kid))
	    continue;

	if (width_of_widgets < kid_geometry->box.width)
	    width_of_widgets = kid_geometry->box.width;

	if (height_of_row[ROW_OF_THIS_WIDGET(i)] < kid_geometry->box.height)
	    height_of_row[ROW_OF_THIS_WIDGET(i)] = kid_geometry->box.height;
    }

    /* then we multiply and add :) */

    *needed_width = (2 * (RC_MarginW(w) + MGR_ShadowThickness(w))
		     + width_of_widgets * number_per_row
		     + RC_Spacing(w) * (number_per_row - 1));

    *needed_height = (2 * (RC_MarginH(w) + MGR_ShadowThickness(w))
		      + RC_Spacing(w) * (RC_NCol(w) - 1));

    for (i = 0; i < RC_NCol(w); i++)
	*needed_height += height_of_row[i];
}

static void
prefered_size_VT(Widget w,
		 Dimension *needed_width,
		 Dimension *needed_height)
{
    int i;
    Dimension current_x = RC_MarginW(w) + MGR_ShadowThickness(w);
    Dimension current_y = RC_MarginH(w) + MGR_ShadowThickness(w);
    Dimension max_width = 0;
    Dimension max_height = 0;

    if (!RC_Boxes(w)) 
    {
	*needed_width = 2 * (RC_MarginW(w) + MGR_ShadowThickness(w));
	*needed_height = 2 * (RC_MarginH(w) + MGR_ShadowThickness(w));
	return;
    }

    for (i = 0; i < MGR_NumChildren(w); i++) 
    {
	XmRCKidGeometry kid_geometry = &(RC_Boxes(w)[i]);
	Widget kid = kid_geometry->kid;

	if (!kid || !XtIsManaged(kid))
	    continue;

	if (XtIsRealized(w) 
	    && current_y + kid_geometry->box.height > XtHeight(w)) 
        {
	    current_x += max_width + RC_Spacing(w);
	    current_y = RC_MarginH(w) + MGR_ShadowThickness(w);
	}

	/* don't add the spacing after the last child */
	if (i == MGR_NumChildren(w) - 1)	
	    current_y += kid_geometry->box.height;
	else
	    current_y += RC_Spacing(w) + kid_geometry->box.height;

	if (current_y > max_height)
	    max_height = current_y;

	if (kid_geometry->box.width + current_x > max_width)
	    max_width = kid_geometry->box.width + current_x;
    }

    *needed_width = max_width + RC_MarginW(w) + MGR_ShadowThickness(w);
    *needed_height = max_height + RC_MarginH(w) + MGR_ShadowThickness(w);

    XdbDebug(__FILE__, w, "PreferredSizeVT => %d %d\n", *needed_width, *needed_height);
}

static void
prefered_size_VC(Widget w,
		 Dimension *needed_width,
		 Dimension *needed_height)
{
    int i;
    int number_per_column;
    Dimension height_of_widgets = 0;
    Dimension *width_of_column;

    if (RC_NCol(w) == 0)
	/* what the hell.  punt and go to .. */
	prefered_size_VT(w, needed_width, needed_height);

    XdbDebug(__FILE__, w, "prefered_size_VC()\n");

    /*
     * This calculation makes the range of column numbers 1 .. NCOL,
     *      and *not* 0..NCOL-1.
     * Therefore, the for loops in this routine must have that range.
     *
     * Danny 19/3/1996.
     */
#define COLUMN_OF_THIS_WIDGET(i) (int)((float)(i) / number_per_column + 0.5)

    /*
     * We need to round up, but don't count the TearOffControl
     */
    if (RC_TearOffControl(w) && XtIsManaged(RC_TearOffControl(w)))
	number_per_column = (int)(((float)MGR_NumChildren(w) - 1) 
				  / RC_NCol(w) + 0.5);
    else
	number_per_column = (int)((float)MGR_NumChildren(w) 
				  / RC_NCol(w) + 0.5);

    width_of_column = (Dimension *) XtCalloc(RC_NCol(w) + 1, 
					     sizeof(Dimension));

    for (i = 0; i < RC_NCol(w) + 1; i++)
	width_of_column[i] = 0;

    /* first, figure out the width of each column and the height of each row */

    for (i = 0; i < MGR_NumChildren(w); i++) 
    {
	XmRCKidGeometry kid_geometry = &(RC_Boxes(w)[i]);
	Widget kid = kid_geometry->kid;

	/*
	 * Don't do this for TearOffControl children - their width should be 
	 * determined by the menu system, not the other way around.
	 */
	if (!kid || !XtIsManaged(kid) || XmIsTearOffButton(kid))
	    continue;

	if (height_of_widgets < kid_geometry->box.height)
	    height_of_widgets = kid_geometry->box.height;

	if (width_of_column[COLUMN_OF_THIS_WIDGET(i)] 
	    < kid_geometry->box.width) 
        {
	    width_of_column[COLUMN_OF_THIS_WIDGET(i)] 
	        = kid_geometry->box.width;
	}
    }

    /* then we multiply and add :),
     * 
     * as a special case, if we're doing tear off stuff, we can decrement
     * the number_per_column count by one, and explicitly add the height
     * of the tear off button. */

    *needed_width = (2 * (RC_MarginW(w) + MGR_ShadowThickness(w))
		     + RC_Spacing(w) * (RC_NCol(w) - 1));

    if ((RC_Type(w) == XmMENU_PULLDOWN || RC_Type(w) == XmMENU_POPUP)
	&& RC_TearOffModel(w) == XmTEAR_OFF_ENABLED && RC_TearOffControl(w) 
	&& XtIsManaged(RC_TearOffControl(w))) 
    {
	*needed_height = 2 * (RC_MarginH(w) + MGR_ShadowThickness(w))
	    + height_of_widgets * number_per_column
	    + XtHeight(RC_TearOffControl(w))
	    + RC_Spacing(w) * (number_per_column - 1);
    }
    else
	*needed_height = 2 * (RC_MarginH(w) + MGR_ShadowThickness(w))
	    + height_of_widgets * number_per_column
	    + RC_Spacing(w) * (number_per_column - 1);

    for (i = 0; i < RC_NCol(w) + 1; i++) 
	*needed_width += width_of_column[i];

    XdbDebug(__FILE__, w, 
	     "  I really need to be %dx%d...\n", 
	     *needed_width, *needed_height);

    XtFree((XtPointer) width_of_column);
#undef	COLUMN_OF_THIS_WIDGET
}

static void
prefered_size(Widget w,
	      Dimension *needed_width,
	      Dimension *needed_height)
{
    *needed_width = 0;
    *needed_height = 0;

    switch (RC_Type(w)) 
    {
    case XmWORK_AREA:
	switch (RC_Orientation(w)) 
        {
	case XmHORIZONTAL:
	    if (RC_Packing(w) == XmPACK_TIGHT)
		prefered_size_HT(w, needed_width, needed_height);
	    else if (RC_Packing(w) == XmPACK_COLUMN)
		prefered_size_HC(w, needed_width, needed_height);
	    break;
	case XmVERTICAL:
	    if (RC_Packing(w) == XmPACK_TIGHT)
		prefered_size_VT(w, needed_width, needed_height);
	    else if (RC_Packing(w) == XmPACK_COLUMN)
		prefered_size_VC(w, needed_width, needed_height);
	    break;
	}
	break;
    case XmMENU_BAR:
	prefered_size_HT(w, needed_width, needed_height);
	break;
    case XmMENU_PULLDOWN:
    case XmMENU_POPUP:
	prefered_size_VC(w, needed_width, needed_height);
	break;
    case XmMENU_OPTION:
	prefered_size_HT(w, needed_width, needed_height);
	break;
    }
}

/*
 * If CW is non-null, make sure not to resize it.
 *
 * The last parameter returns the RC size, to be used in GeometryManager.
 *	Why ? Because GeometryManager needs to call XmRowColumnLayout in test mode,
 *	and figure out from that whether the geometry change is acceptable. If it is,
 *	then it might need to resize the RC itself as a consequence.
 *	And *that* geometry change is a whole different issue from the one we can do
 *	inside XmRowColumnLayout.
 */
static void
XmRowColumnLayout(Widget rc, int mode, Widget cw, XtWidgetGeometry *cg, XtWidgetGeometry *rcgp)
{
    XtWidgetGeometry	rcg;	/* in case rcgp is NULL */
    int i;

    /* No need to do anything else if we don't have any children */
    if (MGR_NumChildren(rc) == 0)
	return;

    XdbDebug(__FILE__, rc, "XmRowColumnLayout (%s) %s %s\n",
	XdbRcType2String(RC_Type(rc)),
	(mode & Mode_Resize) ? "" : "ParentResize",
	(mode & Mode_Test) ? "TestMode" : "");

    rcg.request_mode = 0;
    rcg.width = rcg.height = 0;

    if (RC_FromResize(rc) || !RC_Boxes(rc)) {
	XtFree((XtPointer) RC_Boxes(rc));
	initialize_boxes(rc);
    }

    switch (RC_Type(rc)) {
    case XmWORK_AREA:
	switch (RC_Orientation(rc)) {
	case XmHORIZONTAL:
	    if (RC_Packing(rc) == XmPACK_TIGHT)
		DoLayoutHT(rc, mode, cw, cg, &rcg);
	    else if (RC_Packing(rc) == XmPACK_COLUMN)
		DoLayoutHC(rc);
	    break;
	case XmVERTICAL:
	    if (RC_Packing(rc) == XmPACK_TIGHT)
		DoLayoutVT(rc, mode, cw, cg, &rcg);
	    else if (RC_Packing(rc) == XmPACK_COLUMN)
		DoLayoutVC(rc, mode, cw, cg, &rcg);
	    break;
	}
	break;
    case XmMENU_BAR:
	DoLayoutHT(rc, mode, cw, cg, &rcg);
	break;
    case XmMENU_PULLDOWN:
    case XmMENU_POPUP:
	DoLayoutVT(rc, mode, cw, cg, &rcg);	/* Menus should be tight */
	break;
    case XmMENU_OPTION:
	DoLayoutHT(rc, mode, cw, cg, &rcg);
	break;
    }

/*
 * Resize ourselves
 */
    if (((mode & Mode_Resize) == 0) && (rcg.width != XtWidth(rc) || rcg.height != XtHeight(rc))) {
	XtGeometryResult	result;
	if (((rcg.request_mode & CWWidth) && (rcg.width != XtWidth(rc)))
	 || ((rcg.request_mode & CWHeight) && (rcg.height != XtHeight(rc)))) {
	    XdbDebug(__FILE__, rc, "Need to resize to");
	    if (rcg.request_mode & CWWidth)
		XdbDebug0(__FILE__, rc, " W %d", rcg.width);
	    if (rcg.request_mode & CWHeight)
		XdbDebug0(__FILE__, rc, " H %d", rcg.height);
	    XdbDebug0(__FILE__, rc, "\n");

	    do {
		result = XtMakeResizeRequest(rc, rcg.width, rcg.height,
				&rcg.width, &rcg.height);
	    } while (result == XtGeometryAlmost);

	    if (result == XtGeometryYes) {
		XtWidth(rc) = rcg.width;
		XtHeight(rc) = rcg.height;
	    }
	}
    }

/* Here we actually lay out our children.  */
    for (i = 0; i < MGR_NumChildren(rc); i++) {
	XmRCKidGeometry kid_geometry = &(RC_Boxes(rc)[i]);

	if (kid_geometry->kid == cw)
		continue;

	_XmConfigureObject(kid_geometry->kid,
			   kid_geometry->box.x,
			   kid_geometry->box.y,
			   kid_geometry->box.width,
			   kid_geometry->box.height,
			   XtBorderWidth(kid_geometry->kid));
    }

    if (rcgp)
	*rcgp = rcg;
}

/*
 * This function treats the three following cases :
 *	XmWORK_AREA & XmHORIZONTAL & XmPACK_TIGHT
 *	XmMENU_BAR
 *	XmMENU_OPTION
 * ... which kind of explains what MENU_BAR and MENU_OPTION mean.
 *
 * Finally, it also covers the case where DoLayoutHC is called but no number
 *	of columns is specified.
 *
 * According to the manual page, XmNnumColumns doesn't matter in XmPACK_COLUMN.
 * I guess this means this function shouldn't think about XmNnumColumns.
 */
static void
DoLayoutHT(Widget rc, int mode, Widget cw, XtWidgetGeometry *cg, XtWidgetGeometry *rcg)
{
    int i, j;
    int first_in_row = 0, row;
    Dimension current_x = RC_MarginW(rc) + MGR_ShadowThickness(rc);
    Dimension current_y = RC_MarginH(rc) + MGR_ShadowThickness(rc);
    Dimension max_height = 0;
    XmRCKidGeometry help_kid_geometry;

    if (XdbInDebug(__FILE__, rc)) {
	XdbDebug(__FILE__, rc, "DoLayoutHT(");
	XdbDebug0(__FILE__, rc, "%s", (mode & Mode_Test) ? "Test " : "");
	XdbDebug0(__FILE__, rc, "%s", (mode & Mode_Resize) ? "" : "ParentResize ");
	if (cw) XdbDebug0(__FILE__, rc, "Child %s ", XtName(cw));
	XdbDebug0(__FILE__, rc, ")\n");
    }

/* Calculate the max_height */
    for (i = 0; i < MGR_NumChildren(rc); i++) {
	XmRCKidGeometry kid_geometry = &(RC_Boxes(rc)[i]);
	Widget kid = kid_geometry->kid;

	if (!kid || !XtIsManaged(kid))
	    continue;
	if (kid_geometry->box.height > max_height)
	    max_height = kid_geometry->box.height;
    }
    XdbDebug(__FILE__, rc, "MaxHeight %d\n", max_height);

/* Start layout */
    row = 0;
    for (i = 0; i < MGR_NumChildren(rc); i++) {
	XmRCKidGeometry kid_geometry = &(RC_Boxes(rc)[i]);
	Widget kid = kid_geometry->kid;

	if (!kid || !XtIsManaged(kid))
	    continue;

	/* don't lay the help button out until the end. */
	if (RC_Type(rc) == XmMENU_BAR && kid == RC_HelpPb(rc)) {
	    help_kid_geometry = kid_geometry;
	    continue;
	}

	/* Height treatment */
	kid_geometry->box.height = max_height;

	if (((mode & Mode_Resize) == 0) &&
		current_x + kid_geometry->box.width > XtWidth(rc)) 
        {
	    row++;
	    current_y = row * max_height + RC_Spacing(rc);
	    current_x = RC_MarginW(rc) + MGR_ShadowThickness(rc);
	    first_in_row = i;
	}

	XdbDebug2(__FILE__, rc, kid, "Placing child at %d %d\n", current_x, current_y);

	kid_geometry->box.x = current_x;
	kid_geometry->box.y = current_y;

	current_x += RC_Spacing(rc) + kid_geometry->box.width;
    }

    /* now we fill in the last row, so it takes up the remaining height, 
       if XmNadjustLast is True. */
    if (RC_AdjLast(rc) && RC_FromResize(rc)) {
	for (j = first_in_row; j < MGR_NumChildren(rc); j++) 
        {
	    XmRCKidGeometry kid_geometry = &(RC_Boxes(rc)[j]);
	    Widget kid = kid_geometry->kid;

	    if (!kid || !XtIsManaged(kid))
		continue;

#if 0
	    /* this original code is a problem if XtHeight(rc) is zero */
	    kid_geometry->box.height = XtHeight(rc) - (current_y
						      + RC_MarginH(rc)
						  + MGR_ShadowThickness(rc));
#else
#ifndef MOSAIC_27B4_HACK
	    if ((XtHeight(rc) <= current_y + RC_MarginH(rc))
		|| (max_height - current_y 
		    > XtHeight(rc) - (current_y + RC_MarginH(rc)))) {
		if (current_y + max_height 
		    > XtHeight(rc) - (current_y + RC_MarginH(rc)))
		    kid_geometry->box.height = (XtHeight(rc) 
						- (current_y 
						   + RC_MarginH(rc) 
						   + MGR_ShadowThickness(rc)));
		else
		    kid_geometry->box.height = max_height;
	    }
#else
	    if (XtHeight(rc) < (current_y + RC_MarginH(rc)));
#endif
	    else
		kid_geometry->box.height = (XtHeight(rc) 
					    - (current_y 
					       + RC_MarginH(rc) 
					       + MGR_ShadowThickness(rc)));
#endif
	}
    }

    if (RC_Type(rc) == XmMENU_BAR && RC_HelpPb(rc)) 
    {
#if 0
	/* first we find the help child in the boxes array, 
	   and then set it's position */
	XmRCKidGeometry help_kid_geometry = 
	  &(RC_Boxes(rc)[RC_constraint_IndexPosition(RC_HelpPb(rc))]);
#endif
	help_kid_geometry->box.x = (XtWidth(rc)
				    - MGR_ShadowThickness(rc)
				    - help_kid_geometry->box.width);
	help_kid_geometry->box.y = current_y;

	XdbDebug2(__FILE__, rc, RC_HelpPb(rc), "Set position to %d %d\n",
		help_kid_geometry->box.x, help_kid_geometry->box.y);
    }

    RC_SetFromResize(rc, 0);
}

static void
DoLayoutHC(Widget w)
{
    int i;
    int number_per_row;
    Dimension current_x = RC_MarginW(w) + MGR_ShadowThickness(w);
    Dimension current_y = RC_MarginH(w) + MGR_ShadowThickness(w);
    Dimension width_of_widgets = 0;
    Dimension *height_of_row;
    XmRCKidGeometry kid_geometry;
    Widget kid;

    XdbDebug(__FILE__, w, "DoLayoutHC()\n");

#define ROW_OF_THIS_WIDGET(i) (int)((float)(i) / number_per_row + 0.5)

    if (RC_NCol(w) == 0)
	/* what the hell.  punt and go to .. */
	DoLayoutHT(w, Mode_Resize /* FIX ME */, NULL, NULL, NULL);

    /* we need to round up */
    number_per_row = (int)((float)MGR_NumChildren(w) / RC_NCol(w) + 0.5);

    height_of_row = (Dimension *) XtCalloc(RC_NCol(w), sizeof(Dimension));

    for (i = 0; i < RC_NCol(w); i++)
	height_of_row[i] = 0;

    /* first, figure out the width of each column and the height of each row */
    for (i = 0; i < MGR_NumChildren(w); i++) 
    {
	kid_geometry = &(RC_Boxes(w)[i]);
	kid = kid_geometry->kid;

	if (!kid || !XtIsManaged(kid))
	    continue;

	if (width_of_widgets < kid_geometry->box.width)
	    width_of_widgets = kid_geometry->box.width;

	if (height_of_row[ROW_OF_THIS_WIDGET(i)] < kid_geometry->box.height)
	    height_of_row[ROW_OF_THIS_WIDGET(i)] = kid_geometry->box.height;
    }

    /* now we lay out the children */

    for (i = 0; i < MGR_NumChildren(w); i++) 
    {
	if (i != 0 && (i % number_per_row) == 0) 
        {
	    current_x = RC_MarginW(w) + MGR_ShadowThickness(w);
	    current_y += height_of_row[ROW_OF_THIS_WIDGET(i)] + RC_Spacing(w);
	}

	kid_geometry = &(RC_Boxes(w)[i]);
	kid = kid_geometry->kid;

	if (!kid || !XtIsManaged(kid))
	    continue;

	kid_geometry->box.x = current_x;
	kid_geometry->box.y = current_y;

	kid_geometry->box.width = width_of_widgets;
	kid_geometry->box.height = height_of_row[ROW_OF_THIS_WIDGET(i)];

	current_x += RC_Spacing(w) + width_of_widgets;
    }

    /* now we fill in the last row, so it takes up the remaining height, 
       if XmNadjustLast is True. */
    if (RC_AdjLast(w))
	for (i = MGR_NumChildren(w) - number_per_row; 
	     i < MGR_NumChildren(w); 
	     i++) 
        {
	    kid_geometry = &(RC_Boxes(w)[i]);
	    kid = kid_geometry->kid;

	    if (!kid || !XtIsManaged(kid))
		continue;

	    kid_geometry->box.height = (XtHeight(w) 
					- (current_y 
					   + RC_MarginH(w) 
					   + MGR_ShadowThickness(w)));
	}

#undef ROW_OF_THIS_WIDGET
}

/*
 * Called for cases :
 *	XmVERTICAL & XmPACK_TIGHT
 *	XmMENU_PULLDOWN
 *
 * According to the manual pages, XmNnumColumns should not make a difference in this case.
 */
static void
DoLayoutVT(Widget rc, int mode, Widget cw, XtWidgetGeometry *cg, XtWidgetGeometry *rcg)
{
    int i, j, col;
    int first_in_column = 0;
    Dimension current_x = RC_MarginW(rc) + MGR_ShadowThickness(rc);
    Dimension current_y = RC_MarginH(rc) + MGR_ShadowThickness(rc);
    Dimension max_width = 0;
    XtGeometryResult	result;

    XdbDebug(__FILE__, rc, "DoLayoutVT()\n");

/* Calculate the max_width */
    for (i = 0; i < MGR_NumChildren(rc); i++) {
	XmRCKidGeometry kid_geometry = &(RC_Boxes(rc)[i]);
	Widget kid = kid_geometry->kid;

	if (!kid || !XtIsManaged(kid))
	    continue;
	if (kid_geometry->box.width > max_width)
	    max_width = kid_geometry->box.width;
	XdbDebug2(__FILE__, rc, kid, "Child size %d %d\n",
	    kid_geometry->box.width, kid_geometry->box.height);

    }
    XdbDebug(__FILE__, rc, "MaxWidth %d\n", max_width);

/* After figuring out how much space we need, try to get it */
    if ((mode & Mode_Resize) == 0) {	/* We can try to resize ourselves */
	Dimension	ww, hh;
	hh = 0;
	for (i = 0; i < MGR_NumChildren(rc); i++) {
	    XmRCKidGeometry kid_geometry = &(RC_Boxes(rc)[i]);
	    Widget kid = kid_geometry->kid;

	    if (!kid || !XtIsManaged(kid))
		continue;

	    hh += RC_Spacing(rc) + kid_geometry->box.height;
	}
	if (hh)
	    hh -= RC_Spacing(rc);

	hh += 2 * RC_MarginH(rc) + 2 * MGR_ShadowThickness(rc);
	ww = max_width + 2 * RC_MarginW(rc) + 2 * MGR_ShadowThickness(rc);

	XdbDebug(__FILE__, rc, "Requesting geometry %d %d\n", ww, hh);

	do {
	    result = XtMakeResizeRequest(rc, ww, hh, &ww, &hh);
	} while (result == XtGeometryAlmost);

	XdbDebug(__FILE__, rc, "==> got geometry %d %d (%s)\n",
	    ww, hh, XdbGeometryResult2String(result));
	if (result == XtGeometryYes) {
	    XtWidth(rc) = ww;
	    XtHeight(rc) = hh;
	}
    }

/* Start layout */
    col = 0;
    for (i = 0; i < MGR_NumChildren(rc); i++) {
	XmRCKidGeometry kid_geometry = &(RC_Boxes(rc)[i]);
	Widget kid = kid_geometry->kid;

	if (!kid || !XtIsManaged(kid))
	    continue;

	kid_geometry->box.width = max_width;

	if (current_y + kid_geometry->box.height > XtHeight(rc)) 
        {
	    col++;
	    current_x = col * max_width + RC_Spacing(rc);
	    current_y = RC_MarginH(rc) + MGR_ShadowThickness(rc);
	    first_in_column = i;
	}

	kid_geometry->box.x = current_x;
	kid_geometry->box.y = current_y;

	current_y += RC_Spacing(rc) + kid_geometry->box.height;
#if 0	/* This cannot happen */
	if (kid_geometry->box.width > max_width)
	    max_width = kid_geometry->box.width;
#endif
    }

    /* now we fill in the last column, so it takes up the remaining width, 
       if XmNadjustLast is True. */
    if (RC_AdjLast(rc)) 
    {
	XdbDebug(__FILE__, rc, "DoLayoutVT -- adjusting last column\n");
	for (j = first_in_column; j < MGR_NumChildren(rc); j++) 
        {
	    XmRCKidGeometry kid_geometry = &(RC_Boxes(rc)[j]);
	    Widget kid = kid_geometry->kid;

	    if (!kid || !XtIsManaged(kid))
		continue;

#ifndef MOSAIC_27B4_HACK
	    if ((XtWidth(rc) <= current_x + RC_MarginW(rc))
		|| (max_width - current_x > 
		    XtWidth(rc) - (current_x + RC_MarginW(rc)))) 
           {
		if (current_x + max_width 
		    > XtWidth(rc) - (current_x + RC_MarginW(rc)))
		    kid_geometry->box.width = (XtWidth(rc) 
					       - (current_x 
						  + RC_MarginW(rc) 
						  + MGR_ShadowThickness(rc)));
		else
		    kid_geometry->box.width = max_width;
	    }
#else
	    if (XtWidth(rc) < (current_x + RC_MarginW(rc)));
#endif
	    else
		kid_geometry->box.width = (XtWidth(rc) 
					   - (current_x 
					      + RC_MarginW(rc) 
					      + MGR_ShadowThickness(rc)));
	}
    }
}

/*
 * Vertical Column
 *	XmNnumColumns does determine geometry here.
 *	All children (of certain class) get to have the same height.
 *	It looks like separators and tearOffs don't get treated this way in Motif ...
 *
 * Called when :
 *	XmVERTICAL & XmPACK_COLUMN
 *	XmMENU_POPUP
 */
static void
DoLayoutVC(Widget rc, int mode, Widget cw, XtWidgetGeometry *cg, XtWidgetGeometry *rcg)
{
    int i, j;
    int starting;
    int number_per_column;
    Dimension current_x = RC_MarginW(rc) + MGR_ShadowThickness(rc);
    Dimension current_y = RC_MarginH(rc) + MGR_ShadowThickness(rc);
    Dimension height_of_widgets = 0, toch = 0;
    Dimension *width_of_column;
    Widget kid;
    XmRCKidGeometry kid_geometry;
    int toc;			/* 1 if there is a TearOffControl */

    XdbDebug(__FILE__, rc, "DoLayoutVC()\n");

    toc = XmIsTearOffButton(MGR_Children(rc)[0]) ? 1 : 0;

#define	COLUMN_OF_THIS_WIDGET(i)	((i - toc) / number_per_column)

    if (RC_NCol(rc) == 0)
	/* what the hell.  punt and go to .. */
	DoLayoutVT(rc, Mode_Normal /* FIX ME */, NULL, NULL, NULL);

    /* we need to round up */
    number_per_column = (int)((float)MGR_NumChildren(rc) / RC_NCol(rc) + 0.5);

    width_of_column = (Dimension *) XtCalloc(RC_NCol(rc), sizeof(Dimension));

    for (i = 0; i < RC_NCol(rc); i++)
	width_of_column[i] = 0;

    /* first, figure out the height of each row and the width of each column */
    for (i = 0; i < MGR_NumChildren(rc); i++) 
    {
	kid_geometry = &(RC_Boxes(rc)[i]);
	kid = kid_geometry->kid;

	if (!kid || !XtIsManaged(kid) || XmIsTearOffButton(kid))
	    continue;

	if (height_of_widgets < kid_geometry->box.height)
	    height_of_widgets = kid_geometry->box.height;

	if (width_of_column[COLUMN_OF_THIS_WIDGET(i)] 
	    < kid_geometry->box.width)
	  width_of_column[COLUMN_OF_THIS_WIDGET(i)] = kid_geometry->box.width;
    }

    /*
     * Now we lay out the children
     * First, we do our tear off control, if we have one
     *
     * (That's because the TearOffControl should span all columns)
     */
    if ((RC_Type(rc) == XmMENU_POPUP || RC_Type(rc) == XmMENU_PULLDOWN)
	&& RC_TearOffModel(rc) == XmTEAR_OFF_ENABLED && RC_TearOffControl(rc) 
	&& XtIsManaged(RC_TearOffControl(rc))) {
	kid_geometry = &(RC_Boxes(rc)[0]);	/* the tear off control is here */
	kid = kid_geometry->kid;

	kid_geometry->box.x = current_x;
	kid_geometry->box.y = current_y;
#if 0
	/*
	 * Huh ?? Is this right in the multi-column case ??
	 */
	kid_geometry->box.width = width_of_column[COLUMN_OF_THIS_WIDGET(i)];
#else
	kid_geometry->box.width = XtWidth(rc) - RC_MarginW(rc) - MGR_ShadowThickness(rc);
#endif
	toch = kid_geometry->box.height = XtHeight(kid);

	current_y += RC_Spacing(rc) + kid_geometry->box.height;
	starting = 1;
    }
    else
	starting = 0;

    j = 0;			/* Count the #children already had in this column */
    for (i = starting; i < MGR_NumChildren(rc); i++) {

#if 0
	if (i != 0 && (i % number_per_column) == 0)
#endif
	    if (j == number_per_column) {
		j = 0;
		/*
		 * Current_y should be incremented with the height of TearOffControl if
		 * there is one.
		 */
		current_y = RC_MarginH(rc) + MGR_ShadowThickness(rc);
		if (starting)	/* There is a TearOffControl */
		    current_y += toch;
		current_x += width_of_column[COLUMN_OF_THIS_WIDGET(i)] + RC_Spacing(rc);
	    }

	kid_geometry = &(RC_Boxes(rc)[i]);
	kid = kid_geometry->kid;

	if (!kid || !XtIsManaged(kid) || XmIsTearOffButton(kid))
	    continue;

	kid_geometry->box.x = current_x;
	kid_geometry->box.y = current_y;

	kid_geometry->box.width = width_of_column[COLUMN_OF_THIS_WIDGET(i)];
	kid_geometry->box.height = height_of_widgets;

	current_y += RC_Spacing(rc) + height_of_widgets;

	j++;
    }

    /*
     * Now we fill in the last column 
     * so it takes up the remaining width if XmNadjustLast is True.
     *
     * (The elements in the last column were also processed above, hence no
     *      x, y, height corrections here.)
     */
    if (RC_AdjLast(rc))
	for (j = MGR_NumChildren(rc) - number_per_column;
	     j < MGR_NumChildren(rc);
	     j++) {
	    kid_geometry = &(RC_Boxes(rc)[j]);
	    kid = kid_geometry->kid;

	    if (!kid || !XtIsManaged(kid))
		continue;

	    if (current_x + RC_MarginW(rc) + MGR_ShadowThickness(rc) < XtWidth(rc))
		kid_geometry->box.width = (XtWidth(rc)
					   - current_x
					   - RC_MarginW(rc)
					   - MGR_ShadowThickness(rc));
#if	0
	    XdbDebug2(__FILE__, rc, kid, "adjusting child to width %d\n",
		      kid_geometry->box.width);
#endif
	}
}

static void
resize(Widget w)
{
    RC_SetFromResize(w, 1);

/*
 * Mode_Normal means don't resize the RC,
 * and we're not just testing: do change the children
 */
    XmRowColumnLayout(w, Mode_Normal, NULL, NULL, NULL);
}

/*
 * FIX ME
 *
 * This sucker badly needs work.
 *
 * Note (as indicated in Form.c) that we are called from the widget "w" here,
 * and that the processing of RcGeometryManager should NEVER result in geometry
 * changes to that widget.
 */
static XtGeometryResult
RcGeometryManager(Widget w,
		 XtWidgetGeometry * request,
		 XtWidgetGeometry * reply)
{
    Widget		rc = XtParent(w);
    XtWidgetGeometry	geo;
    XtGeometryResult	result;

    if (XmIsMenuShell(XtParent(rc)))
	/* we're the direct child of a menu shell, so allow any resizing -- FIX ME */
    {
	Widget msw = XtParent(rc);

	XdbDebug2(__FILE__, rc, w, "RC in MenuShell GeometryManager - REQ %d %d RC %d %d Child %d %d\n",
		request->width, request->height,
		XtWidth(rc), XtHeight(rc),
		XtWidth(w), XtHeight(w));

	reply->width = request->width;
	reply->height = request->height;

#if 0
/* This simply cannot be right */
	XtWidth(rc) += (request->width - XtWidth(w));
	XtHeight(rc) += (request->height - XtHeight(w));

	XtWidth(msw) += (request->width - XtWidth(w));
	XtHeight(msw) += (request->height - XtHeight(w));
#else
/*
 * What about borders/margins ???
 * FIX ME
 */
	XtWidth(rc) = request->width;
	XtHeight(rc) = request->height;

	XtWidth(msw) = request->width;
	XtHeight(msw) = request->height;
#endif

	XdbDebug2(__FILE__, rc, w, "RC GeometryManager => YES, RC %d %d MS %d %d\n",
		XtWidth(rc), XtHeight(rc),
		XtWidth(msw), XtHeight(msw));

	return XtGeometryYes;
    }

/* We're not a menu */
#if 0
    XdbDebug2(__FILE__, rc, w, "RC GeometryManager: Preferred_Size\n");
    prefered_size(rc, &geo.width, &geo.height);
#else
    reply->request_mode = CWWidth | CWHeight;
    geo.width = XtWidth(rc);
    geo.height = XtHeight(rc);
#endif

    if (request->width < geo.width || request->height < geo.height) {
	Dimension	ww, hh;

	XdbDebug2(__FILE__, rc, w, "RC GeometryManager: error REQ %d %d < PREF %d %d\n",
		request->width, request->height, geo.width, geo.height);
	if (request->width < geo.width)
	     ww = geo.width;
	else
	     ww = request->width;

	if (request->height < geo.height)
	     hh = geo.height;
	else
	     hh = request->height;
#if 1
	do {
	    XdbDebug2(__FILE__, rc, w, "RC GeometryManager: XtMakeResizeRequest(%d, %d)\n", ww, hh);

	    result = XtMakeResizeRequest(rc, ww, hh, &ww, &hh);

	    XdbDebug2(__FILE__, rc, w,
			"RC GeometryManager: XtMakeResizeRequest => %s (%d, %d)\n",
			XdbGeometryResult2String(result), ww, hh);

	} while (result == XtGeometryAlmost);

	if (result == XtGeometryNo)
		return XtGeometryNo;

	if (result == XtGeometryYes) {
		XtWidth(rc) = ww;
		XtHeight(rc) = hh;
	}
#endif
    }

#if 1
    XdbDebug2(__FILE__, rc, w, "RC GeometryManager: XmRowColumnLayout\n");
    RC_SetFromResize(w, 0);
    XmRowColumnLayout(rc, Mode_Resize, w, request, NULL);
#endif

    reply->width = request->width;
    reply->height = request->height;
    reply->request_mode = request->request_mode;
    return _XmGMReplyToQueryGeometry(w, request, reply);
}

static void
change_managed(Widget w)
{
    Dimension my_width, my_height;
    XtGeometryResult result;

    XdbDebug(__FILE__, w, "change_managed()\n");

    XtFree((XtPointer) RC_Boxes(w));
    initialize_boxes(w);

    prefered_size(w, &my_width, &my_height);

#if 0
    if (!XtIsRealized(w))
    {
	XtWidth(w) = my_width;
	XtHeight(w) = my_height;

	if (XmIsMenuShell(XtParent(w))) {
	    XtWidth(XtParent(w)) = my_width;
	    XtHeight(XtParent(w)) = my_height;
	}
    }
#else
    /* this is a total and complete HACK!  FIXME */
    /* I hacked it further. Is it still a hack now ? (Danny 31/7/1996)
       Every widget in LessTif currently does XtGeometryYes, not XtGeometryDone
       style of geometry management. Therefore after XtMakeResizeRequest you
       still need to do something.
    */

    do {
	result = XtMakeResizeRequest(w,
				     my_width, my_height,
				     &my_width, &my_height);
    } while (result == XtGeometryAlmost);

    if (result == XtGeometryYes) {
	XdbDebug(__FILE__, w, "ChangeManaged: setting size to %d %d\n", my_width, my_height);

	XtWidth(w) = my_width;
	XtHeight(w) = my_height;
    }
#endif

    XmRowColumnLayout(w, Mode_Normal, NULL, NULL, NULL);

    _XmNavigChangeManaged(w);
}

static void
_XmRadioCallback(Widget w, 
                 XtPointer cd,
                 XtPointer cs)
{
    Widget tgl;
    Widget rc = (Widget) cd;
    XmToggleButtonCallbackStruct *cbs = (XmToggleButtonCallbackStruct *) cs;
    int i;

    for (i = 0; i < MGR_NumChildren(rc); i++) {
	tgl = MGR_Children(rc)[i];
	if (RC_RadioBehavior(rc) || RC_RadioAlwaysOne(rc)) {
	    if (XmIsToggleButton(tgl) && XmToggleButtonGetState(tgl)) {
		XdbDebug(__FILE__, w, "is toggle and set\n");
               if (w != tgl && cbs && cbs->set) {
                   XmToggleButtonSetState(tgl, False, True);
		   XdbDebug(__FILE__, w, "clearing toggle %d\n", i);
            }
            else if (w == tgl && cbs && cbs->set)
                XmToggleButtonSetState(tgl, True, False);
            }
            if (XmIsToggleButtonGadget(tgl) && XmToggleButtonGadgetGetState(w))
            { 
                if (w != tgl && cbs && cbs->set)
                    XmToggleButtonGadgetSetState(tgl, False, True);
            }
        }
    } 
}

static void
_XmEntryCallback(Widget w, 
                 XtPointer cd,
                 XtPointer cs)
{
    XmRowColumnCallbackStruct rcb;
    XmToggleButtonCallbackStruct *cbs = (XmToggleButtonCallbackStruct *)cs;
    Widget rc = (Widget)cd;

#if 0
    if (XmIsToggleButton(w) || XmIsToggleButtonGadget(w)) {
	if (cbs->set == 0)
	    return;
    }
#endif
    rcb.reason = XmCR_ACTIVATE;
    rcb.event = cbs->event;
    rcb.widget = w;
    rcb.data = (char *)cd;
    rcb.callbackstruct = (char *)cs;
    XtCallCallbackList(rc, RC_Entry_cb(rc), (XtPointer)&rcb);
}

static Cardinal 
_XmRowColumnOrderProc(Widget widget)
{
  if (RC_constraint_IndexPosition(widget) == XmLAST_POSITION
      || RC_constraint_IndexPosition(widget) >= MGR_NumChildren(XtParent(widget)))
  {
      return MGR_NumChildren(XtParent(widget));
  }
  else 
  {
      return RC_constraint_IndexPosition(widget);
  }
}

static void
insert_child(Widget w)
{
    Widget rc = (Widget) XtParent(w);
    Arg args[1];

/*
  This may be correct, but it keeps me from adding CascadeButtonGadgets
  to a menu bar...   Chris

    if (rc->row_column.is_homogeneous
        && (rc->row_column.entry_class != XtClass(w))) 
    {
       _XmWarning(w, "Child of homogeneous rowcolumn is of wrong class!");
       return;
    }
*/
    if ((RC_RadioBehavior(rc) || RC_RadioAlwaysOne(rc)) &&
	((XtClass(w) == xmToggleButtonWidgetClass)
	 || (XtClass(w) == xmToggleButtonGadgetClass)))
	XtAddCallback(w, XmNvalueChangedCallback, _XmRadioCallback, rc);

#define superclass (&xmManagerClassRec)
    (*superclass->composite_class.insert_child) (w);
#undef superclass

    /*
     * if it's a label subclass, set its alignment to our
     * RC_EntryAlignment resource
     */
    XtSetArg(args[0], XmNalignment, RC_EntryAlignment(rc));
    if (XmIsLabel(w) 
	|| XmIsLabelGadget(w))
      XtSetValues(w, args, 1);


    if (RC_Entry_cb(rc)) 
    {
	if (XmIsDrawnButton(w)) 
        {
	    XtRemoveCallbacks(w, XmNactivateCallback, DB_ActivateCallback(w));
	    XtAddCallback(w, XmNactivateCallback, _XmEntryCallback, rc);
	}
	else if (XmIsCascadeButton(w)) 
        {
	    XtRemoveCallbacks(w, XmNactivateCallback, CB_ActivateCall(w));
	    XtAddCallback(w, XmNactivateCallback, _XmEntryCallback, rc);
	}
	else if (XmIsCascadeButtonGadget(w)) 
        {
	    XtRemoveCallbacks(w, XmNactivateCallback, CBG_ActivateCall(w));
	    XtAddCallback(w, XmNactivateCallback, _XmEntryCallback, rc);
	}
	else if (XmIsPushButton(w)) 
        {
	    XtRemoveCallbacks(w, XmNactivateCallback, PB_ActivateCallback(w));
	    XtAddCallback(w, XmNactivateCallback, _XmEntryCallback, rc);
	}
	else if (XmIsPushButtonGadget(w)) 
        {
	    XtRemoveCallbacks(w, XmNactivateCallback, PBG_ActivateCallback(w));
	    XtAddCallback(w, XmNactivateCallback, _XmEntryCallback, rc);
	}
	else if (XmIsToggleButton(w)) 
        {
	    XtRemoveCallbacks(w, XmNvalueChangedCallback, TB_ValueChangedCallback(w));
	    XtAddCallback(w, XmNvalueChangedCallback, _XmEntryCallback, rc);
	}
	else if (XmIsToggleButtonGadget(w)) 
        {
	    XtRemoveCallbacks(w, XmNvalueChangedCallback, TBG_ValueChangedCallback(w));
	    XtAddCallback(w, XmNvalueChangedCallback, _XmEntryCallback, rc);
	}
    }

    /* This broke something when it was done at the beginning of this func */
    XtFree((XtPointer) RC_Boxes(rc));
    initialize_boxes(rc);
}

/* Determine if widget w is a top level menu pane/bar, and that the given event
 * (assumed to be a button event !) occured outside widget w's child widgets */
static Boolean
ExternalBtnEvent(Widget w,
		 XEvent * event)
{
    Widget event_widget;
    Boolean is_child = False;

    /* Determine the top-level widget for 'w' */
    Widget w_top_level = RC_LastSelectToplevel(w);

    XdbDebug(__FILE__, w, 
	     "  Determining if the button event was external to the menu system\n");

    /* Determine if w is a top level menu pane/bar */
    if ((w_top_level == w) || !w_top_level) {
	/* Determine the event window */
	Window event_window = (event->xbutton.subwindow) ?
	event->xbutton.subwindow : event->xbutton.window;

	/* Convert window to a widget ID */
	event_widget = XtWindowToWidget(XtDisplay(w), event_window);

	/* Now determine if the event widget is a child (direct of indirect)
	 * of widget w. */
	if (event_widget && (event_widget != w)) {
	    Widget parent = XtParent(event_widget);

	    /* Search event widgets parents, to see if 'w' is an ancestor */
	    while (parent && (parent != w))
		parent = XtParent(parent);

	    is_child = (parent == w);
	}			/* if event widget is not 'w' */
    }				/* If event widget is a top level menu pane/bar */

    return (!is_child);
}

/* action routines */

static void
MenuBtnDown(Widget w,
	    XEvent * event,
	    String * params,
	    Cardinal * num_params)
{
    Widget gadget;

    /* If event was in a child gadget, then arm it, otherwise determine
     * whether the event occured outside the menu system, and do a cleanup
     * if it did. */
    gadget = (Widget) _XmInputInGadget(w,
				       event->xkey.x,
				       event->xkey.y);

    if (gadget) {
	/* Event was in a gadget */
	XdbDebug2(__FILE__, w, gadget, "MenuBtnDown() in gadget\n");

	MGR_HighlightedWidget(w) = gadget;

	/* turn on mouse traversal */
	_XmSetInDragMode(w, True);

	/* Now we arm the widget */
	_XmGadgetArm(w, event, params, num_params);
    }
    else if (ExternalBtnEvent(w, event)) {
	XdbDebug(__FILE__, w, "MenuBtnDown(event external to menu system)\n");
	/*
	 * Event was external to the menu system;
	 * Do a cleanup, and replay the event to the outside world.
	 */
	DoBtnEventCleanupReplay(w, event, params, num_params);
    }
    else {
	XdbDebug(__FILE__, w, "MenuBtnDown(not in gadget)\n");
    }
}

static void
DoBtnEventCleanupReplay(Widget w,
			XEvent *event,
			String *params,
			Cardinal *num_params)
{
    Widget menu_shell = NULL;

    XdbDebug(__FILE__, w, "DoBtnEventCleanupReplay\n");

    if (RC_PopupPosted(w))
      menu_shell = XtParent(RC_PopupPosted(w));

    if (menu_shell == NULL) {	/* Think this is always happening - Danny 9/5/1996 */
	Widget mb = w;

	while (mb && !XmIsMenuBar(mb))
	    mb = XtParent(mb);

	if (RC_PopupPosted(mb))
	  menu_shell = XtParent(RC_PopupPosted(mb));

	XdbDebug(__FILE__, w, "DoBtnEventCleanupReplay: menu bar is '%s'\n",
		 mb ? XtName(mb) : "(null)");

	/* disarm the menu bar */
	RC_SetArmed(mb, 1);
    }

    if (menu_shell) {

	if (menu_shell) {
	    XdbDebug(__FILE__, w, 
		     "DoBtnEventCleanupReplay: calling MenuShellPopdownDone\n");
	    XtCallActionProc(menu_shell, "MenuShellPopdownDone",
			     event, params, *num_params);
	}

	/* Replay the event */
	XdbDebug(__FILE__, w, "DoBtnEventCleanupReplay: XAllowEvents (replay)\n");
	XAllowEvents(XtDisplay(w), ReplayPointer, CurrentTime);

	_XmMenuFocus(w, XmMENU_FOCUS_RESTORE, CurrentTime);

	_XmUngrabKeyboard(w, CurrentTime);

	_XmUngrabPointer(w, CurrentTime);

	_XmSetInDragMode(w, False);

    }
    else {
	XdbDebug(__FILE__, w, "DoBtnEventCleanupReplay: RC_PopupPosted was False\n");
    }
}

static void
MenuBtnUp(Widget w,
	  XEvent * event,
	  String * params,
	  Cardinal * num_params)
{
    Widget gadget;

    /*
     * If event was in a child gadget, then arm it, otherwise determine
     * whether the event occured outside the menu system, and do a cleanup if it did.
     */
    gadget = (Widget) _XmInputInGadget(w,
				       event->xkey.x,
				       event->xkey.y);

    if (gadget) {

	XdbDebug2(__FILE__, w, gadget, "MenuBtnUp(in gadget)\n");

	/* Now we activate the gadget */
	_XmGadgetActivate(w, event, params, num_params);
    }
    else if (ExternalBtnEvent(w, event)) {

	XdbDebug(__FILE__, w, "MenuBtnUp(event external to menu system)\n");

	/*
	 * Event was external to the menu system;
	 * Do a cleanup, and replay the event to the outside world.
	 */
	DoBtnEventCleanupReplay(w, event, params, num_params);
    }
    else {
	XdbDebug(__FILE__, w, "MenuBtnUp(not in gadget)\n");
    }
}

static void
MenuEnter(Widget w,
	  XEvent * event,
	  String * params,
	  Cardinal * num_params)
{
    XdbDebug(__FILE__, w, "MenuEnter()\n");
}

static void
MenuUnmap(Widget w,
	  XEvent * event,
	  String * params,
	  Cardinal * num_params)
{
    XdbDebug(__FILE__, w, "MenuUnmap()\n");
    _XmCallRowColumnUnmapCallback(w, event);
}

static void
MenuFocusIn(Widget w,
	    XEvent * event,
	    String * params,
	    Cardinal * num_params)
{
    XdbDebug(__FILE__, w, "MenuFocusIn\n");
}

static void
MenuFocusOut(Widget w,
	     XEvent * event,
	     String * params,
	     Cardinal * num_params)
{
    XdbDebug(__FILE__, w, "MenuFocusOut\n");
    /* restore the focus to the widget that had it before moving to the menu bar */
}

static void
ManagerGadgetSelect(Widget w,
		    XEvent * event,
		    String * params,
		    Cardinal * num_params)
{
    XdbDebug(__FILE__, w, "ManagerGadgetSelect()\n");
}

static void
MenuGadgetEscape(Widget w,
		 XEvent * event,
		 String * params,
		 Cardinal * num_params)
{
    XdbDebug(__FILE__, w, "MenuGadgetEscape()\n");
}

static void
MenuBarEnter(Widget w,
	     XEvent * event,
	     String * params,
	     Cardinal * num_params)
{
    XdbDebug(__FILE__, w, "MenuBarEnter()\n");
}

static void
MenuBarLeave(Widget w,
	     XEvent * event,
	     String * params,
	     Cardinal * num_params)
{
    XdbDebug(__FILE__, w, "MenuBarLeave()\n");
}

static void
MenuBarGadgetSelect(Widget w,
		    XEvent * event,
		    String * params,
		    Cardinal * num_params)
{
    XdbDebug(__FILE__, w, "MenuBarGadgetSelect()\n");
}

/* convenience functions */

Widget
XmCreateMenuBar(Widget parent,
		char *name,
		Arg * arglist,
		Cardinal argcount)
{
    /* menu bar's have the their rowColumnType set to XmMENU_BAR and
       are homogeneous (they only accept CascadeButtons, and 
       CascadeButtonGadgets */

    Widget w;
    Arg myArgList[4];
    int n = 0;

    ArgList combined;

    XtSetArg(myArgList[n], XmNrowColumnType, XmMENU_BAR); n++;
    XtSetArg(myArgList[n], XmNorientation, XmHORIZONTAL); n++;
    XtSetArg(myArgList[n], XmNisHomogeneous, True); n++;
    XtSetArg(myArgList[n], XmNentryClass, xmCascadeButtonWidgetClass); n++;

    combined = XtMergeArgLists(myArgList, n, arglist, argcount);

    w = XtCreateWidget(name,
		       xmRowColumnWidgetClass,
		       parent,
		       combined, n + argcount);

    XtFree((char *)combined);

    return w;
}

Widget
XmCreateOptionMenu(Widget parent,
		   char *name,
		   Arg * arglist,
		   Cardinal argcount)
{
    /* option menus have the their rowColumnType set to XmMENU_OPTION */
    Widget w;
    Arg myArgList[1];
    int n = 0;

    ArgList combined;

    XtSetArg(myArgList[n], XmNrowColumnType, XmMENU_OPTION); n++;

    combined = XtMergeArgLists(myArgList, n, arglist, argcount);

    w = XtCreateWidget(name,
		       xmRowColumnWidgetClass,
		       parent,
		       combined, n + argcount);

    XtFree((char *)combined);

    return w;
}

Widget
XmCreatePopupMenu(Widget parent,
		  char *name,
		  Arg * arglist,
		  Cardinal argcount)
{
    /* popup menus have the their rowColumnType set to XmMENU_POPUP */

    Widget w, ms;
    Arg myArgList[1];
    Arg shell_args[3];
    int shell_ac;
    int n=0;
    char *popup_name = XtMalloc(strlen("popup_") + strlen(name) + 1);

    ArgList combined;

    strcpy(popup_name, "popup_");
    strcat(popup_name, name);


    XtSetArg(myArgList[n], XmNrowColumnType, XmMENU_POPUP); n++;

    combined = XtMergeArgLists(myArgList, n, arglist, argcount);

    shell_ac = 0;
    XtSetArg(shell_args[shell_ac], XmNwidth, 10); shell_ac ++; /* FIX ME */
    XtSetArg(shell_args[shell_ac], XmNheight, 10); shell_ac ++; /* FIX ME */
    XtSetArg(shell_args[shell_ac], XmNallowShellResize, True); shell_ac ++;

    ms = XtCreatePopupShell(popup_name,
			    xmMenuShellWidgetClass,
			    parent, shell_args, shell_ac);

    XtFree(popup_name);

    w = XtCreateWidget(name, 
		       xmRowColumnWidgetClass,
		       ms,
		       combined, n+argcount);


    /* since we're not managed -- we leave that up to the application
       to get us to pop up, we realize ourselves here. */
    /*XtRealizeWidget(w);*/

    XtAddEventHandler(parent, ButtonPressMask, False,
		      _XmPopupButtonPressHandler, (XtPointer) w);

    XtFree((char *)combined);

    return w;
}

Widget
XmCreatePulldownMenu(Widget parent,
		     char *name,
		     Arg * arglist,
		     Cardinal argcount)
{
/* pulldown menus have their rowColumnType set to XmMENU_PULLDOWN */
    Widget w, ms;
    Arg myArgList[5];
    int n = 0;
    char *popup_name = XtMalloc(strlen("popup_") + strlen(name) + 1);
    ArgList combined;
    Arg shell_args[3];
    int shell_ac;

    strcpy(popup_name, "popup_");
    strcat(popup_name, name);

    XtSetArg(myArgList[n], XmNrowColumnType, XmMENU_PULLDOWN); n++;
    XtSetArg(myArgList[n], XmNorientation, XmVERTICAL); n++;
    XtSetArg(myArgList[n], XmNpacking, XmPACK_COLUMN); n++;
    XtSetArg(myArgList[n], XmNnumColumns, 1); n++;

    combined = XtMergeArgLists(myArgList, n, arglist, argcount);

    shell_ac = 0;
    XtSetArg(shell_args[shell_ac], XmNwidth, 10); shell_ac ++; /* FIX ME */
    XtSetArg(shell_args[shell_ac], XmNheight, 10); shell_ac ++; /* FIX ME */
    XtSetArg(shell_args[shell_ac], XmNallowShellResize, True); shell_ac ++;

    ms = XtCreatePopupShell(popup_name,
			    xmMenuShellWidgetClass,
			    parent, shell_args, shell_ac);

    XtFree(popup_name);

    w = XtCreateManagedWidget(name, 
			      xmRowColumnWidgetClass,
			      ms,
			      combined, n + argcount);

    XtFree((char *)combined);

    return w;
}

Widget
XmCreateRadioBox(Widget parent,
		 char *name,
		 Arg * arglist,
		 Cardinal argcount)
{
    /* radio boxes have their rowColumnType set to XmWORK_AREA, and their 
       radioBehavior set to true.  The also are homogeneous and accept 
       ToggleButtons, and ToggleButtonGadgets */

    Widget w;
    Arg myArgList[2];
    int n = 0;

    ArgList combined;

    XtSetArg(myArgList[n], XmNrowColumnType, XmWORK_AREA); n++;
    XtSetArg(myArgList[n], XmNradioBehavior, True); n++;

    combined = XtMergeArgLists(myArgList, n, arglist, argcount);

    w = XtCreateWidget(name,
		       xmRowColumnWidgetClass,
		       parent,
		       combined, n + argcount);

    XtFree((char *)combined);

    return w;
}

Widget
XmCreateRowColumn(Widget parent,
		  char *name,
		  Arg * arglist,
		  Cardinal argcount)
{
    return XtCreateWidget(name,
			  xmRowColumnWidgetClass,
			  parent,
			  arglist,
			  argcount);
}

Widget
XmCreateWorkArea(Widget parent,
		 char *name,
		 Arg * arglist,
		 Cardinal argcount)
{
    Widget w;
    Arg myArgList[2];
    int n = 0;

    ArgList combined;

    XtSetArg(myArgList[n], XmNrowColumnType, XmWORK_AREA);
    n++;

    combined = XtMergeArgLists(myArgList, n, arglist, argcount);

    w = XtCreateWidget(name,
		       xmRowColumnWidgetClass,
		       parent,
		       combined, n + argcount);

    XtFree((char *)combined);

    return w;
}

Widget
XmOptionButtonGadget(Widget option_menu)
{
    return XtNameToWidget(option_menu, RC_OPTION_CBG);
}

Widget
XmOptionLabelGadget(Widget option_menu)
{
    return XtNameToWidget(option_menu, RC_OPTION_LABEL);
}

void
XmMenuPosition(Widget menu,
	       XButtonPressedEvent * event)
{
    if (!XmIsMenuShell(XtParent(menu))
	|| RC_Type(menu) != XmMENU_POPUP)
    {
        _XmWarning(menu, "XmMenuPosition called with a non popup menu.");
	return;
    }

    /* should be all that's needed */
    _XmMoveObject(XtParent(menu),
		  event->x_root,
		  event->y_root);
}

Widget
XmGetTearOffControl(Widget menu)
{
    if (XmIsRowColumn(menu))
	return RC_TearOffControl(menu);
    return NULL;
}

void
XmAddToPostFromList(Widget menu_wid, 
		    Widget widget)
{
  /* well, we could either be adding menu_wid to widget's post
     from list or adding widget to menu_wid's post from list.
     I'll opt for the latter. */
  if (RC_PostFromList(menu_wid) == NULL)
  {
    RC_PostFromListSize(menu_wid) = 5;

    RC_PostFromList(menu_wid) = (Widget*)XtMalloc(sizeof(Widget)
						  * RC_PostFromListSize(menu_wid));
    RC_PostFromCount(menu_wid) = 0;
  }

  RC_PostFromList(menu_wid) [ RC_PostFromCount(menu_wid)++ ] = widget;

  if (RC_PostFromCount(menu_wid) == RC_PostFromListSize(menu_wid))
  {
    RC_PostFromListSize(menu_wid) *= 2;

    RC_PostFromList(menu_wid) = (Widget*)XtRealloc((char*) RC_PostFromList(menu_wid),
						   sizeof(Widget)
						   * RC_PostFromListSize(menu_wid));
  }
}

void
XmRemoveFromPostFromList(Widget menu_wid, 
			 Widget widget)
{
  int i;

  if (RC_PostFromList(menu_wid) == NULL)
    return; /* should we post an error/warning? */

  for (i = 0; i < RC_PostFromCount(menu_wid); i++ )
  {
    if (RC_PostFromList(menu_wid)[i] == widget)
    {
        int j;

	/* found it, now slide everything down. */
        for (j = i ; j < RC_PostFromCount(menu_wid) - 1 ; j ++)
	  RC_PostFromList(menu_wid)[j] = RC_PostFromList(menu_wid)[j + 1];

	RC_PostFromCount(menu_wid) --;

	return;
    }
  }

  /* should getting to this point also be an error? */
}

/*
 * The result of XmGetPostedFromWidget when the callback is activated by
 * an accelerator is not defined in the manual.
 */
Widget
XmGetPostedFromWidget(Widget menu)
{
    if (XmIsRowColumn(menu)) 
    {
	if (RC_LastSelectToplevel(menu))
	    return RC_LastSelectToplevel(menu);
	/* return something sensible */
	return menu;
    }
    return NULL;
}

/*
 * most of these are already written under different names
 */
void
_XmPostPopupMenu(Widget wid,
		 XEvent * event)
{
    if (!XmIsRowColumn(wid)
	|| RC_Type(wid) != XmMENU_POPUP)
    {
      _XmWarning(wid,
		 "_XmPostPopupMenu sent non rowcolumn widget\n");
      return;
    }

    _XmSetInDragMode(wid, True);
    _XmSetInPMMode(wid, True);

    XtPopupSpringLoaded(XtParent(wid));
}

void
_XmSetPopupMenuClick(Widget wid, 
		     Boolean popupMenuClick)
{
}

Boolean
_XmGetPopupMenuClick(Widget wid)
{
    return False;
}

void
_XmAllowAcceleratedInsensitiveUnmanagedMenuItems(Widget wid,
						 Boolean allowed)
{
  XmMenuState state = _XmGetMenuState(wid);

  state->RC_allowAcceleratedInsensitiveUnmanagedMenuItems = allowed;
}

void
_XmSetSwallowEventHandler(Widget widget, 
			  Boolean add_handler)
{
}

void
_XmMenuFocus(Widget w, 
	     int operation, 
	     Time _time)
{
    XmMenuState state = _XmGetMenuState(w);
    Window	window_return;
    int		revert_to_return;
    /* nothing but a guess here.  There are three operations that seem
       to take place with respect to focus and menus.
       
       1 - saving the keyboard focus (done at the start of the menu stuff.)
       2 - setting the keyboard focus (done at several points.)
       3 - restoring the keyboard focus (done at the end of the menu stuff.)
       
       So, those are the three "operations" that are valid to this function.
       */
    switch (operation) 
    {
    case XmMENU_FOCUS_SAVE:
      XGetInputFocus(XtDisplay(w),
		     &state->RC_menuFocus.oldFocus,
		     &state->RC_menuFocus.oldRevert);

      state->RC_menuFocus.oldWidget = XtWindowToWidget(XtDisplay(w),
   				               state->RC_menuFocus.oldFocus);
      break;
    case XmMENU_FOCUS_RESTORE:
	XSetInputFocus(XtDisplay(w),
		       state->RC_menuFocus.oldFocus,
		       state->RC_menuFocus.oldRevert,
		       _time);

	XGetInputFocus(XtDisplay(w),
		       &window_return,
		       &revert_to_return);

	if (state->RC_menuFocus.oldFocus != window_return
	    || state->RC_menuFocus.oldRevert != revert_to_return)
	  XdbDebug(__FILE__, w, "  SetInputFocus call failed.\n");
      break;
    case XmMENU_FOCUS_SET:
      XSetInputFocus(XtDisplay(w),
		     XtWindow(w),
		     RevertToPointerRoot,
		     _time);

      XGetInputFocus(XtDisplay(w),
		     &window_return,
		     &revert_to_return);

      if (window_return != XtWindow(w)
	  || revert_to_return != RevertToPointerRoot)
      {
	XdbDebug(__FILE__, w, "  setting input focus failed\n");
	_XmUngrabKeyboard(w, _time);
	return;
      }
      break;
    }
}

void
_XmGetActiveTopLevelMenu(Widget wid, 
			 Widget * rwid)
{
}

Boolean
_XmMatchBSelectEvent(Widget wid, 
		     XEvent * event)
{
    return False;
}

Boolean
_XmMatchBDragEvent(Widget wid, 
		   XEvent * event)
{
    return False;
}

void
_XmHandleMenuButtonPress(Widget wid, 
			 XEvent * event)
{
}

void
_XmMenuBtnDown(Widget wid, 
	       XEvent * event,
	       String * params, 
	       Cardinal * num_params)
{
}

void
_XmMenuBtnUp(Widget wid, 
	     XEvent * event,
	     String * params, 
	     Cardinal * num_params)
{
}

void
_XmCallRowColumnMapCallback(Widget wid, XEvent * event)
{
    XmRowColumnCallbackStruct cbs;

    cbs.reason = XmCR_MAP;
    cbs.event = event;
    cbs.widget = NULL;
    cbs.data = NULL;
    cbs.callbackstruct = NULL;

    XtCallCallbackList(wid, RC_Map_cb(wid), (XtPointer) & cbs);
}

void
_XmCallRowColumnUnmapCallback(Widget wid, XEvent * event)
{
    XmRowColumnCallbackStruct cbs;

    cbs.reason = XmCR_UNMAP;
    cbs.event = event;
    cbs.widget = NULL;
    cbs.data = NULL;
    cbs.callbackstruct = NULL;

    XtCallCallbackList(wid, RC_Unmap_cb(wid), (XtPointer) & cbs);
}

void
_XmMenuPopDown(Widget w,
	       XEvent * event,
	       Boolean * popped_up)
{
}

Boolean
_XmIsActiveTearOff(Widget w)
{
    return RC_TearOffActive(w);
}

void
_XmMenuHelp(Widget w, 
	    XEvent *event, 
	    String *params, 
	    Cardinal *num_params)
{
    XtCallCallbacks(w,
		    XmNhelpCallback,
		    NULL);
}

/* synthetic resource converters */
static XmImportOperator
_XmToMenuPost(Widget widget,
	      int offset,
	      XtArgVal *value)
{
    String menuPost = (String)*value; /* the string we're checking/parsing. */
    Boolean valid_expression = True; /* we're optimistic :) */
    int numberIndex; /* the index of the button number */
    int button;
    
    XdbDebug(__FILE__, widget, "In _XmToMenuPost(%s)\n", menuPost);
    
    if (menuPost == NULL)
    {
        XdbDebug(__FILE__, widget, "  Default case -- menupost == null\n");
	return XmSYNTHETIC_NONE;
    }

    /* first we check the basic syntax of the expression. 
       for now we only allow things of the type <(Btn|Button)[0-5](Down)?> */

    if (valid_expression)
      if (!strncmp(menuPost, "<Btn", 4))
	numberIndex = 4;
      else if (!strncmp(menuPost, "<Button", 7))
	numberIndex = 7;
      else
      {
  	  XdbDebug(__FILE__, widget, "  Invalid beginning of string.\n");
	  valid_expression = False;
      }

    if (valid_expression)
      if (numberIndex + 1 >= strlen(menuPost))
      {
	  XdbDebug(__FILE__, widget, "  Invalid ending of string.\n");
	  valid_expression = False;
      }
      else if (!strncmp(&(menuPost[numberIndex+1]), "Down>", 5))
	valid_expression = True;
      else if (menuPost[numberIndex+1] == '>')
	valid_expression = True;
      else 
      {
	  XdbDebug(__FILE__, widget, "  Invalid ending of string.\n");
	  valid_expression = False;
      }
  
    if (valid_expression)
    {
        button = menuPost[numberIndex] - '0';
	
	if (button < 1 || button > 5)
        {
	    XdbDebug(__FILE__, widget, "  Invalid button in string.\n");
	    valid_expression = False;
	}
    }

    if (valid_expression)
    {
        XdbDebug(__FILE__, widget, "  Valid expression.  Button is %d\n", button);
	
	*value = (XtArgVal)menuPost;
	/* does this one ever change? */
	RC_PostEventType(widget) = ButtonPressMask;
	RC_PostButton(widget) = button;
	RC_PostModifiers(widget) = 0; /* FIX ME */
	
	return XmSYNTHETIC_LOAD;
    }
    else
        return XmSYNTHETIC_NONE;
}

static void 
_XmFromMenuPost(Widget widget, 
		int offset, 
		XtArgVal *value)
{
    /* This function assumes that the menuPost string is a button event.
       Not sure if this is in general the case.. FIX ME -- Chris */
    String return_value = (String)XtMalloc(sizeof(char)
					   * (strlen("<Btn>")
					      + 2  /* for the <>'s */
					      + 1  /* the the NULL */));

    XdbDebug(__FILE__, widget, "In _XmFromMenuPost()\n");

    sprintf(return_value, "<Btn%d>", RC_PostButton(widget));
    
    *value = (XtArgVal)return_value;
}

static void 
_XmPopupButtonPressHandler(Widget w, 
			   XtPointer client_data,
			   XEvent *event, 
			   Boolean *cont)
{
    Widget rc = (Widget) client_data;
    XmMenuState state = _XmGetMenuState(w);

    /* first we check if the event is valid with respect to our menupost resource*/
    if (event->xany.type != ButtonPress
	|| event->xbutton.button != RC_PostButton(rc)
	/* should also check the modifiers, but since we don't set 
	   them to a useful value... */)
        return;

    /* if we're already waiting to be managed, do nothing. */
    if (state->RC_ButtonEventStatus.waiting_to_be_managed)
        return;

    XdbDebug(__FILE__, rc, "  Filling in ButtonEventStatusRec\n");

    state->RC_ButtonEventStatus.waiting_to_be_managed = True;
    state->RC_ButtonEventStatus.event = event->xbutton;
    state->RC_ButtonEventStatus.time = event->xbutton.time;
    state->RC_ButtonEventStatus.verified = True; /* Is it really? FIX ME */
#if 0
    RC_PopupTimer(rc) = XtAppAddTimeOut(XtWidgetToApplicationContext(rc),
					3000 /* 3 seconds */,
					foo,
					(XtPointer) rc);
#endif
}
