/**
 *
 * $Id: Text.c,v 1.4 1996/08/14 05:47:39 u27113 Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static char rcsid[] = "$Id: Text.c,v 1.4 1996/08/14 05:47:39 u27113 Exp $";

#define	DO_MOSAIC_HACK
#define I18N

#include <LTconfig.h>
#include <XmI/XmI.h>
#include <XmI/DebugUtil.h>

#include <Xm/XmP.h>
#include <Xm/TextP.h>
#include <Xm/ScrolledW.h>
#include <stdio.h>
#include <limits.h>		/* for INT_MAX */
#ifdef HAVE_STRING_H
#include <string.h>
#define ANSI_STRING
#else
#include <strings.h>
#endif
#include <stdlib.h>

#ifdef	DO_MOSAIC_HACK
#include <Xm/TextF.h>
#endif

#ifndef XmUNSPECIFIED
#define XmUNSPECIFIED (~0)
#endif

/* Forward Declarations */

static void ClassInitialize();
static void ClassPartInitialize(WidgetClass class);
static void Initialize(Widget request, Widget new, ArgList args, Cardinal * num_args);
static void Destroy(Widget w);
static void Resize(Widget w);
static void Realize(Widget w, XtValueMask * value_mask, XSetWindowAttributes * attributes);
static void Redisplay(Widget w, XEvent * event, Region region);
static XtGeometryResult QueryGeometry(Widget w, XtWidgetGeometry * proposed, XtWidgetGeometry * answer);
static void _XmTextSetEditable(Widget w, Boolean e);

static Boolean SetValues(Widget current, Widget request, Widget new, ArgList args, Cardinal * num_args);

static void LineTable(XmTextWidget w);


#define Offset(_name) XtOffsetOf(XmTextRec, text._name)

/* Resources for the Text  class */
static XtResource resources[] =
{
  {
    XmNsource, XmCSource, XmRPointer,
    sizeof(XmTextSource), Offset(source),
    XmRPointer, (XtPointer) NULL	/* FIXME: Motif has a wacky value here */
  },
  {
    XmNactivateCallback, XmCCallback, XmRCallback,
    sizeof(XtCallbackList), Offset(activate_callback),
    XmRCallback, (XtPointer) NULL
  },
  {
    XmNfocusCallback, XmCCallback, XmRCallback,
    sizeof(XtCallbackList), Offset(focus_callback),
    XmRCallback, (XtPointer) NULL
  },
  {
    XmNlosingFocusCallback, XmCCallback, XmRCallback,
    sizeof(XtCallbackList), Offset(losing_focus_callback),
    XmRCallback, (XtPointer) NULL
  },
  {
    XmNvalueChangedCallback, XmCCallback, XmRCallback,
    sizeof(XtCallbackList), Offset(value_changed_callback),
    XmRCallback, (XtPointer) NULL
  },
  {
    XmNmodifyVerifyCallback, XmCCallback, XmRCallback,
    sizeof(XtCallbackList), Offset(modify_verify_callback),
    XmRCallback, (XtPointer) NULL
  },
  {
    XmNmodifyVerifyCallbackWcs, XmCCallback, XmRCallback,
    sizeof(XtCallbackList), Offset(wcs_modify_verify_callback),
    XmRCallback, (XtPointer) NULL
  },
  {
    XmNmotionVerifyCallback, XmCCallback, XmRCallback,
    sizeof(XtCallbackList), Offset(motion_verify_callback),
    XmRCallback, (XtPointer) NULL
  },
  {
    XmNgainPrimaryCallback, XmCCallback, XmRCallback,
    sizeof(XtCallbackList), Offset(gain_primary_callback),
    XmRCallback, (XtPointer) NULL
  },
  {
    XmNlosePrimaryCallback, XmCCallback, XmRCallback,
    sizeof(XtCallbackList), Offset(lose_primary_callback),
    XmRCallback, (XtPointer) NULL
  },
  {
    XmNvalue, XmCValue, XmRString,
    sizeof(String), Offset(value),
    XmRString, (XtPointer) NULL	/* FIXME: Motif has a wacky value here */
  },
  {
    XmNvalueWcs, XmCValueWcs, XmRValueWcs,
    sizeof(wchar_t *), Offset(wc_value),
    XmRString, (XtPointer) NULL
  },
  {
    XmNmaxLength, XmCMaxLength, XmRInt,
    sizeof(int), Offset(max_length),
    XmRImmediate, (XtPointer) INT_MAX
  },
  {
    XmNmarginHeight, XmCMarginHeight, XmRVerticalDimension,
    sizeof(Dimension), Offset(margin_height),
    XmRImmediate, (XtPointer) 5
  },
  {
    XmNmarginWidth, XmCMarginWidth, XmRHorizontalDimension,
    sizeof(Dimension), Offset(margin_width),
    XmRImmediate, (XtPointer) 5
  },
  {
    XmNoutputCreate, XmCOutputCreate, XmRFunction,
    sizeof(OutputCreateProc), Offset(output_create),
    XmRFunction, (XtPointer) _XmTextOutputCreate
  },
  {
    XmNinputCreate, XmCInputCreate, XmRFunction,
    sizeof(InputCreateProc), Offset(input_create),
    XmRFunction, (XtPointer) _XmTextInputCreate
  },
  {
    XmNtopCharacter, XmCTopCharacter, XmRTextPosition,
    sizeof(XmTextPosition), Offset(top_character),
    XmRImmediate, (XtPointer) 0
  },
  {
    XmNcursorPosition, XmCCursorPosition, XmRTextPosition,
    sizeof(XmTextPosition), Offset(cursor_position),
    XmRImmediate, (XtPointer) 0
  },
  {
    XmNeditMode, XmCEditMode, XmREditMode,
    sizeof(int), Offset(edit_mode),
    XmRImmediate, (XtPointer) XmSINGLE_LINE_EDIT
  },
  {
    XmNautoShowCursorPosition, XmCAutoShowCursorPosition, XmRBoolean,
    sizeof(Boolean), Offset(auto_show_cursor_position),
    XmRImmediate, (XtPointer) True
  },
  {
    XmNeditable, XmCEditable, XmRBoolean,
    sizeof(Boolean), Offset(editable),
    XmRImmediate, (XtPointer) True
  },
  {
    XmNverifyBell, XmCVerifyBell, XmRBoolean,
    sizeof(Boolean), Offset(verify_bell),
    XtRImmediate, (XtPointer) ((unsigned char) XmUNSPECIFIED)
  },
  {
    XmNnavigationType, XmCNavigationType, XmRNavigationType,
    sizeof(XmNavigationType), XtOffsetOf(XmTextRec, primitive.navigation_type),
    XmRImmediate, (XtPointer) XmTAB_GROUP
  }
};

static void _XmTextSynGetString(Widget wid, int resource_offset, XtArgVal *value);

static XmSyntheticResource syn_resources[] =
{
  {
    XmNmarginWidth,
    sizeof(Dimension), Offset(margin_width),
    _XmFromHorizontalPixels, _XmToHorizontalPixels
  },
  {
    XmNmarginHeight,
    sizeof(Dimension), Offset(margin_height),
    _XmFromVerticalPixels, _XmToVerticalPixels
  },
  {
    XmNvalue,
    sizeof(String), Offset(value),
    _XmTextSynGetString, NULL
  }
};
#undef Offset

static void activate();
static void backward_character();
static void backward_paragraph();
static void backward_word();
static void beep();
static void beginning_of_file();
static void beginning_of_line();
static void clear_selection();
static void copy_clipboard();
static void copy_primary();
static void copy_to();
static void cut_clipboard();
static void cut_primary();
static void delete_next_character();
static void delete_previous_character();
static void delete_next_word();
static void delete_previous_word();
static void delete_selection();
static void delete_to_end_of_line();
static void delete_to_start_of_line();
static void delete_all();
static void deselect_all();
static void do_quick_action();
static void end_of_file();
static void end_of_line();
static void extend_adjust();
static void extend_end();
static void extend_start();
static void forward_character();
static void forward_paragraph();
static void forward_word();
static void grab_focus();
static void focus_in(Widget, XEvent *, String *, Cardinal *);
static void focus_out(Widget, XEvent *, String *, Cardinal *);
static void Help();
static void insert_string();
static void key_select();
static void kill_next_character();
static void kill_next_word();
static void kill_previous_character();
static void kill_previous_word();
static void kill_selection();
static void kill_to_end_of_line();
static void kill_to_start_of_line();
static void move_destination();
static void move_to();
static void newline();
static void newline_and_backup();
static void newline_and_indent();
static void next_line();
static void next_page();
static void next_tab_group();
static void page_left();
static void page_right();
static void paste_clipboard();
static void prev_tab_group();
static void previous_line();
static void previous_page();
static void process_bdrag();
static void process_cancel();
static void process_down();
static void process_up();
static void process_home();
static void process_return();
static void process_shift_down();
static void process_shift_up();
static void process_tab();
static void quick_copy_set();
static void quick_cut_set();
static void redraw_display();
static void scroll_one_line_down();
static void scroll_one_line_up();
static void secondary_adjust();
static void secondary_notify();
static void secondary_start();
static void select_adjust();
static void select_all();
static void select_end();
static void select_start();
static void self_insert();
static void set_anchor();
static void set_insertion_point();
static void set_selection_hint();
static void toggle_add_mode();
static void traverse_home();
static void traverse_next();
static void traverse_prev();
static void unkill();
static void unmap();

static char defaultTranslations[] =
"Ctrl<Key>1:    backward-word()\n" \
"Ctrl<Key>2:    backward-paragraph()\n" \
"Ctrl<Key>3:    beginning-of-line()\n" \
"Ctrl<Key>4:    forward-word()\n" \
"Ctrl<Key>6:	end-of-line()\n" \
"Ctrl<Key>5:    forward-paragraph()\n" \
"Ctrl<Key>A:    beginning-of-line()\n" \
"Ctrl<Key>B:    backward-character()\n" \
"Ctrl<Key>C:    paste-clipboard()\n" \
"Ctrl<Key>D:    delete-next-character()\n" \
"Ctrl<Key>E:	end-of-line()\n" \
"Ctrl<Key>F:    forward-character()\n" \
"Ctrl<Key>H:    delete-previous-character()\n" \
"Ctrl<Key>J:	newline-and-indent()\n" \
"Ctrl<Key>K:    kill-to-end-of-line()\n" \
"Ctrl<Key>L:    redraw-display()\n" \
"Ctrl<Key>M:	newline()\n" \
"Ctrl<Key>N:	next-line()\n" \
"Ctrl<Key>O:	newline-and-backup()\n" \
"Ctrl<Key>P:	previous-line()\n" \
"Ctrl<Key>V:	next-page()\n" \
"Ctrl<Key>W:    kill-selection()\n" \
"Meta<Key>B:    backward-word()\n" \
"Meta<Key>F:    forward-word()\n" \
"Meta<Key>V:    previous-page()\n" \
":Meta<Key>d:	delete-next-word()\n" \
":Meta<Key>h:	delete-previous-word()\n" \
"<Key>Left:     backward-character()\n" \
"<Key>Right:    forward-character()\n" \
"<Key>Up:	previous-line()\n" \
"<Key>Down:	next-line()\n" \
"<Key>End:      end-of-line()\n" \
"<Key>Home:     beginning-of-line()\n" \
"<Key>Return:   process-return()\n" \
"<Key>BackSpace: delete-previous-character()\n" \
"<Key>Delete:   delete-next-character()\n" \
"<Key>:         self-insert()\n" \
"<EnterWindow>:	enter-window()\n" \
"<LeaveWindow>:	leave-window()\n" \
"<Key>osfActivate:	activate()\n" \
"<FocusIn>:	focus-in()\n" \
"<FocusOut>:	focus-out()\n" \
"<Btn1Down>:	select-start()\n" \
"<Btn1Motion>:	extend-adjust()\n" \
"<Btn1Up>:	select-end()\n" \
"<Btn2Down>:	copy-primary()\n";

char _XmText_EventBindings1[] =
"<Key>F2:    backward-word()\n" \
"<Key>F3:    backward-paragraph()\n" \
"<Key>F4:    beginning-of-line()\n" \
"<Key>F5:    forward-word()\n" \
"<Key>F6:    forward-paragraph()\n" \
"<Key>F7:	end-of-line()\n" \
"Ctrl<Key>5:    forward-paragraph()\n" \
"Ctrl<Key>A:    beginning-of-line()\n" \
"Ctrl<Key>B:    backward-character()\n" \
"Ctrl<Key>C:    paste-clipboard()\n" \
"Ctrl<Key>D:    delete-next-character()\n" \
"Ctrl<Key>E:	end-of-line()\n" \
"Ctrl<Key>F:    forward-character()\n" \
"Ctrl<Key>H:    delete-previous-character()\n" \
"Ctrl<Key>J:	newline-and-indent()\n" \
"Ctrl<Key>K:    kill-to-end-of-line()\n" \
"Ctrl<Key>L:    redraw-display()\n" \
"Ctrl<Key>M:	newline()\n" \
"Ctrl<Key>N:	next-line()\n" \
"Ctrl<Key>O:	newline-and-backup()\n" \
"Ctrl<Key>P:	previous-line()\n" \
"Ctrl<Key>V:	next-page()\n" \
"Ctrl<Key>W:    kill-selection()\n" \
"Meta<Key>B:    backward-word()\n" \
"Meta<Key>F:    forward-word()\n" \
"Meta<Key>V:    previous-page()\n" \
":Meta<Key>d:	delete-next-word()\n" \
":Meta<Key>h:	delete-previous-word()\n" \
":m <Key>osfPrimaryPaste: cut-primary()\n\
:a <Key>osfPrimaryPaste: cut-primary()\n\
:<Key>osfPrimaryPaste:	copy-primary()\n\
:m <Key>osfCut:		cut-primary()\n\
:a <Key>osfCut:		cut-primary()\n\
:<Key>osfCut:		cut-clipboard()\n\
:<Key>osfPaste:		paste-clipboard()\n\
:m <Key>osfCopy:	copy-primary()\n\
:a <Key>osfCopy:	copy-primary()\n\
:<Key>osfCopy:		copy-clipboard()\n\
s <Key>osfBeginLine:	beginning-of-line(extend)\n\
:<Key>osfBeginLine:	beginning-of-line()\n\
s <Key>osfEndLine:	end-of-line(extend)\n\
:<Key>osfEndLine:	end-of-line()\n\
s <Key>osfPageLeft:	page-left(extend)\n\
:<Key>osfPageLeft:	page-left()\n\
s c<Key>osfPageUp:	previous-page(extend)\n\
:c <Key>osfPageUp:	previous-page()\n\
s <Key>osfPageRight:	page-right(extend)\n\
:<Key>osfPageRight:	page-right()\n\
s c <Key>osfPageDown:	next-page(extend)\n\
:c <Key>osfPageDown:	next-page()\n\
:<Key>osfClear:		clear-selection()\n\
:<Key>osfBackSpace:	delete-previous-character()\n\
s m <Key>osfDelete:	cut-primary()\n\
s a <Key>osfDelete:	cut-primary()\n\
s <Key>osfDelete:	cut-clipboard()\n\
:c <Key>osfDelete:	delete-to-end-of-line()\n\
:<Key>osfDelete:	delete-next-character()";

char _XmText_EventBindings2[] =
":c m <Key>osfInsert:	copy-primary()\n\
:c a <Key>osfInsert:	copy-primary()\n\
s <Key>osfInsert:	paste-clipboard()\n\
:c <Key>osfInsert:	copy-clipboard()\n\
:s <Key>osfSelect:	key-select()\n\
:<Key>osfSelect:	set-anchor()\n\
:<Key>osfActivate:	activate()\n\
:<Key>osfAddMode:	toggle-add-mode()\n\
:<Key>osfHelp:		Help()\n\
:<Key>osfCancel:	process-cancel()\n\
s c <Key>osfLeft:	backward-word(extend)\n\
:c <Key>osfLeft:	backward-word()\n\
s <Key>osfLeft:		key-select(left)\n\
:<Key>osfLeft:		backward-character()\n\
s c <Key>osfRight:	forward-word(extend)\n\
:c <Key>osfRight:	forward-word()\n\
s <Key>osfRight:	key-select(right)\n\
:<Key>osfRight:		forward-character()\n\
:<Key>osfUp:		previous-line()\n\
:<Key>osfDown:		next-line()\n\
c ~m ~a <Key>slash:	select-all()\n\
c ~m ~a <Key>backslash:	deselect-all()\n\
s ~m ~a <Key>Tab:	prev-tab-group()\n\
~m ~a <Key>Tab:		next-tab-group()\n\
~s ~m ~a <Key>Return:	newline()\n\
c ~s ~m ~a <Key>space:	set-anchor()\n\
c s ~m ~a <Key>space:	key-select()\n\
s ~c ~m ~a <Key>space:	self-insert()\n\
<Key>:			self-insert()";

char _XmText_EventBindings3[] =
"<Unmap>:		unmap()\n\
<Enter>:		enter-window()\n\
<Leave>:		leave-window()\n\
<FocusIn>:		focus-in()\n\
<FocusOut>:		focus-out()\n\
~c s ~m ~a <Btn1Down>:	extend-start()\n\
c ~s ~m ~a <Btn1Down>:	move-destination()\n\
~c ~s ~m ~a <Btn1Down>:	grab-focus()\n\
~c ~m ~a <Btn1Motion>:	extend-adjust()\n\
~c ~m ~a <Btn1Up>:	extend-end()\n\
<Btn2Down>:		process-bdrag()\n\
m ~a <Btn2Motion>:	secondary-adjust()\n\
~m a <Btn2Motion>:	secondary-adjust()\n\
~s <Btn2Up>:		copy-to()\n\
~c <Btn2Up>:		move-to()";

/* action table table */

static XtActionsRec actions[] =
{
  {"activate", activate},
  {"backward-character", backward_character},
  {"backward-paragraph", backward_paragraph},
  {"backward-word", backward_word},
  {"beep", beep},
  {"beginning-of-file", beginning_of_file},
  {"beginning-of-line", beginning_of_line},
  {"clear-selection", clear_selection},
  {"copy-clipboard", copy_clipboard},
  {"copy-primary", copy_primary},
  {"copy-to", copy_to},
  {"cut-clipboard", cut_clipboard},
  {"cut-primary", cut_primary},
  {"delete-next-character", delete_next_character},
  {"delete-previous-character", delete_previous_character},
  {"delete-next-word", delete_next_word},
  {"delete-previous-word", delete_previous_word},
  {"delete-selection", delete_selection},
  {"delete-to-end-of-line", delete_to_end_of_line},
  {"delete-to-start-of-line", delete_to_start_of_line},
  {"delete-all", delete_all},
  {"deselect-all", deselect_all},
  {"do-quick-action", do_quick_action},
  {"end-of-file", end_of_file},
  {"end-of-line", end_of_line},
  {"extend-adjust", extend_adjust},
  {"extend-end", extend_end},
  {"extend-start", extend_start},
  {"forward-character", forward_character},
  {"forward-paragraph", forward_paragraph},
  {"forward-word", forward_word},
  {"focus-in", focus_in},
  {"focus-out", focus_out},
  {"grab-focus", grab_focus},
  {"Help", Help},
  {"insert-string", insert_string},
  {"key-select", key_select},
  {"kill-next-character", kill_next_character},
  {"kill-next-word", kill_next_word},
  {"kill-previous-character", kill_previous_character},
  {"kill-previous-word", kill_previous_word},
  {"kill-selection", kill_selection},
  {"kill-to-end-of-line", kill_to_end_of_line},
  {"kill-to-start-of-line", kill_to_start_of_line},
  {"move-destination", move_destination},
  {"move-to", move_to},
  {"newline", newline},
  {"newline-and-backup", newline_and_backup},
  {"newline-and-indent", newline_and_indent},
  {"next-line", next_line},
  {"next-page", next_page},
  {"next-tab-group", next_tab_group},
  {"page-left", page_left},
  {"page-right", page_right},
  {"paste-clipboard", paste_clipboard},
  {"prev-tab-group", prev_tab_group},
  {"previous-line", previous_line},
  {"previous-page", previous_page},
  {"process-bdrag", process_bdrag},
  {"process-cancel", process_cancel},
  {"process-down", process_down},
  {"process-up", process_up},
  {"process-home", process_home},
  {"process-return", process_return},
  {"process-shift_down", process_shift_down},
  {"process-shift-up", process_shift_up},
  {"process-tab", process_tab},
  {"quick-copy-set", quick_copy_set},
  {"quick-cut-set", quick_cut_set},
  {"redraw-display", redraw_display},
  {"scroll-one-line-down", scroll_one_line_down},
  {"scroll-one-line-up", scroll_one_line_up},
  {"secondary-adjust", secondary_adjust},
  {"secondary-notify", secondary_notify},
  {"secondary-start", secondary_start},
  {"select-adjust", select_adjust},
  {"select-all", select_all},
  {"select-end", select_end},
  {"select-start", select_start},
  {"self-insert", self_insert},
  {"set-anchor", set_anchor},
  {"set-insertion-point", set_insertion_point},
  {"set-selection-hint", set_selection_hint},
  {"toggle-add-mode", toggle_add_mode},
  {"traverse-home", traverse_home},
  {"traverse-next", traverse_next},
  {"traverse-prev", traverse_prev},
  {"unkill", unkill},
  {"unmap", unmap},
};


static XmBaseClassExtRec _XmTextCoreClassExtRec = {
    /* next_extension            */ NULL,
    /* record_type               */ NULLQUARK,                             
    /* version                   */ XmBaseClassExtVersion,
    /* size                      */ sizeof(XmBaseClassExtRec),
    /* initialize_prehook        */ NULL, /* FIXME */
    /* set_values_prehook        */ NULL, /* FIXME */
    /* initialize_posthook       */ NULL, /* FIXME */
    /* set_values_posthook       */ NULL, /* FIXME */
    /* secondary_object_class    */ NULL, /* FIXME */
    /* secondary_object_create   */ NULL, /* FIXME */
    /* get_secondary_resources   */ NULL, /* FIXME */
    /* fast_subclass             */ { 0 }, /* FIXME */
    /* get_values_prehook        */ NULL, /* FIXME */
    /* get_values_posthook       */ NULL, /* FIXME */
    /* class_part_init_prehook   */ NULL, /* FIXME */
    /* class_part_init_posthook  */ NULL, /* FIXME */
    /* ext_resources             */ NULL, /* FIXME */
    /* compiled_ext_resources    */ NULL, /* FIXME */
    /* num_ext_resources         */ 0, /* FIXME */
    /* use_sub_resources         */ FALSE, /* FIXME */
    /* widget_navigable          */ XmInheritWidgetNavigable,
    /* focus_change              */ XmInheritFocusChange,
    /* wrapper_data              */ NULL
};

XmPrimitiveClassExtRec _XmTextPrimClassExtRec =
{
    /* next_extension      */ NULL,
    /* record_type         */ NULLQUARK,
    /* version             */ XmPrimitiveClassExtVersion,
    /* record_size         */ sizeof(XmPrimitiveClassExtRec),
    /* widget_baseline     */ NULL, /* FIX ME */
    /* widget_display_rect */ NULL, /* FIX ME */
    /* widget_margins      */ NULL /* FIX ME */
};

XmTextClassRec xmTextClassRec =
{
    /* Core class part */
  {
	/* superclass            */ (WidgetClass) & xmPrimitiveClassRec,
	/* class_name            */ "XmText",
	/* widget_size           */ sizeof(XmTextRec),
	/* class_initialize      */ ClassInitialize,
	/* class_part_initialize */ ClassPartInitialize,
	/* class_inited          */ FALSE,
	/* initialize            */ Initialize,
	/* initialize_hook       */ NULL,
	/* realize               */ Realize,
	/* actions               */ actions,
	/* num_actions           */ XtNumber(actions),
	/* resources             */ resources,
	/* num_resources         */ XtNumber(resources),
	/* xrm_class             */ NULLQUARK,
	/* compress_motion       */ TRUE,
	/* compress_exposure     */ XtExposeCompressMaximal,
	/* compress_enterleave   */ TRUE,
	/* visible_interest      */ FALSE,
	/* destroy               */ Destroy,
	/* resize                */ Resize,
	/* expose                */ Redisplay,
	/* set_values            */ SetValues,
	/* set_values_hook       */ NULL,
	/* set_values_almost     */ XtInheritSetValuesAlmost,
	/* get_values_hook       */ NULL,
	/* accept_focus          */ NULL,
	/* version               */ XtVersion,
	/* callback offsets      */ NULL,
	/* tm_table              */ _XmText_EventBindings1,
	/* query_geometry        */ QueryGeometry,
	/* display_accelerator   */ XtInheritDisplayAccelerator,
	/* extension             */ (XtPointer) & _XmTextCoreClassExtRec
  },
    /* Primitive Class part */
  {
	/* border_highlight      */ XmInheritBorderHighlight,
	/* border_unhighlight    */ XmInheritBorderUnhighlight,
	/* translations          */ NULL,
	/* arm_and_activate_proc */ XmInheritArmAndActivate,
	/* synthetic resources   */ syn_resources,
	/* num syn res           */ XtNumber(syn_resources),
	/* extension             */ (XtPointer) & _XmTextPrimClassExtRec,
  },
    /* Text Class part */
  {
	/* extension */ NULL
  }
};

WidgetClass xmTextWidgetClass = (WidgetClass) & xmTextClassRec;



/* Inner Widget--------------------------------------------------------------
 * the base for the Output and Input objects
 */


#define Offset(field) XtOffsetOf(XmTextInnerRec, inner.field)

/* Resources for the TextInner class */
static XtResource inner_resources[] =
{
    /* Input Resources */
  {
    XmNpendingDelete, XmCPendingDelete, XmRBoolean,
    sizeof(Boolean), Offset(in.pendingdelete),
    XmRImmediate, (XtPointer) True
  },
  {
    XmNselectionArray, XmCSelectionArray, XmRPointer,
    sizeof(XtPointer), Offset(in.sarray),
    XmRImmediate, (XtPointer) NULL
	/* FIXME: Motif has XmRInt, wacky value here */
  },
  {
    XmNselectionArrayCount, XmCSelectionArrayCount, XmRInt,
    sizeof(int), Offset(in.sarraycount),
    XmRImmediate, (XtPointer) 3
	/* FIXME: Motif has XmRInt, (XtPointer)whacko value here */
  },
  {
    XmNselectThreshold, XmCSelectThreshold, XmRInt,
    sizeof(int), Offset(in.threshold),
    XmRImmediate, (XtPointer) 5
  },

    /* Output Resources */
  {
    XmNblinkRate, XmCBlinkRate, XmRInt,
    sizeof(int), Offset(out.blinkrate),
    XmRImmediate, (XtPointer) 500
  },
  {
    XmNcolumns, XmCColumns, XmRShort,
    sizeof(short), Offset(out.columns),
    XmRImmediate, (XtPointer) 20
  },
  {
    XmNcursorPositionVisible, XmCCursorPositionVisible, XmRBoolean,
    sizeof(Boolean), Offset(out.cursor_position_visible),
    XmRImmediate, (XtPointer) True
  },
  {
    XmNfontList, XmCFontList, XmRFontList,
    sizeof(XmFontList), Offset(out.fontlist),
    XmRImmediate, (XtPointer) XmUNSPECIFIED
  },
  {
    XmNresizeHeight, XmCResizeHeight, XmRBoolean,
    sizeof(Boolean), Offset(out.resizeheight),
    XmRImmediate, (XtPointer) False
  },
  {
    XmNresizeWidth, XmCResizeWidth, XmRBoolean,
    sizeof(Boolean), Offset(out.resizewidth),
    XmRImmediate, (XtPointer) False
  },
  {
    XmNrows, XmCRows, XmRShort,
    sizeof(short), Offset(out.rows),
    XmRImmediate, (XtPointer) 20
  },
  {
    XmNwordWrap, XmCWordWrap, XmRBoolean,
    sizeof(Boolean), Offset(out.wordwrap),
    XmRImmediate, (XtPointer) False
  },

    /* ScrolledText resources */
  {
    XmNscrollHorizontal, XmCScroll, XmRBoolean,
    sizeof(Boolean), Offset(out.scrollhorizontal),
    XmRImmediate, (XtPointer) False
  },
  {
    XmNscrollLeftSide, XmCScrollSide, XmRBoolean,
    sizeof(Boolean), Offset(out.scrollleftside),
    XmRImmediate, (XtPointer) False
  },
  {
    XmNscrollTopSide, XmCScrollSide, XmRBoolean,
    sizeof(Boolean), Offset(out.scrolltopside),
    XmRImmediate, (XtPointer) False
  },
  {
    XmNscrollVertical, XmCScroll, XmRBoolean,
    sizeof(Boolean), Offset(out.scrollvertical),
    XmRImmediate, (XtPointer) False
  },  
};



static void
InnerInitialize(Widget request, Widget new, ArgList args, Cardinal * num_args)
{
  XmTextInnerWidget iw = (XmTextInnerWidget) new;
  XmTextWidget w = (XmTextWidget)XtParent(iw);

  DBGW("InnerInitialize");

}

static void
InnerDestroy(Widget w)
{
    /* Destroy Output and Input */
}

static Boolean
InnerSetValues(Widget old, Widget request, Widget new,
  ArgList args, Cardinal * num_args)
{
  Boolean refresh_needed = False;

  XdbDebug(__FILE__, new, "XmTextInner %s SetValues\n", XtName(new));

  return refresh_needed;
}

XmTextInnerClassRec xmTextInnerClassRec =
{
    /* core_class fields */
  {
    /* superclass               */      (WidgetClass) (&objectClassRec),
    /* class_name               */      "XmTextInner",
    /* widget_size              */      sizeof(XmTextInnerRec),
    /* class_initialize         */      NULL,
    /* class_part_initialize    */      NULL,
    /* class_inited             */      FALSE,
    /* initialize               */      InnerInitialize,
    /* initialize_hook          */      NULL,
    /* obj1                     */      NULL,
    /* obj2                     */      NULL,
    /* obj3                     */      0,
    /* resources                */      inner_resources,
    /* num_resources            */      XtNumber(inner_resources),
    /* xrm_class                */      NULLQUARK,
    /* obj4                     */      FALSE,
    /* obj5                     */      FALSE,
    /* obj6                     */      FALSE,
    /* obj7                     */      FALSE,
    /* destroy                  */      InnerDestroy,
    /* obj8                     */      NULL,
    /* obj9                     */      NULL,
    /* set_values               */      InnerSetValues,
    /* set_values_hook          */      NULL,
    /* obj10                    */      NULL,
    /* get_values_hook          */      NULL,
    /* obj11                    */      NULL,
    /* version                  */      XtVersion,
    /* callback_private         */      NULL,
    /* obj12                    */      NULL,
    /* obj13                    */      NULL,
    /* obj14                    */      NULL,
    /* extension                */      NULL
  },
    /* TextInner Class part */
  {
        /* extension */ NULL
  }
};

WidgetClass xmTextInnerObjectClass = (WidgetClass)&xmTextInnerClassRec;




/* Text-----------------------------------------------------------------------
 *
 * Widget methods
 */



extern Boolean _XmCvtStringToXmString(Display *, XrmValue *, Cardinal *, XrmValue *, XrmValue *, XtPointer *);
extern XmFontList _XmFontListCreateDefault(Display *);

static void
ClassInitialize()
{
  _XmTextCoreClassExtRec.record_type = XmQmotif;
}

static void
ClassPartInitialize(WidgetClass widget_class)
{
  _XmFastSubclassInit(widget_class, XmTEXT_BIT);
}

static void
Initialize(Widget request,
  Widget new,
  ArgList args,
  Cardinal * num_args)
{
  XmTextWidget w = (XmTextWidget) new;

  XtAugmentTranslations((Widget) w, XtParseTranslationTable(_XmText_EventBindings2));
  XtAugmentTranslations((Widget) w, XtParseTranslationTable(_XmText_EventBindings3));

  w->text.inner_widget = XtCreateWidget( "inner", xmTextInnerObjectClass,
                                  new, args, *num_args );

  Text_LineCount(w)=Text_LineMax(w)=0;
  Text_Line(w)=NULL;
  Text_TableSize(w)=Text_TableIndex(w)=0;
  Text_LineTable(w)=NULL;

  Text_TopPos(w)=Text_CursorPos(w)=0;

    /* FIXME: why doesn't it pick this up from the resources? */
  Text_OutputCreate(w)=_XmTextOutputCreate;
  (*Text_OutputCreate(w))((Widget)w,args,*num_args);

    /* FIXME: why doesn't it pick this up from the resources? */
  Text_InputCreate(w)=_XmTextInputCreate;
  (*Text_InputCreate(w))((Widget)w,args,*num_args);
}

static void
Destroy(Widget w)
{
  (*Text_Output(w)->destroy)(w);
  (*Text_Input(w)->destroy)(w);
}

#define	CHECK_CLASS(w)	XmTextCheckClass(w, __FILE__, __LINE__)

static void
XmTextCheckClass(Widget w, char *f, int l)
{
  if (!XtIsSubclass(w, xmTextWidgetClass)) {
    XdbDebug(__FILE__, w, "Don't take LessTif for a fool !\n");
    XdbDebug(__FILE__, w,
      "XmTextField != XmText. Widget %s misused at %s %d\n",
      XtName(w), f, l);
  }
}

static Boolean
SetValues(Widget old,
  Widget request,
  Widget new,
  ArgList args,
  Cardinal * num_args)
{
  Boolean refresh_needed = False;
  XmTextWidget nw = (XmTextWidget) new;
  XmTextWidget ow = (XmTextWidget) old;

  XdbDebug(__FILE__, new, "XmText %s SetValues\n", XtName(new));

  if (Text_Value(nw) != Text_Value(ow)) {
    
    refresh_needed = True;
  }

  if (Text_Editable(nw) != Text_Editable(ow)) {
    _XmTextSetEditable(new, Text_Editable(nw));
    refresh_needed = True;
  }
  /* FIX ME - lots more cases */

  return refresh_needed;
}

static void
Realize(Widget aw, XtValueMask * value_mask, XSetWindowAttributes * attributes)
{
  XmTextWidget w=(XmTextWidget)aw;
  
#define superclass (&xmPrimitiveClassRec)
  (*superclass->core_class.realize) (aw, value_mask, attributes);
#undef superclass

  LineTable(w);
  (*Text_Output(w)->realize)(aw, value_mask, attributes);
  (*Text_Output(w)->expose)(aw, NULL, NULL);
  (*Text_Output(w)->DrawInsertionPoint)(w, Text_CursorPos(w), on);

  _XmTextSetEditable(aw, Text_Editable(w));
}


static void
Redisplay(Widget aw, XEvent * event, Region region)
{
  XmTextWidget w=(XmTextWidget)aw;
  
  if (!XtIsRealized(w))
    return;

  (*Text_Output(w)->expose)(aw, event, region);
  (*Text_Output(w)->DrawInsertionPoint)(w, Text_CursorPos(w), on);

  _XmDrawShadows(XtDisplay(w),
    XtWindow(w),
    Prim_TopShadowGC(w),
    Prim_BottomShadowGC(w),
    Prim_HighlightThickness(w), Prim_HighlightThickness(w),
    XtWidth(w) - 2 * Prim_HighlightThickness(w),
    XtHeight(w) - 2 * Prim_HighlightThickness(w),
    Prim_ShadowThickness(w),
    XmSHADOW_IN);
}

static void
Resize(Widget aw)
{
  XmTextWidget w = (XmTextWidget) aw;

  DBGW("resize");

  (*Text_Output(w)->resize)(aw, True);
  LineTable(w);
}

static XtGeometryResult
QueryGeometry(Widget w, XtWidgetGeometry * proposed, XtWidgetGeometry * answer)
{
  XmTextWidget tw = (XmTextWidget) w;
  XtWidgetGeometry a;

#define Wants(x)   (proposed->request_mode & x)

  if (proposed->request_mode != 0) {	/* NULL case should not yet end here ! */
    if ((!(Wants(CWWidth))) && (!Wants(CWHeight))) {
      /* If they don't ask width/height, let them have whatever they like */
      if (answer)
	*answer = *proposed;
      return XtGeometryYes;
    }
  }

  a.request_mode = CWWidth | CWHeight;
  a.width = XtWidth(tw);
  a.height = XtHeight(tw);

  if (answer)
    *answer = a;

  if (proposed->request_mode == 0)	/* NULL proposed -> return Width+Height */
    return XtGeometryAlmost;

  if (proposed->width >= answer->width && proposed->height >= answer->height)
    return XtGeometryYes;
  else if (answer->width == XtWidth(tw) && answer->height == XtHeight(tw)) {
    if (answer)
      answer->request_mode = 0;
    return XtGeometryNo;
  }
  else
    return XtGeometryAlmost;
}





/* Line Table Routines ---------------------------------------------------- */

static void
LineIncrease(XmTextWidget w,LineNum num)
{
  if (num>Text_LineMax(w)) {
    LineNum start=Text_LineMax(w);
    int i;
    
    Text_Line(w)=(Line)XtRealloc((char *)Text_Line(w),sizeof(LineRec)*num);
    Text_LineMax(w)=num;

    for (i=start; i<num; i++) {
      Line line=&Text_Line(w)[i];

      line->extra=NULL;
    }
  }
}

/*
 * Set up the entire line table, starting with the first visible position.
 *
 * Note that there is 1 extra line in the table.  This is so we can find
 * out what the last position in any line is by looking at the first
 * position of the next line.  Oh, the limitations of Motif compatibility!
 */

static void
LineTable(XmTextWidget w)
{
  XmTextPosition pos,next;
  LineTableExtra extra;
  Boolean more;
  Line line;
  LineNum count;
  
  pos=Text_TopPos(w);
  count=0;
  XdbDebug(__FILE__, w, "LineTable: pos=%d max=%d\n",pos,Text_LineMax(w));
  
  do {
    if (count>=Text_LineMax(w))
      LineIncrease(w,count+16);
    
    line=&Text_Line(w)[count];
    if (line->extra) XtFree((char *)line->extra);
    extra=NULL;
    more=(*Text_Output(w)->MeasureLine)((XmTextWidget)w, count,
      pos, &next, &extra);
    line->start=pos;
    line->changed=False;
    line->changed_position=0;
    line->past_end=(next==PASTENDPOS);
    line->extra=extra;

    if (more) {
      pos=next;
      count++;
    }
  } while (more);

  if (Text_EditMode(w)==XmSINGLE_LINE_EDIT) {
    count=1;
  }
  Text_LineCount(w)=count;
  
  if (XdbInDebug(__FILE__, (Widget)w)) {
    int i;

    for (i=0; i<Text_LineCount(w); i++) {
      Line line=&Text_Line(w)[i];
      
      XdbDebug(__FILE__, (Widget)w,
        "LineTable: line[%d]: start=%d width=%d\n",
        i, line->start, line->extra->width);
    }
  } 
}

static LineNum
LineTableFindNum(XmTextWidget w, XmTextPosition pos)
{
  LineNum i;
  
  if (pos<Text_TopPos(w)) return NOLINE;
  
  for (i=0; i<Text_LineCount(w); i++) {
    Line line=&Text_Line(w)[i];
    Line next=&Text_Line(w)[i+1];

    if (pos>=line->start && pos <next->start) {
      return i;
    }
  }
  
  return NOLINE;
}



/* Scan type utilities --------------------------------------------------- */

static XmTextScanType
ScanType(XmTextWidget w)
{
  InputData i=Text_InputData(w);
  XmTextScanType type;

  if (In_SelArray(i)) {
    if (In_SelArrayIndex(i)>=In_SelArrayCount(i))
      In_SelArrayIndex(i)=0;
    type=In_SelArray(i)[In_SelArrayIndex(i)];
  }
  else {
    switch (In_SelArrayIndex(i)) {
      case 1:
        type=XmSELECT_WORD;
        break;
      case 2:
        type=XmSELECT_ALL;
        break;
      default:
        type=XmSELECT_POSITION;
        In_SelArrayIndex(i)=0;
        break;
    }
  }

  return type;
}






/* High level text insert and delete routines ------------------------------ */

static void
VerifyBell(XmTextWidget w)
{
  if (Text_VerifyBell(w))
    XBell(XtDisplay((Widget) w), 50);
}

static Boolean
DoCursorMove(XmTextWidget w, XEvent * ev, XmTextPosition pos)
{
  XmTextVerifyCallbackStruct cbs;

  if (pos > Text_LastPos(w))
    pos = Text_LastPos(w);

  cbs.doit = True;
  cbs.newInsert = pos;
  if (Text_MotionVerifyCallback(w)) {
    cbs.reason = XmCR_MOVING_INSERT_CURSOR;
    cbs.event = ev;
    cbs.currInsert = Text_CursorPos(w);
    cbs.startPos = cbs.endPos = 0;
    cbs.text = NULL;

    XtCallCallbacks((Widget) w, XmNmotionVerifyCallback, &cbs);

    if (cbs.doit) {
      pos = cbs.newInsert;
    }
  }

  if (cbs.doit) {
    Text_CursorPos(w)=cbs.newInsert;      
    Redisplay(w,NULL,(Region)NULL);
    (*Text_Output(w)->DrawInsertionPoint)(w, Text_CursorPos(w), on);
  }
  else {
    VerifyBell(w);
  }

  return cbs.doit;
}

static void
DoDelete(XmTextWidget w, XEvent * ev, XmTextPosition start, XmTextPosition end)
{
  XmTextVerifyCallbackStruct cbs;
  XmTextBlockRec blockrec;

  if (end <= 0)
    end=0;

  blockrec.ptr = NULL;
  blockrec.length = 0;
  blockrec.format = XmFMT_8_BIT;

  cbs.reason = XmCR_MODIFYING_TEXT_VALUE;
  cbs.event = ev;
  cbs.startPos = cbs.newInsert = start;
  cbs.endPos = end;
  cbs.currInsert = Text_CursorPos(w);
  cbs.text = &blockrec;
  cbs.doit = True;

  if (Text_ModifyVerifyCallback(w)) {
    XtCallCallbacks((Widget) w, XmNmodifyVerifyCallback, &cbs);

    if (!cbs.doit) {
      VerifyBell(w);
    }
  }

  if (cbs.doit) {
    XmTextStatus status;

    (*Text_Output(w)->DrawInsertionPoint)(w, Text_CursorPos(w), off);
    start=cbs.startPos;
    end=cbs.endPos;
    status=(*Text_Source(w)->Replace)((XmTextWidget)w,ev,&start,&end,&blockrec,True);

    LineTable(w);
    if (status==EditDone) {
      Text_CursorPos(w)=cbs.newInsert;      
      Redisplay(w,NULL,(Region)NULL);
    }
    (*Text_Output(w)->DrawInsertionPoint)(w, Text_CursorPos(w), on);
  }

    /* FIXME:  Are we required to free this if it exists?  It can only exist
     * if the user made it, but does Motif think that the user malloced it?
     */
/*   if (blockrec.ptr) XtFree(blockrec.ptr); */
}

static void
DoInsert(XmTextWidget w, XEvent * ev, char *buf, int len)
{
  XmTextVerifyCallbackStruct cbs;
  XmTextBlockRec blockrec;

  if (len <= 0)
    return;

  blockrec.ptr = XtMalloc(len+1);
  blockrec.length = len;
  blockrec.format = XmFMT_8_BIT;
  strncpy(blockrec.ptr, buf, len);
  blockrec.ptr[len]='\0';

  cbs.reason = XmCR_MODIFYING_TEXT_VALUE;
  cbs.event = ev;
  cbs.currInsert = cbs.startPos = cbs.endPos = Text_CursorPos(w);
  cbs.newInsert = Text_CursorPos(w)+=blockrec.length;
  cbs.text = &blockrec;
  cbs.doit = True;
  
  if (Text_ModifyVerifyCallback(w)) {
    XtCallCallbacks((Widget) w, XmNmodifyVerifyCallback, &cbs);

    if (!cbs.doit) {
      VerifyBell(w);
    }
  }

  if (cbs.doit) {
    XmTextStatus status;
    XmTextPosition start,end;

    (*Text_Output(w)->DrawInsertionPoint)(w, Text_CursorPos(w), off);
    start=cbs.startPos;
    end=cbs.endPos;
    status=(*Text_Source(w)->Replace)((XmTextWidget)w,ev,&start,&end,&blockrec,True);
    
    LineTable(w);
    if (status==EditDone) {
      Text_CursorPos(w)=cbs.newInsert;      
      Redisplay(w,NULL,(Region)NULL);
    }
    (*Text_Output(w)->DrawInsertionPoint)(w, Text_CursorPos(w), on);
  }
  XtFree(blockrec.ptr);
}

static void
DoScanType(XmTextWidget w, XEvent *ev, XmTextPosition pos)
{
  InputData i=Text_InputData(w);

  switch (ScanType(w)) {
    case XmSELECT_POSITION:
      DoCursorMove(w, ev, pos);
      
      In_HighlightPivot(i) = Text_CursorPos(w);
      break;
    default:
#ifdef NEW_STUFF
      In_HighlightPivot(i) = In_HighlightStart(i) = ScanTypeStart(w,pos);
      In_HighlightEnd(i) = ScanTypeEnd(w,pos);

        /* FIXME - what event type should this be? */
      PrimarySelectionComplete(w, ev->xbutton.time);
      Draw(w);
#endif
      break;
  }
}




/* Action Routines -------------------------------------------------------- */

static void
activate(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
  XmTextWidget tw = (XmTextWidget) w;
  XmAnyCallbackStruct cb;

  if (Text_EditMode(tw) != XmSINGLE_LINE_EDIT)
    return;

  cb.reason = XmCR_ACTIVATE;
  cb.event = ev;

  XtCallCallbackList(w, tw->text.activate_callback, (XtPointer) & cb);
  if (XmIsManager(XtParent(w))) {
    /* FIXME: Send event to manager. How do we do this ? XSendEvent ? */
  }
}

static void
backward_character(XmTextWidget w, XEvent * ev, String * params, Cardinal * num_params)
{
  DBGW("backward_character");

  if (Text_CursorPos(w) > 0) {
    XmTextPosition pos;

    pos = Text_CursorPos(w) - 1;
    DoCursorMove(w, ev, pos);
  }
}

static void
backward_paragraph(XmTextWidget w, XEvent * ev, String * params, Cardinal * num_params)
{
  DBGW("backward_paragraph");

  if (Text_CursorPos(w) > 0) {
    XmTextPosition pos;
    
    pos=(*Text_Source(w)->Scan)(Text_Source(w),Text_CursorPos(w),
      XmSELECT_WHITESPACE,XmsdLeft,-1,False);
    pos=(*Text_Source(w)->Scan)(Text_Source(w),pos,
      XmSELECT_PARAGRAPH,XmsdLeft,-1,False);
    DoCursorMove(w, ev, pos);
  }
}

static void
backward_word(XmTextWidget w, XEvent * ev, String * params, Cardinal * num_params)
{
  DBGW("backward_word");

  if (Text_CursorPos(w) > 0) {
    XmTextPosition pos;

    pos=(*Text_Source(w)->Scan)(Text_Source(w),Text_CursorPos(w),
      XmSELECT_WHITESPACE,XmsdLeft,-1,False);
    pos=(*Text_Source(w)->Scan)(Text_Source(w),pos,
      XmSELECT_WORD,XmsdLeft,-1,False);
    DoCursorMove(w, ev, pos);
  }
}

static void 
beep()
{
}

static void 
beginning_of_file(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  DBGW("backward_character");

  if (!Text_Editable(w))
    return;

  if (Text_CursorPos(w) > 0) {
    XmTextPosition pos;

    pos = 0;
    DoCursorMove(w, ev, pos);
  }
}

static void 
beginning_of_line(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  XmTextPosition pos;

  DBGW("beginning_of_line");

  pos=(*Text_Source(w)->Scan)(Text_Source(w),Text_CursorPos(w),
    XmSELECT_LINE,XmsdLeft,-1,False);
  DoCursorMove(w, ev, pos);
}

static void 
clear_selection()
{
}

static void 
copy_clipboard()
{
}

static void 
copy_primary()
{
}

static void 
copy_to()
{
}

static void 
cut_clipboard()
{
}

static void 
cut_primary()
{
}

static void 
delete_next_character(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  DBGW("delete_next_character");

  if (!Text_Editable(w)) {
    VerifyBell(w);
  }

  DoDelete(w,ev,Text_CursorPos(w),Text_CursorPos(w)+1);
}

static void 
delete_previous_character(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  DBGW("delete_previous_character");

  if (!Text_Editable(w)) {
    VerifyBell(w);
  }

  DoDelete(w,ev,Text_CursorPos(w)-1,Text_CursorPos(w));
}

static void 
delete_next_word()
{

}

static void 
delete_previous_word()
{
}
static void 
delete_selection()
{
}

static void 
delete_to_end_of_line(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
}

static void 
delete_to_start_of_line(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
}

static void 
delete_all(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
}

static void 
deselect_all(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
}

static void 
do_quick_action()
{
}

static void 
end_of_line(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  XmTextPosition pos;

  DBGW("end_of_line");

  pos=(*Text_Source(w)->Scan)(Text_Source(w),Text_CursorPos(w),
    XmSELECT_LINE,XmsdRight,-1,False);
  DoCursorMove(w, ev, pos);
}

static void 
end_of_file(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  XmTextPosition pos;

  DBGW("end_of_file");

  pos=Text_LastPos(w);
  DoCursorMove(w, ev, pos);
}


static void 
extend_adjust(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  if (ev->type != MotionNotify)
    return;
}


static void 
extend_end(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
}

static void 
extend_start(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
}

static void 
forward_character(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
  DBGW("forward_character");

  if (!Text_Editable(w))
    return;

  if (Text_CursorPos(w) < Text_LastPos(w)) {
    XmTextPosition pos;

    pos = Text_CursorPos(w) + 1;
    DoCursorMove(w, ev, pos);
  }
}

static void 
forward_paragraph(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  XmTextPosition pos;

  DBGW("foreward_paragraph");

  pos=(*Text_Source(w)->Scan)(Text_Source(w),Text_CursorPos(w),
    XmSELECT_PARAGRAPH,XmsdRight,-1,False);
  pos=(*Text_Source(w)->Scan)(Text_Source(w),pos,
    XmSELECT_WHITESPACE,XmsdRight,-1,False);
  DoCursorMove(w, ev, pos);
}


static void 
forward_word(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  XmTextPosition pos;

  DBGW("foreward_word");

  pos=(*Text_Source(w)->Scan)(Text_Source(w),Text_CursorPos(w),
    XmSELECT_WORD,XmsdRight,-1,False);
  pos=(*Text_Source(w)->Scan)(Text_Source(w),pos,
    XmSELECT_WHITESPACE,XmsdRight,-1,False);
  DoCursorMove(w, ev, pos);
}

static void 
grab_focus(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
  OutputData o=Text_OutputData(w);
  InputData i=Text_InputData(w);
  XmTextPosition pos;

  DBGW("grab_focus");

  XmProcessTraversal((Widget) w, XmTRAVERSE_CURRENT);

  pos=(*Text_Output(w)->XYToPos)(w, ev->xbutton.x, ev->xbutton.y);

  if ((In_LastTime(i)+XtGetMultiClickTime(XtDisplay((Widget) w))) >
    ev->xbutton.time)
    In_SelArrayIndex(i)++;
  else
    In_SelArrayIndex(i)=0;

  DoScanType(w,ev,pos);

  In_LastTime(i)=ev->xbutton.time;
}

static void 
Help(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
}

static void 
insert_string(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
}

static void 
key_select()
{
}

static void 
kill_next_character(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  delete_next_character(w, ev, params, num_params);
}

static void 
kill_next_word(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  delete_next_word(w, ev, params, num_params);
}

static void 
kill_previous_character(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  delete_previous_character(w, ev, params, num_params);
}

static void 
kill_previous_word(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  delete_previous_word(w, ev, params, num_params);
}

static void 
kill_selection()
{
}

static void 
kill_to_end_of_line(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  delete_to_end_of_line(w, ev, params, num_params);
}

static void 
kill_to_start_of_line()
{
}

static void 
move_destination()
{
}

static void 
move_to()
{
}

static void 
next_tab_group()
{
}

static void 
next_line(Widget aw, XEvent * ev, String * params, Cardinal * num_params)
{
  XmTextWidget w = (XmTextWidget) aw;
  Position x,y;
  XmTextPosition pos;
  Boolean visible;
  LineNum num;
  
  DBGW("next_line");

  if (!Text_Editable(w))
    return;

  num=LineTableFindNum(w,Text_CursorPos(w));

/* FIXME: This really knows too much about Output & isn't really right anyway*/
  if (num!=NOLINE) {
    OutputData o=Text_OutputData(w);

    visible=(*Text_Output(w)->PosToXY)(w, Text_CursorPos(w), &x, &y);
    y+=Out_FontHeight(o);
    pos=(*Text_Output(w)->XYToPos)(w, x, y);
    DoCursorMove(w, ev, pos);
  }
}

static void 
next_page(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
}

static void
newline(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
  if (!Text_Editable(w))
    return;			/* if we can't edit */
  else if (Text_EditMode(w) == XmSINGLE_LINE_EDIT) {
    activate(w, ev, params, num_params);
    return;
  }

  DoInsert(w, ev, "\n", 1);
}

static void
newline_and_backup(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
  if ((!Text_Editable(w)) || (Text_EditMode(w) == XmSINGLE_LINE_EDIT))
    return;			/* if we can't edit, or we are in single line only mode. */

  DoInsert(w, ev, "\n", 1);
}

static void 
newline_and_indent(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  if ((!Text_Editable(w)) || (Text_EditMode(w) == XmSINGLE_LINE_EDIT))
    return;			/* if we can't edit, or we are in single line only mode. */

  DoInsert(w, ev, "\n", 1);
}

static void 
page_left()
{
}

static void 
page_right()
{
}

static void 
paste_clipboard()
{
}

static void 
prev_tab_group()
{
}


static void 
previous_line(Widget aw, XEvent * ev, String * params, Cardinal * num_params)
{
  XmTextWidget w = (XmTextWidget) aw;
  Position x,y;
  XmTextPosition pos;
  Boolean visible;
  LineNum num;
  
  DBGW("next_line");

  if (!Text_Editable(w))
    return;

  num=LineTableFindNum(w,Text_CursorPos(w));

/* FIXME: This really knows too much about Output & isn't really right anyway*/
  if (num!=NOLINE) {
    OutputData o=Text_OutputData(w);

    visible=(*Text_Output(w)->PosToXY)(w, Text_CursorPos(w), &x, &y);
    y-=Out_FontHeight(o);
    pos=(*Text_Output(w)->XYToPos)(w, x, y);
    DoCursorMove(w, ev, pos);
  }
}

static void 
previous_page()
{
}

static void 
process_bdrag()
{
}

static void 
process_cancel()
{
}

static void 
process_down(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  next_line(w, ev, params, num_params);
}

static void 
process_home(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  beginning_of_line(w, ev, params, num_params);
}

static void 
process_return(Widget w, XEvent * ev, String * params,
  Cardinal * num_params)
{
  newline(w, ev, params, num_params);
}

static void 
process_shift_down()
{
}

static void 
process_shift_up()
{
}

static void 
process_tab()
{
}

static void 
process_up()
{
}

static void 
quick_copy_set()
{
}

static void 
quick_cut_set()
{
}

static void
redraw_display(Widget w,
  XEvent * ev,
  String * params,
  Cardinal * num_params)
{
  Redisplay(w, NULL, (Region) NULL);
}

static void
scroll_one_line_up()
{
}

static void
secondary_adjust()
{
}

static void
secondary_notify()
{
}

static void
secondary_start()
{
}

static void
select_adjust()
{
}

static void
select_all()
{
}

static void
select_start(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
  DBGW("select_start");

  grab_focus(w,ev,params,num_params);
}

static void
select_end(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
}



static void
self_insert(Widget w,XEvent * ev,String * params,Cardinal * num_params)
{
#define INSERTCHARBUFSIZ 32
  char buf[INSERTCHARBUFSIZ];
  KeySym keysym;
  int len, status;

  DBGW("self_insert");

  if (!Text_Editable(w)) {
    VerifyBell(w);
  }

  len = XmImMbLookupString((Widget)w, (XKeyPressedEvent *)ev, buf, INSERTCHARBUFSIZ, &keysym, &status);

  if (XdbInDebug(__FILE__, w)) {
    XdbDebug(__FILE__, w, "XmImMbLookupString => %d\n", len);
    XdbDebug0(__FILE__, w, "\tStatus %s\n",
      (status == XLookupNone) ? "none" :
      (status == XLookupChars) ? "chars" :
      (status == XLookupBoth) ? "both" :
      (status == XLookupKeySym) ? "keysym" :
      (status == XBufferOverflow) ? "overflow" : "????");
    if (status == XLookupBoth || status == XLookupKeySym)
      XdbDebug0(__FILE__, w, "\tKeySym 0x%X\n", keysym);
    if (len > 0) {
      int i;

      XdbDebug0(__FILE__, w, "\tBuffer ");
      for (i = 0; i < len; i++)
	XdbDebug(__FILE__, w, " %X", 0xFF & buf[i]);
      XdbDebug0(__FILE__, w, "\n");
    }
  }

  if (len > 0) {		/* FIXME */
    if (status == XLookupBoth || status == XLookupChars)
      DoInsert(w, ev, buf, len);
  }
}

static void
set_anchor()
{
}

static void
set_insertion_point()
{
}

static void
set_selection_hint()
{
}

static void
scroll_one_line_down()
{
}

static void
toggle_add_mode()
{
}

static void
traverse_home()
{
}

static void
traverse_next()
{
}

static void
traverse_prev()
{
}

static void
unkill()
{
}

static void
unmap()
{
}


static void
focus_in(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
  XmTextVerifyCallbackStruct cbs;
  OutputData o=Text_OutputData(w);

  DBGW("focus_in");

#ifdef  I18N
  if (Text_Editable(w))
    XmImSetFocusValues(w, NULL, 0);
#endif

  if (!Out_HasFocus(o)) {
    if (Text_FocusCallback(w)) {
      cbs.reason = XmCR_FOCUS;
      cbs.event = ev;
      cbs.currInsert = cbs.newInsert = Text_CursorPos(w);
      cbs.startPos = cbs.endPos = 0;
      cbs.text = NULL;

      XtCallCallbackList((Widget) w, Text_FocusCallback(w), &cbs);
    }
    
    XtCallActionProc((Widget) w, "PrimitiveFocusIn", ev, params, *num_params);

    Out_HasFocus(o)=True;
  }
}

static void
focus_out(Widget w, XEvent * ev, String * params, Cardinal * num_params)
{
  XmTextVerifyCallbackStruct cbs;
  OutputData o=Text_OutputData(w);

  DBGW("focus_out");

#ifdef  I18N
  if (Text_Editable(w))
    XmImUnsetFocus(w);
#endif

  if (Out_HasFocus(o)) {
    if (Text_LosingFocusCallback(w)) {
      cbs.reason = XmCR_LOSING_FOCUS;
      cbs.event = ev;
      cbs.currInsert = cbs.newInsert = Text_CursorPos(w);
      cbs.startPos = cbs.endPos = 0;
      cbs.text = NULL;

      XtCallCallbackList((Widget) w, Text_LosingFocusCallback(w), &cbs);
    }

    XtCallActionProc((Widget) w, "PrimitiveFocusOut", ev, params, *num_params);

    Out_HasFocus(o)=False;
  }
}


/* Public Functions ------------------------------------------------------ */

Widget
XmCreateText(Widget parent,char *name,Arg * arglist,Cardinal argCount)
{
  return XtCreateWidget(name, xmTextWidgetClass, parent, arglist, argCount);
}

Widget
XmCreateScrolledText(Widget parent, char *name,Arg * arglist,Cardinal argcount)
{
  Widget sw;
  char *sname;
  Arg a;

  sname = XtMalloc(strlen(name) + 4);
  strcpy(sname, name);
  strcat(sname, "_SW");

  sw = XtCreateWidget(sname, xmScrolledWindowWidgetClass, parent, arglist, argcount);

/* No merge operation: we're sure we can use SetValues on this without harm. */
  XtSetArg(a, XmNscrollingPolicy, XmAPPLICATION_DEFINED);
  XtSetValues(sw, &a, 1);

  XtManageChild(sw);

  return XtCreateWidget(name, xmTextWidgetClass, sw, arglist, argcount);
}

void
XmTextClearSelection(Widget w, Time time)
{
}

Boolean
XmTextCopy(Widget w, Time time)
{
  return False;
}

Boolean
XmTextCut(Widget w, Time time)
{
  return False;
}

Boolean
XmTextFindString(Widget w, XmTextPosition start, char *string,
  XmTextDirection direction, XmTextPosition * position)
{
  return False;
}

int
XmTextGetBaseline(Widget w)
{
  return 0;
}


XmTextPosition
XmTextGetCursorPosition(Widget w)
{
  return Text_CursorPos(w);
}

Boolean
XmTextGetEditable(Widget w)
{
  return Text_Editable(w);
}

XmTextPosition
XmTextGetInsertionPosition(Widget w)
{
  return Text_CursorPos(w);
}

XmTextPosition
XmTextGetLastPosition(Widget w)
{
  return Text_LastPos(w);
}

int
XmTextGetMaxLength(Widget w)
{
  return _XmStringSourceGetMaxLength(Text_Source(w));
}

char *
XmTextGetSelection(Widget w)
{
  return NULL;
}

Boolean
XmTextGetSelectionPosition(Widget w, XmTextPosition * left, XmTextPosition * right)
{
  return False;
}

wchar_t *
XmTextGetSelectionWcs(Widget w)
{
  return NULL;
}


char *
XmTextGetString(Widget w)
{
  return _XmStringSourceGetValue(Text_Source(w),False);
}

wchar_t *
XmTextGetStringWcs(Widget w)
{
  return (wchar_t *)_XmStringSourceGetValue(Text_Source(w),True);
}

/*
 * XmTextGetSubstring()
 * 
 * Gets a substring of the text 
 */
int
XmTextGetSubstring(Widget w, XmTextPosition start, int num_chars, int buffer_size, char *buffer)
{
  int len;
  int retval = XmCOPY_SUCCEEDED;

  if (num_chars > buffer_size)
    len = num_chars;
  else {
    len = buffer_size - 1;
    retval = XmCOPY_TRUNCATED;
  }
/* FIXME */
  
  return retval;
}

int
XmTextGetSubstringWcs(Widget w, XmTextPosition start, int num_chars, int buffer_size, wchar_t * buffer)
{
  return 0;
}

XmTextPosition
XmTextGetTopCharacter(Widget w)
{
  return Text_TopPos(w);
}

void
XmTextInsert(Widget w, XmTextPosition position, char *string)
{
  CHECK_CLASS(w);
}

void
_XmTextInsert(Widget w, XmTextPosition position, char *string, XEvent * evp)
{
  CHECK_CLASS(w);
}

void
XmTextInsertWcs(Widget w, XmTextPosition position, wchar_t * wcstring)
{
}

Boolean
XmTextPaste(Widget w)
{
  return False;
}

Boolean
XmTextPosToXY(Widget w, XmTextPosition position, Position * x, Position * y)
{
  return False;
}

Boolean
XmTextRemove(Widget w)
{
  return False;
}

void
XmTextReplace(Widget w, XmTextPosition from_pos, XmTextPosition to_pos, char *value)
{
}

void
XmTextReplaceWcs(Widget w, XmTextPosition from_pos, XmTextPosition to_pos, wchar_t * wcstring)
{
}

void
XmTextSetAddMode(Widget w, Boolean state)
{
}

void
XmTextSetCursorPosition(Widget w, XmTextPosition position)
{
  Text_CursorPos(w) = position;
  Redisplay(w, NULL, (Region) NULL);
}

static void
_XmTextSetEditable(Widget w, Boolean e)
{
  if (!XtIsRealized(w))
    return;

#ifdef  I18N
  if (e) {			/* becomes editable */
    Arg args[10];
    int nargs;

    XmImRegister(w, 0);

    nargs = 0;
    XtSetArg(args[nargs], XmNbackground, XtBackground(w));
    nargs++;
    XtSetArg(args[nargs], XmNforeground, Prim_Foreground(w));
    nargs++;
    XmImSetValues(w, args, nargs);
  }
  else {			/* Becomes un-editable */
    XmImUnregister(w);
  }
  _XmStringSourceSetEditable(Text_Source(w),e);
}
#endif

void
XmTextSetEditable(Widget w, Boolean editable)
{
  if (!XmIsText(w))
    return;

  if (Text_Editable(w) != editable)
    _XmTextSetEditable(w, editable);

  Text_Editable(w) = editable;
}

void
XmTextSetHighlight(Widget w, XmTextPosition left, XmTextPosition right, XmHighlightMode mode)
{
}

void
XmTextSetInsertionPosition(Widget w, XmTextPosition position)
{
  CHECK_CLASS(w);

  Text_CursorPos(w) = position;
  Redisplay(w, NULL, (Region) NULL);
}

void
XmTextSetMaxLength(Widget w, int max_length)
{
  CHECK_CLASS(w);

  _XmStringSourceSetMaxLength(Text_Source(w),max_length);
}

void
XmTextSetSelection(Widget w, XmTextPosition first, XmTextPosition last, Time time)
{
}

void
XmTextSetString(Widget w, char *value)
{
  CHECK_CLASS(w);

#ifdef	DO_MOSAIC_HACK
/* This is a hack for Mosaic to work ! */
  if (XtIsSubclass(w, xmTextFieldWidgetClass))
    XmTextFieldSetString(w, value);
  else
#endif
  {
  }
  
}

void
XmTextDisableRedisplay(Widget w)
{
  XdbDebug(__FILE__, w, "XmTextDisableRedisplay\n");
}

void
XmTextEnableRedisplay(Widget w)
{
  XdbDebug(__FILE__, w, "XmTextEnableRedisplay\n");
}


XmTextSource
XmTextGetSource(Widget w)
{
  XdbDebug(__FILE__, w, "XmTextGetSource\n");

  return Text_Source(w);
}

void
XmTextSetSource(Widget w, XmTextSource s, XmTextPosition top, XmTextPosition curs)
{
  XdbDebug(__FILE__, w, "XmTextSetSource\n");
}

void
XmTextScroll(Widget w, int n)
{
  XdbDebug(__FILE__, w, "XmTextScroll\n");
}

void
XmTextSetStringWcs(Widget w, wchar_t * wcstring)
{
}

void
XmTextSetTopCharacter(Widget w, XmTextPosition top_character)
{
  CHECK_CLASS(w);
}

void
XmTextShowPosition(Widget w, XmTextPosition position)
{
}

XmTextPosition
XmTextXYToPos(Widget w, Position x, Position y)
{
  return 0;
}

static void
_XmTextSynGetString(Widget w, int resource_offset, XtArgVal *value)
{
	*value = _XmStringSourceGetValue(Text_Source(w),False);
}
