/**
 *
 * $Id: TextOut.c,v 1.4 1996/08/15 02:07:31 miers Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static char rcsid[] = "$Id: TextOut.c,v 1.4 1996/08/15 02:07:31 miers Exp $";

#define	I18N

#include <LTconfig.h>
#include <XmI/XmI.h>
#include <XmI/DebugUtil.h>

#include <Xm/XmP.h>
#include <Xm/TextP.h>
#include <Xm/TextOutP.h>
#include <Xm/ScrolledW.h>
#include <stdio.h>
#include <limits.h>		/* for INT_MAX */
#ifdef HAVE_STRING_H
#include <string.h>
#define ANSI_STRING
#else
#include <strings.h>
#endif
#include <stdlib.h>

#ifndef XmUNSPECIFIED
#define XmUNSPECIFIED (~0)
#endif

static void Draw(XmTextWidget w, LineNum num, XmTextPosition start, XmTextPosition end, XmHighlightMode highlight);

/*
 * Common initialization procedures -------------------------------------------
 */

static void
FontInit(XmTextWidget w)
{
  OutputData o=Text_OutputData(w);
  XmFontListEntry entry = NULL;
  XFontStruct *fs;
  int i;

  DBGW("FontInit");

  for (i=0; Out_FontList(o)[i].tag != NULL; i++)
  {
    if (!strcmp(XmFONTLIST_DEFAULT_TAG, Out_FontList(o)[i].tag))
    {
      entry = &Out_FontList(o)[i];
      break;
    }
  }
  if (!entry)
    Out_FontList(o) = entry = _XmFontListCreateDefault(XtDisplay((Widget)w));

  fs = (XFontStruct *) entry->font;
  Out_Font(o) = fs;
  Out_FontHeight(o) = Out_FontAscent(o) + Out_FontDescent(o);
  Out_FontAverageWidth(o) = (fs->max_bounds.width + fs->min_bounds.width) / 2;
  Out_TabWidth(o)=8*Out_FontAverageWidth(o);
  if (XdbInDebug(__FILE__, (Widget)w)) {
    XdbDebug(__FILE__, (Widget)w,
      "font: ascent=%d descent=%d height=%d avewidth=%d\n",
      Out_FontAscent(o), Out_FontDescent(o),
      Out_FontHeight(o), Out_FontAverageWidth(o));
  }
}


static void
GCInitialize(XmTextWidget w)
{
  OutputData o=Text_OutputData(w);
  XGCValues values;
  XtGCMask mask, dynamic, dontcare;

/*
 * DrawGC
 *      Need to change :        ClipRectangles
 */
  values.line_style = LineSolid;
  values.line_width = 0;
  values.fill_style = FillSolid;
  values.font = Out_Font(o)->fid;
  values.foreground = Prim_Foreground(w);
  values.background = XtBackground(w);
  values.clip_x_origin = 0;
  values.clip_y_origin = 0;
  mask = GCLineStyle | GCLineWidth | GCFillStyle | GCFont | GCForeground | GCBackground;
  dynamic = GCClipMask | GCClipXOrigin | GCClipYOrigin | GCForeground | GCBackground;
  dontcare = 0;
/*   Out_DrawGC(o) = XtAllocateGC((Widget) w, 0, mask, &values, dynamic, dontcare); */
  Out_DrawGC(o) = XCreateGC(XtDisplay(w), XtWindow(w), mask, &values);
  Out_DrawGCInverted(o) = False;

/*
 * Cursor GC
 *      Need to change :        Tile/Stipple Origin, Stipple
 */
  values.line_style = LineSolid;
  values.line_width = 0;
  values.fill_style = FillStippled;
  values.foreground = Prim_Foreground(w);
  values.background = XtBackground(w);
  mask = GCLineStyle | GCLineWidth | GCFillStyle | GCForeground | GCBackground;
  dynamic = GCTileStipXOrigin | GCTileStipYOrigin | GCStipple;
  dontcare = 0;
/*   Out_CursorGC(o) = XtAllocateGC((Widget) w, 0, mask, &values, dynamic, dontcare); */
  Out_CursorGC(o) = XCreateGC(XtDisplay(w), XtWindow(w), mask, &values);

/*
 * CopyGC
 *      At least this one is not changed anywhere :-)
 */
  values.line_style = LineSolid;
  values.line_width = 0;
  values.fill_style = FillSolid;
  values.foreground = XtBackground(w);
  values.background = Prim_Foreground(w);
  mask = GCLineStyle | GCLineWidth | GCFillStyle | GCForeground | GCBackground;
  Out_CopyGC(o) = XtGetGC((Widget) w, mask, &values);
}

static void
GCPixmap(XmTextWidget w)
{
  OutputData o=Text_OutputData(w);
  static char dots[] = {2, 1, 1};
  GC gc;
  int width, width_erase, height;
  Display *display;
  XGCValues values;
  XtGCMask mask;

  values.line_style = LineSolid;
  values.line_width = 0;
  values.fill_style = FillSolid;
  values.foreground = 0;
  values.background = 0;

  display = XtDisplay((Widget) w);
  width = 5;
  height = Out_FontHeight(o);
  width_erase = 2 * Out_Font(o)->max_bounds.width;

  if (Out_CursorIBeam(o))
    XFreePixmap(XtDisplay((Widget)w), Out_CursorIBeam(o));
  Out_CursorIBeam(o) = (Pixmap) NULL;
  if (Out_CursorStipple(o))
    XFreePixmap(XtDisplay((Widget)w), Out_CursorStipple(o));
  Out_CursorStipple(o) = (Pixmap) NULL;
  if (Out_CursorSave(o))
    XFreePixmap(XtDisplay((Widget)w), Out_CursorSave(o));
  Out_CursorSave(o) = (Pixmap) NULL;
  Out_CursorSaveValid(o) = False;

  if (height > 0) {
    Out_CursorIBeam(o) = XCreatePixmap(display,
      RootWindowOfScreen(XtScreen((Widget) w)),
      width, height, 1);
    Out_CursorStipple(o) = XCreatePixmap(display,
      RootWindowOfScreen(XtScreen((Widget) w)),
      width, height, 1);
    Out_CursorSave(o) = XCreatePixmap(display,
      RootWindowOfScreen(XtScreen((Widget) w)),
      width_erase, height, w->core.depth);
    values.line_style = LineSolid;
    values.line_width = 0;
    values.fill_style = FillSolid;
    values.foreground = 0;
    values.background = 0;
    mask = GCLineStyle | GCLineWidth | GCFillStyle | GCForeground | GCBackground;
    gc = XCreateGC(display, Out_CursorIBeam(o), mask, &values);
    XFillRectangle(display, Out_CursorIBeam(o), gc, 0, 0, width, height);
    XFillRectangle(display, Out_CursorStipple(o), gc, 0, 0, width, height);
    XSetForeground(display, gc, 1);
    XDrawLine(display, Out_CursorIBeam(o), gc, 2, 0, 2, height - 2);
    XDrawLine(display, Out_CursorIBeam(o), gc, 0, 0, 4, 0);
    XDrawLine(display, Out_CursorIBeam(o), gc, 0, height - 2, 4, height - 2);

    XSetLineAttributes(display, gc, 0, LineOnOffDash, CapRound, JoinRound);
    XSetDashes(display, gc, 0, &dots[1], (int) dots[0]);
    XDrawLine(display, Out_CursorStipple(o), gc, 2, 0, 2, height - 2);
    XDrawLine(display, Out_CursorStipple(o), gc, 1, 0, 3, 0);
    XDrawLine(display, Out_CursorStipple(o), gc, 1, height - 2, 3, height - 2);

    XFreeGC(display, gc);
  }
}

static void
GCClip(XmTextWidget w)
{
  OutputData o=Text_OutputData(w);
  XRectangle clip;

  clip.x = 0;
  clip.y = 0;
  clip.width = XtWidth(w)-
    2 * Prim_ShadowThickness(w) - 2 * Prim_HighlightThickness(w) -
    2 * Text_MarginWidth(w);
  clip.height = XtHeight(w)-
    2 * Prim_ShadowThickness(w) - 2 * Prim_HighlightThickness(w) -
    2 * Text_MarginHeight(w);

/* These calls change the GCs */
  XSetClipRectangles(XtDisplay((Widget) w), Out_DrawGC(o),
    Out_XDraw(o), Out_YDraw(o) - Out_FontAscent(o), &clip, 1, Unsorted);

  GCPixmap(w);
}


static void
SizeRecalc(XmTextWidget w)
{
  OutputData o=Text_OutputData(w);

  Out_XDraw(o) = Prim_ShadowThickness(w) + Prim_HighlightThickness(w) +
    Text_MarginWidth(w);

    /* Y offset for a row R is calculated as Out_YDraw(o) - R*fontHeight,
       and I add FontAscent here because y positions for fonts are
       measured from the baseline of the font. */
  Out_YDraw(o) = Prim_ShadowThickness(w) + Prim_HighlightThickness(w) +
    Text_MarginHeight(w) + Out_FontAscent(o);
  
  Out_Columns(o) = (XtWidth(w)
    - 2 * Prim_ShadowThickness(w)
    - 2 * Text_MarginWidth(w)) / Out_FontAverageWidth(o);

  if (Text_EditMode(w)==XmSINGLE_LINE_EDIT)
    Out_Rows(o)=1;
  else
    Out_Rows(o) = (XtHeight(w)
      - 2 * Prim_ShadowThickness(w)
      - 2 * Text_MarginHeight(w)) / Out_FontHeight(o);
}




/* Text Drawing functions ------------------------------------------------*/

static void
DrawAll(XmTextWidget w)
{
  int i;
  
  for (i=0; i<Text_LineCount(w); i++) {
    Line line=&Text_Line(w)[i];
    Line next=&Text_Line(w)[i+1];
    
    Draw(w,i,line->start,next->start-1,0);
    
    if (XdbInDebug(__FILE__, (Widget)w)) {
      XdbDebug(__FILE__, (Widget)w,
        "DrawAll: line[%d]: start=%d end=%d\n",
        i, line->start, next->start-1);
    }
  }
}




/* Utilities ------------------------------------------------------------- */


#define Out_NextTabStop(o,x) (Out_TabWidth(o)*((x+Out_TabWidth(o))/Out_TabWidth(o)))
  
/* XTextWidth clone, but taking TABS into account.
 * FIXME: Must start from col 0 (or a tab boundary) for tab position to be
 *        accurate.
 */
int
_XmTextNextX(XmTextWidget w,int x,char *ptr, int len)
{
  OutputData o=Text_OutputData(w);
  XFontStruct *fs=Out_Font(o);
  
  while (len>0) {
    char c;
    
    c=*ptr;
    if (c=='\t') {
      x=Out_NextTabStop(o,x);
    }
    else {
      if (c<fs->min_char_or_byte2 || c>fs->max_char_or_byte2)
        c=fs->default_char;
      if (fs->per_char) {
        c-=fs->min_char_or_byte2;
        x+=fs->per_char[c].width;
      }
      else
        x+=fs->max_bounds.width;
    }
    
    ptr++;
    len--;
  }
  
  return x;
}

static Dimension
FontTextWidth(XmTextWidget w,XmTextPosition pos, XmTextPosition end)
{
  XmTextBlockRec block;
  Dimension width;
  
  (*Text_Source(w)->ReadSource)(Text_Source(w),pos,end,&block);
  width=_XmTextNextX(w,0,block.ptr,block.length);
  XtFree(block.ptr);
  return width;
}

static XmTextPosition
XToPos(XmTextWidget w, LineNum row, int x)
{
  OutputData o=Text_OutputData(w);
  Line line=&Text_Line(w)[row];
  Line next=&Text_Line(w)[row+1];
  XmTextPosition start,end,pos;
  XmTextBlockRec block;
  
  start=line->start;
  end=next->start-1;

    /* x is now the width from the start of the line to the pixel pos */
  x -= Out_XDraw(o) - Out_XOffset(o);
  
  (*Text_Source(w)->ReadSource)(Text_Source(w),start,end,&block);

/* check if the cursor is before the 1st character */
  if (x <= 0) {
    pos = start;
  }

/* OK, how 'bout after the last character */
  else if (x > _XmTextNextX(w,0,block.ptr,block.length)) {
    pos = end;
  }

/* must be in between somewhere... */
  else {
    int tot,count,i;
    
    tot = 0;
    count=end-start;
    pos = -1;
    for (i = 0; i < count; i++) {
      tot = _XmTextNextX(w,tot,&block.ptr[i],1);
      if (x < tot) {
	pos = start + i;
	break;
      }
    }
    if (pos < 0)
      pos = end;
  }

  XtFree(block.ptr);

  if (XdbInDebug(__FILE__, (Widget)w)) {
    XdbDebug(__FILE__, (Widget)w,
      "XToPos: x=%d start=%d end=%d pos=%d\n",x,start,end,pos);
  }
  
  return pos;
}


static LineNum
YToLineNum(XmTextWidget w, Position y)
{
  OutputData o=Text_OutputData(w);
  LineNum line;
  
  line = (LineNum)((y - Prim_ShadowThickness(w) - Prim_HighlightThickness(w) -
    Text_MarginHeight(w)) / Out_FontHeight(o));

  if (line>=Text_LineCount(w)) line=Text_LineCount(w)-1;
  
  if (XdbInDebug(__FILE__, (Widget)w)) {
    XdbDebug(__FILE__, (Widget)w,
      "YToLineNum: y=%d line=%d\n",y,line);
  }
  
  return line;
}

static void
DrawText(XmTextWidget w, int x, int y, char *ptr, int len,
  XmHighlightMode highlight)
{
  OutputData o=Text_OutputData(w);
  int xoff,xstart=x,xend;

  xoff = Out_XDraw(o) + Out_XOffset(o);
  if (highlight==XmHIGHLIGHT_SELECTED && !Out_DrawGCInverted(o)) {
    XSetForeground(XtDisplay(w),Out_DrawGC(o),XtBackground(w));
    XSetBackground(XtDisplay(w),Out_DrawGC(o),Prim_Foreground(w));
    Out_DrawGCInverted(o)=True;
  }
  else if (Out_DrawGCInverted(o)) {
    XSetForeground(XtDisplay(w),Out_DrawGC(o),Prim_Foreground(w));
    XSetBackground(XtDisplay(w),Out_DrawGC(o),XtBackground(w));
    Out_DrawGCInverted(o)=False;
  }
  
  while (len>0) {
    char *start;
    Dimension piecewidth;
    int count;
  
    start=ptr;
    count=0;
    while (len>0 && *ptr!='\t') ptr++,count++,len--;
    if (count>0) {
      piecewidth=_XmTextNextX(w,0,start,count);
      XDrawImageString(XtDisplay(w), XtWindow(w), Out_DrawGC(o),
        xoff+x, y, start, count);
      if (XdbInDebug(__FILE__, (Widget)w)) {
        char temp[256];
        
        strncpy(temp,start,count);
        temp[count]='\0';
        XdbDebug(__FILE__, (Widget)w, "Draw: x=%d y=%d -->%s<--\n",x,y,temp);
      }
    
      x+=piecewidth;
    }
    if (len>0 && *ptr=='\t') {
      x=Out_NextTabStop(o,x);
      ptr++,len--;
    }
  }
  xend=x;
  
  if (highlight==XmHIGHLIGHT_SECONDARY_SELECTED)
    XDrawLine(XtDisplay(w), XtWindow(w), Out_DrawGC(o),
      xoff+xstart, y+1, xoff+xend, y+1);
}




/* Cursor functions -------------------------------------------------------- */

static void
CursorInit(XmTextWidget w)
{
  OutputData o=Text_OutputData(w);
  
  Out_CursorIBeam(o) = (Pixmap) NULL;
  Out_CursorStipple(o) = (Pixmap) NULL;
  Out_CursorSave(o) = (Pixmap) NULL;

  Out_CursorPositionVisible(o) = Out_CursorSaveValid(o) = False;
}

static void
CursorSet(XmTextWidget w, Boolean focus)
{
  OutputData o=Text_OutputData(w);
  
  Out_HasFocus(o) = focus;

    /* These change the GC !! (Danny) */
  if (focus)
    XSetStipple(XtDisplay((Widget) w), Out_CursorGC(o), Out_CursorIBeam(o));
  else
    XSetStipple(XtDisplay((Widget) w), Out_CursorGC(o), Out_CursorStipple(o));
}

static void
CursorSaveUnderIBeam(XmTextWidget w, Position x, Position y)
{
  OutputData o=Text_OutputData(w);

  if (! XtIsRealized((Widget)w))
	return;
    /* Make sure that the text is drawn before saving the cursor */
  XFlush(XtDisplay((Widget) w));

    /* save the area under the cursor */
  XFillRectangle(XtDisplay((Widget) w), Out_CursorSave(o), Out_CopyGC(o),
    0, 0, 5, Out_FontHeight(o));
  XCopyArea(XtDisplay((Widget) w), XtWindow((Widget) w), Out_CursorSave(o),
    Out_CursorGC(o),
    x - 2, y,
    5, Out_FontHeight(o), 0, 0);

  Out_CursorSaveValid(o) = True;
}

static void
CursorRestoreUnderIBeam(XmTextWidget w, Position x, Position y)
{
  OutputData o=Text_OutputData(w);
  
  if (! XtIsRealized((Widget)w))
	return;
    /* Make sure that the text is drawn before erasing the cursor */
  XFlush(XtDisplay((Widget) w));
  
  XCopyArea(XtDisplay((Widget) w), Out_CursorSave(o), XtWindow((Widget) w),
    Out_CursorGC(o), 0, 0, 5, Out_FontHeight(o),
    x - 2, y);
  Out_CursorSaveValid(o) = False;
}

static void
CursorDrawIBeam(XmTextWidget w, int x, int y)
{
  OutputData o=Text_OutputData(w);
  
    /* save the area under the cursor */
  CursorSaveUnderIBeam(w, x, y);

    /* These change the GC !! (Danny) */
  XSetTSOrigin(XtDisplay((Widget) w), Out_CursorGC(o),
    x - 2, y);
  XFillRectangle(XtDisplay((Widget) w), XtWindow((Widget) w), Out_CursorGC(o),
    x - 2, y, 5, Out_FontHeight(o));
}

static void
CursorDraw(XmTextWidget w)
{
  OutputData o=Text_OutputData(w);

  if (Out_CursorPositionVisible(o)) {
    CursorDrawIBeam(w, Out_CursorX(o), Out_CursorY(o));

    Out_BlinkOn(o) = True;
  }
}

static void
CursorErase(XmTextWidget w)
{
  OutputData o=Text_OutputData(w);
  
  if (Out_CursorSaveValid(o)) {
    CursorRestoreUnderIBeam(w, Out_CursorX(o), Out_CursorY(o));
  }
  Out_BlinkOn(o) = False;
}



/* Output Methods --------------------------------------------------------- */

static XmTextPosition
XYToPos(XmTextWidget w, Position x, Position y)
{
  LineNum row=YToLineNum(w,y);

  return XToPos(w,row,x);
}


static Boolean
PosToXY(XmTextWidget w, XmTextPosition pos, Position * x, Position * y)
{
  OutputData o=Text_OutputData(w);
  int i;
  
  if (pos<Text_TopPos(w)) return False;
  
  for (i=0; i<Text_LineCount(w); i++) {
    Line line=&Text_Line(w)[i];
    Line next=&Text_Line(w)[i+1];

    if (pos>=line->start && pos <next->start) {
      *y = Out_YDraw(o) + i*Out_FontHeight(o);
      *x = Out_XDraw(o) + Out_XOffset(o) + FontTextWidth(w,line->start,pos);
      return True;
    }
  }
  
  return False;
}

static void
MakePositionVisible(XmTextWidget w, XmTextPosition pos)
{
}


static Boolean
MoveLines(XmTextWidget w, LineNum from, LineNum to, LineNum dest)
{
  return False;
}

static Boolean
MeasureLine(XmTextWidget w, LineNum num, XmTextPosition pos,
  XmTextPosition * next, LineTableExtraRec ** extra)
{
  OutputData o=Text_OutputData(w);
  Dimension width;
  XmTextPosition end;
  LineTableExtra e;

  if (num>Out_Rows(o) || pos==PASTENDPOS) return False;
  if (pos<Text_LastPos(w))
    end=(*Text_Source(w)->Scan)(Text_Source(w),pos,
      XmSELECT_LINE,XmsdRight,-1,False);
  else
    pos=end=Text_LastPos(w);
  
  if (next) {
    if (end==Text_LastPos(w))
      *next=PASTENDPOS;
    else
      *next=end+1;
  }
  
  width=FontTextWidth(w,pos,end);
    
  if (extra) {
    e=(LineTableExtra)XtMalloc(sizeof(LineTableExtraRec));
    e->width=width;
    e->wrappedbychar=False;
    
    *extra=e;
  }
  
  if (XdbInDebug(__FILE__, (Widget)w)) {
    XdbDebug(__FILE__, (Widget)w,
      "MeasureLine: pos=%d next=%d width=%d\n",
      pos, end, width);
  }

  return True;
}


static void
DrawInsertionPoint(XmTextWidget w, XmTextPosition pos, OnOrOff state)
{
  OutputData o=Text_OutputData(w);
  Position x,y;

#ifdef NOT_YET
    /* erase previously visible cursor */
  if (Out_CursorPositionVisible(o))
    CursorErase(w);
#endif
  
    /* draw new cursor */
  Out_CursorPositionVisible(o)=PosToXY(w,pos,&x,&y);
  if (Out_CursorPositionVisible(o)) {
    Out_CursorX(o)=x;
    Out_CursorY(o)=y - Out_FontAscent(o) + 1;
    CursorDraw(w);
  }
  
  if (XdbInDebug(__FILE__, (Widget)w)) {
    XdbDebug(__FILE__, (Widget)w,
      "DrawInsertionPoint: x=%d pos=%d state=%s\n",
      Out_CursorX(o), pos, (state==on?"on":"off"));
  }
}


static void
Draw(XmTextWidget w, LineNum num, XmTextPosition start, XmTextPosition end,
  XmHighlightMode highlight)
{
  OutputData o=Text_OutputData(w);
  XmTextBlockRec block;
  Line line=&Text_Line(w)[num];

  if (line->past_end) {
    end=Text_LastPos(w);
    if (line->start>end) return;
  }
  (*Text_Source(w)->ReadSource)(Text_Source(w),line->start,end,&block);
  if (block.length > 0) {
    int x,y;
    char *ptr;
    int len;
    
    ptr=block.ptr+start-line->start;
    len=end-start;
    if (start>line->start)
      x=_XmTextNextX(w,0,block.ptr,start-line->start);
    else
      x=0;
    
    y = Out_YDraw(o) + num*Out_FontHeight(o);

    DrawText(w,x,y,ptr,len,highlight);
  }
  XtFree(block.ptr);
}


static void
Invalidate(XmTextWidget w, XmTextPosition start, XmTextPosition end,
  long delta)
{
}

static void
GetPreferredSize(Widget aw, Dimension * width, Dimension * height)
{
  XmTextWidget w = (XmTextWidget) aw;
  OutputData o=Text_OutputData(w);
  
  *width = (2 * Prim_ShadowThickness(w) + 2 * Prim_HighlightThickness(w) +
    2 * Text_MarginWidth(w) +
    Out_Columns(o) * Out_FontAverageWidth(o));

  *height = (2 * Prim_ShadowThickness(w) + 2 * Prim_HighlightThickness(w) +
    2 * Text_MarginHeight(w) +
    Out_Rows(o) * Out_FontHeight(o));
}

static void
GetValues(Widget aw, ArgList args, Cardinal num_args)
{
}

static Boolean
SetValues(Widget old, Widget request, Widget new,
  ArgList args, Cardinal * num_args)
{
  return False;
}

static void
destroy(Widget aw)
{
  XmTextWidget w = (XmTextWidget) aw;
  OutputData o=Text_OutputData(w);
  
  DBGW("destroy");
  XtReleaseGC((Widget) w, Out_DrawGC(o));
  XtReleaseGC((Widget) w, Out_CursorGC(o));
  XtReleaseGC((Widget) w, Out_CopyGC(o));
  if (Out_CursorIBeam(o))
    XFreePixmap(XtDisplay((Widget)w), Out_CursorIBeam(o));
  if (Out_CursorStipple(o))
    XFreePixmap(XtDisplay((Widget)w), Out_CursorStipple(o));
  if (Out_CursorSave(o))
    XFreePixmap(XtDisplay((Widget)w), Out_CursorSave(o));

  XtFree((char *)Text_Output(w));
}

static void
expose(Widget aw,XEvent * event,Region region)
{
  XmTextWidget w = (XmTextWidget) aw;
  
  XClearArea(XtDisplay(w), XtWindow(w),
    Prim_HighlightThickness(w) + Prim_ShadowThickness(w),
    Prim_HighlightThickness(w) + Prim_ShadowThickness(w),
    XtWidth(w) - 2 * (Prim_HighlightThickness(w) + Prim_ShadowThickness(w)),
    XtHeight(w) - 2 * (Prim_HighlightThickness(w) + Prim_ShadowThickness(w)),
    False);
  DrawAll(w);
}

static void
realize(Widget aw, XtValueMask * value_mask, XSetWindowAttributes * attributes)
{
  XmTextWidget w = (XmTextWidget) aw;
  
  GCInitialize(w);
  GCClip(w);
  CursorSet(w,False);
}

static void
resize(Widget aw, Boolean state)
{
  XmTextWidget w = (XmTextWidget) aw;
  OutputData o=Text_OutputData(w);
  
  SizeRecalc(w);

  /* Only adjust the GCs if they have already been created! */
  if (Out_DrawGC(o))
    GCClip(w);
}




OutputRec outputRec =
{
    /* _OutputDataRec            */ NULL,
    /* XYToPosProc               */ XYToPos,
    /* PosToXYProc               */ PosToXY,
    /* MeasureLineProc           */ MeasureLine,
    /* DrawProc                  */ Draw,
    /* DrawInsertionPointProc    */ DrawInsertionPoint,
    /* MakePositionVisibleProc   */ MakePositionVisible,
    /* MoveLinesProc             */ MoveLines,
    /* InvalidateProc            */ Invalidate,
    /* GetPreferredSizeProc      */ GetPreferredSize,
    /* GetValuesProc             */ GetValues,
    /* SetValuesProc             */ SetValues,
    /* XmRealizeOutProc          */ realize,
    /* XtWidgetProc              */ destroy,
    /* XmResizeFlagProc          */ resize,
    /* XtExposeProc              */ expose
};

void
_XmTextOutputCreate(Widget aw, ArgList args, Cardinal num_args)
{
  XmTextWidget w = (XmTextWidget) aw;
  OutputData o;
  XmTextInnerWidget iw = (XmTextInnerWidget) w->text.inner_widget;
  Dimension width,height;

  DBGW("_XmTextOutputCreate");

  Text_Output(w) = (Output) XtMalloc(sizeof(OutputRec));
  memcpy(Text_Output(w), &outputRec, sizeof(OutputRec));

  Text_OutputData(w) = o = &iw->inner.out;
  Out_XOffset(o)=0;
  
  CursorInit(w);
  
  /* Use DrawGC as a flag to see if the GCs have been created */
  Out_DrawGC(o) = NULL;


  /* Get the font information */
  if (Out_FontList(o) == (XmFontList) XmUNSPECIFIED || !Out_FontList(o)) {
    Out_FontList(o) = _XmGetDefaultFontList((Widget) w, XmTEXT_FONTLIST);
    Out_FontListCreated(o) = False;	/* we didn't create this, so don't free it */
  }
  else
    Out_FontListCreated(o) = True;	/* we created this, so free it when the time comes */

  FontInit(w);

    /* Check window dimensions */
  if (XtWidth(w) == (Dimension) 0)
    if (Out_Columns(o)<=0) Out_Columns(o) = 20;
  else
    Out_Columns(o) = (XtWidth(w)
      - 2 * Prim_ShadowThickness(w)
      - 2 * Text_MarginWidth(w)) / Out_FontAverageWidth(o);

  if (XtHeight(w) == (Dimension) 0)
    if (Out_Rows(o)<=0) Out_Rows(o) = 10;
  else
    Out_Rows(o) = (XtHeight(w)
      - 2 * Prim_ShadowThickness(w)
      - 2 * Text_MarginHeight(w)) / Out_FontHeight(o);
  if (Text_EditMode(w)==XmSINGLE_LINE_EDIT) Out_Rows(o)=1;

  GetPreferredSize((Widget)w,&width,&height);
  XtWidth(w) = width;
  XtHeight(w) = height;
  
  SizeRecalc(w);
}



/*
 * Quasi-Public functions -----------------------------------------------------
 */


void
_XmTextDrawDestination(XmTextWidget widget)
{
}

void
_XmTextClearDestination(XmTextWidget widget, Boolean ignore_sens)
{
}

void
_XmTextDestinationVisible(Widget w, Boolean turn_on)
{
}

void
_XmTextChangeBlinkBehavior(XmTextWidget widget, Boolean newvalue)
{
}

void
_XmTextAdjustGC(XmTextWidget w)
{
}

Boolean
_XmTextShouldWordWrap(XmTextWidget w)
{
  OutputData o=Text_OutputData(w);

  return o->wordwrap;
}

Boolean
_XmTextScrollable(XmTextWidget w)
{
  OutputData o=Text_OutputData(w);

  return o->scrollvertical;
}

void
_XmTextOutputGetSecResData(XmSecondaryResourceData * secResDataRtn)
{
}

int
_XmTextGetNumberLines(XmTextWidget w)
{
  return w->text.total_lines;
}

Boolean
_XmTextGetDisplayRect(Widget w, XRectangle * display_rect)
{
  return False;
}

void
_XmTextMarginsProc(Widget w, XmBaselineMargins * margins_rec)
{
}

void
_XmTextChangeHOffset(XmTextWidget widget, int length)
{
}

void
_XmTextToggleCursorGC(Widget widget)
{
}

void
_XmTextFreeContextData(Widget w, XtPointer clientData, XtPointer callData)
{
}

void
_XmTextResetClipOrigin(XmTextWidget tw, XmTextPosition position,
  Boolean clip_mask_reset)
{
}

XmTextPosition
_XmTextFindLineEnd(XmTextWidget widget, XmTextPosition position,
  LineTableExtra * extra)
{
  return 0;
}

void
_XmTextMovingCursorPosition(XmTextWidget tw, XmTextPosition position)
{
}

Boolean
_XmTextGetBaselines(Widget widget, Dimension ** baselines, int *line_count)
{
  return False;
}

